/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.client.multiplayer.api;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

import au.edu.unimelb.csse.mugle.shared.api.KeyError;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.GameInfo;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.GameState;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.Transaction;

import com.google.gwt.user.client.rpc.AsyncCallback;

/**
 * GameManager is the primary class for multi-player games. It is the
 * factory for Message instances, and contains methods for access of the
 * multi-player game services.
 */
public class GameManager {
    private String gameToken;
    private String playerName;
    private boolean isCreator;
    private GameInfo currentGame;
    private GameHandler handler;
    private HashMap<String, GameInfo> games = new HashMap<String, GameInfo>();
    private HashSet<String> players = new HashSet<String>();

    private static Channel channel;

    /**
     * Create a game manager.
     * 
     * @param playerName
     *            The url name for the current player.
     * @param gameToken
     *            The secret token for the current game.
     * @param handler
     *            The {@link GameHandler} for handling events.
     */
    public GameManager(String playerName, String gameToken, GameHandler handler) {
        this.gameToken = gameToken;
        this.playerName = playerName;
        this.handler = handler;
        this.isCreator = false;
        openChannel();
    }

    /**
     * The current player name.
     * 
     * @return the url name of the current player.
     */
    public String playerName() {
        return this.playerName;
    }

    /**
     * The current game instance.
     * 
     * @return information of the current game instance, or null if the player
     *         is not in any game instances.
     */
    public GameInfo currentGame() {
        return currentGame;
    }

    /**
     * Returns true if the player is within a game instance.
     * 
     * @return true if the player is within a game instance.
     */
    public boolean inGame() {
        return this.currentGame != null;
    }

    /**
     * Returns true if the player is the current game creator
     * 
     * @return true if the player is the current game creator.
     */
    public boolean isCreator() {
        return this.isCreator;
    }

    /**
     * The list of players in the current game.
     * 
     * @return the list of players in the current game.
     */
    public List<String> getPlayers() {
        return new ArrayList<String>(this.players);
    }

    /**
     * The player count of the current game.
     * 
     * @return the player count of the current game.
     */
    public int playerCount() {
        return this.players.size();
    }

    /**
     * @param name
     *            the name of the game.
     * @return the game information of the game.
     */
    public GameInfo getGame(String name) {
        return games.get(name);
    }

    /**
     * Create a new Message.
     * 
     * @return a new Message.
     */
    public Message newMessage() {
        Message m = new Message();
        m.setSender(playerName);
        m.setInstanceToken(gameToken);
        if (currentGame != null)
            m.setInstanceToken(currentGame.getToken());
        return m;
    }

    private void openChannel() {
        if (channel == null) {
            channel = new Channel();
            channel.open(new ChannelListener() {

                @Override
                public void onOpen() {
                    handler.onChannelOpen();
                }

                @Override
                public void onMessage(String result) {
                    Message message = new Message(result);
                    if (message.serverCommand() != null)
                        switch (message.serverCommand()) {
                        case PLAYER_CONNECTED:
                            break;
                        case PLAYER_DISCONNECTED:
                            players.remove(message.sender());
                            deductPlayerCount();
                            handler.onPlayerDisconnect(message.sender());
                            break;
                        case PLAYER_JOINED:
                            addPlayerCount();
                            players.add(message.sender());
                            handler.onPlayerJoin(message.sender());
                            break;
                        case PLAYER_QUIT:
                            players.remove(message.sender());
                            deductPlayerCount();
                            handler.onPlayerQuit(message.sender());
                            break;
                        case PLAYER_REMOVED:
                            players.remove(message.sender());
                            if (message.sender().equals(playerName)) {
                                reset();
                            } else {
                                deductPlayerCount();
                            }
                            handler.onPlayerRemove(message.sender());
                            break;
                        case CREATOR_DISCONNECTED:
                            // TODO Remove?
                            break;
                        case NEW_GAME:
                            games.put(message.gameInfo().getName(),
                                    message.gameInfo());
                            handler.onNewGame(message.gameInfo().getName());
                            break;
                        case GAME_UPDATED:
                            games.put(message.gameInfo().getName(),
                                    message.gameInfo());
                            handler.onGameUpdate(message.gameInfo());
                            break;
                        case GAME_REMOVED:
                            games.remove(message.sender());
                            players.clear();
                            handler.onGameRemove(message.sender());
                            break;
                        case TRANSACTION_CREATED:
                            handler.onTransactionCreated(
                                    message.getAttribute("ID"), message.body());
                            break;
                        default:
                            break;
                        }

                    else {
                        // notice the handler on new message
                        handler.onMessage(message);
                    }

                }

                @Override
                public void onError(ChannelError error) {
                    handler.onError(new Throwable("Channel error"));
                    reopenChannel();
                }

                @Override
                public void onClose() {
                    handler.onError(new Throwable("Channel closed"));
                    reopenChannel();
                }
            });
        }
    }

    /**
     * Reopen the channel.
     */
    public void reopenChannel() {
        channel.reOpen();
    }

    /** Get the list of active game instances. */
    public void getGameList(final AsyncCallback<GameInfo[]> callback) {
        Services.multigame.getGameList(gameToken,
                new AsyncCallback<GameInfo[]>() {

                    @Override
                    public void onSuccess(GameInfo[] result) {
                        games.clear();
                        for (GameInfo game : result) {
                            if (game.canContinue()) {
                            }
                            games.put(game.getName(), game);
                        }
                        callback.onSuccess(result);
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        callback.onFailure(caught);
                    }
                });
    }

    private void reset() {
        currentGame = null;
        isCreator = false;
        players.clear();
    }

    // private void print(String line) {
    // System.out.println(line);
    // }
    //
    // private void printError(String error) {
    // System.err.println(error);
    // }

    private void addPlayerCount() {
        currentGame.setPlayerCount(currentGame.getPlayerCount() + 1);
    }

    private void deductPlayerCount() {
        currentGame.setPlayerCount(currentGame.getPlayerCount() - 1);
    }

    /**
     * Create a new game instance.
     * 
     * @param gameToken
     *            The secret token for the current game.
     * @param size
     *            The maximum number of player for the game.
     * @param isPublic
     *            True if the game can be joined public.
     * @param callback
     *            return the name of the game instance.
     */
    public void createGameInstance(final String gameToken, final int size,
            final boolean isPublic, final AsyncCallback<String> callback) {
        Services.multigame.createGameInstance(gameToken, size, isPublic,
                new AsyncCallback<String>() {

                    @Override
                    public void onSuccess(String result) {
                        isCreator = true;
                        GameInfo game = new GameInfo();
                        game.setToken(result);
                        game.setName(playerName);
                        game.setSize(size);
                        game.setPlayerCount(1);
                        games.put(game.getName(), game);
                        currentGame = game;
                        players.add(playerName);
                        callback.onSuccess(game.getName());
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        callback.onFailure(caught);
                    }
                });

    }

    /**
     * Join a previous joined game instance.
     * 
     * @param callback
     *            return a list of player name for this game.
     */
    // public void continueGame(AsyncCallback<String[]> callback) {
    // joinGame(continueGame.getName(), callback);
    // }

    /**
     * Join a game instance.
     * 
     * @param gameName
     *            the name of the game.
     * @param callback
     *            return a list of all the player names for this game.
     */
    public void joinGame(final String gameName,
            final AsyncCallback<String[]> callback) {
        final String token = getGame(gameName).getToken();
        Services.multigame.joinGame(token, new AsyncCallback<String[]>() {

            @Override
            public void onSuccess(String[] playerNames) {
                currentGame = getGame(gameName);
                for (String name : playerNames)
                    players.add(name);

                addPlayerCount();
                callback.onSuccess(playerNames);
            }

            @Override
            public void onFailure(Throwable caught) {
                callback.onFailure(caught);
            }
        });
    }

    /**
     * Exit the current game instance.
     */
    public void quitGame(final AsyncCallback<Void> callback) {
        Services.multigame.quitGame(currentGame.getToken(),
                new AsyncCallback<Void>() {

                    @Override
                    public void onSuccess(Void result) {
                        reset();
                        callback.onSuccess(result);
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        callback.onFailure(caught);
                    }
                });

    }

    /**
     * Remove a player from the current if this player is the game creator.
     * 
     * @param name
     *            the url name of the player.
     */
    public void removePlayer(final String name,
            final AsyncCallback<Void> callback) {
        Services.multigame.removePlayer(currentGame.getToken(), name,
                new AsyncCallback<Void>() {

                    @Override
                    public void onSuccess(Void result) {
                        players.remove(name);
                        callback.onSuccess(result);
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        callback.onFailure(caught);
                    }
                });
    }

    /**
     * Set whether the game can be joined.
     * 
     * @param joinable
     *            true if the game can be joined.
     */
    public void setGameJoinable(boolean joinable, AsyncCallback<Void> callback) {
        Services.multigame.setGameJoinable(currentGame.getToken(), joinable,
                callback);
    }

    /**
     * Retrieve a message with the given counter.
     * 
     * @param counter
     *            the counter of the message.
     * @param callback
     *            return the message.
     */
    public void getMessage(int counter, final AsyncCallback<Message> callback) {
        Services.channel.getMessage(currentGame.getToken(), counter,
                new AsyncCallback<String>() {

                    @Override
                    public void onSuccess(String result) {
                        callback.onSuccess(new Message(result));
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        callback.onFailure(caught);
                    }
                });
    }

    /**
     * Retrieve a list of messages with the given counters.
     * 
     * @param sequence
     *            the counters of the messages.
     * @param callback
     *            return the messages.
     */
    public void getMessages(int[] sequence,
            final AsyncCallback<Message[]> callback) {
        Services.channel.getMessages(currentGame.getToken(), sequence,
                new AsyncCallback<String[]>() {

                    @Override
                    public void onSuccess(String[] result) {
                        Message[] list = new Message[result.length];
                        for (int i = 0; i < result.length; i++)
                            list[i] = new Message(result[i]);
                        callback.onSuccess(list);
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        callback.onFailure(caught);
                    }
                });
    }

    /**
     * Retrieve all the messages of the current game instance.
     * 
     * @param callback
     *            return all the messages of the current game instance.
     */
    public void getAllMessages(final AsyncCallback<Message[]> callback) {
        Services.channel.getAllMessages(currentGame.getToken(),
                new AsyncCallback<String[]>() {

                    @Override
                    public void onSuccess(String[] result) {
                        Message[] list = new Message[result.length];
                        for (int i = 0; i < result.length; i++)
                            list[i] = new Message(result[i]);
                        callback.onSuccess(list);
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        callback.onFailure(caught);
                    }
                });
    }

    /**
     * Store an object in the game key-value store. The value can be accessed by
     * all players in the same game instance.
     * 
     * @param key
     *            The key to associate the object with.
     * @param value
     *            The object to store.
     * @param callback
     *            return true if succeed.
     */
    public void putValue(String key, Serializable value,
            AsyncCallback<Boolean> callback) {
        Services.keyValue
                .putValue(currentGame.getToken(), key, value, callback);
    }

    /**
     * Retrieve an object from the game key-value store.
     * 
     * @param gameToken
     *            The secret token for the current game.
     * @param key
     *            The key the object is associated with.
     * @param callback
     *            return the object associated with the key, or throws
     *            {@link KeyError} if no object is associated with that key.
     */
    public void getValue(String key, AsyncCallback<Serializable> callback) {
        Services.keyValue.getValue(currentGame.getToken(), key, callback);
    }

    /**
     * Create a log for the current game.
     * 
     * @param state
     *            the game state.
     * @param hasLogMessages
     *            whether the log contains messages which are sent during the
     *            game play.
     */
    public void log(GameState state, boolean hasLogMessages,
            AsyncCallback<Void> callback) {
        Services.logging.log(currentGame.getToken(), state, hasLogMessages,
                callback);
    }

    /**
     * Create a log for the current game.
     * 
     * @param transaction
     * 
     * @param callback
     *            the transaction ID.
     */
    public void createTransaction(Transaction transaction,
            AsyncCallback<String> callback) {
        transaction.setInstanceToken(currentGame.getToken());
        Services.transaction.createTransaction(currentGame.getToken(),
                transaction, callback);
    }

    /**
     * Confirm a transaction.
     * 
     * @param transactionId
     * @param callback
     */
    public void confirmTransaction(String transactionId,
            AsyncCallback<Void> callback) {
        Services.transaction.confirmTransaction(currentGame.getToken(),
                transactionId, callback);
    }
}
