/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.client.multiplayer.api;

import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import au.edu.unimelb.csse.mugle.shared.multiplayer.api.*;

import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.web.bindery.autobean.shared.*;

/**
 * Message sent between players.
 */
public class Message implements HasGameState {

    private MessageBean messageBean;

    private final ChannelServiceAsync channelService = Services.channel;
    private MessageBeanFactory factory = Services.messageFactory;
    private String instanceToken;

    /**
     * Message should be created with {@link GameManager#newMessage()}
     */
    protected Message() {
        messageBean = makeMessageBean();
    }

    /**
     * Message should be created with {@link GameManager#newMessage()}
     */
    protected Message(String message) {
        messageBean = deserialiseFromJson(message);
    }

    /**
     * The body of the message.
     * 
     * @return the message body.
     */
    public String body() {
        return messageBean.getBody();
    }

    /**
     * The command of the message.
     * 
     * @return the command of the message.
     */
    public String command() {
        return messageBean.getCommand();
    }

    /**
     * Return an attribute.
     * 
     * @param name
     *            the attribute name.
     */
    public String getAttribute(String name) {
        if (messageBean.getAttributeValuePairs() == null)
            return null;
        else
            return messageBean.getAttributeValuePairs().get(name);
    }

    /**
     * Return the attribute-value pairs.
     * 
     * @return the attribute-value pairs.
     */
    public Map<String, String> getAttributeValuePairs() {
        return messageBean.getAttributeValuePairs();
    }

    /**
     * Return an attribute list.
     * 
     * @param attribute
     *            the attribute of the list.
     */
    public List<String> getList(String attribute) {
        if (messageBean.getAttributeListPairs() == null)
            return null;
        else
            return messageBean.getAttributeListPairs().get(attribute);
    }

    /**
     * Return the attribute-list pairs.
     * 
     * @return the attribute-list pairs.
     */
    public Map<String, List<String>> getAttributeListPairs() {
        return messageBean.getAttributeListPairs();
    }

    public ServerCommand serverCommand() {
        return messageBean.getServerCommand();
    }

    /**
     * The sender of the message.
     * 
     * @return the sender of the message.
     */
    public String sender() {
        return messageBean.getSender();
    }

    /**
     * The receivers of the message.
     * 
     * @return the receivers of the message, or null if the message is sent to
     *         all players in the game instance.
     */
    public Set<String> receivers() {
        return messageBean.getReceivers();
    }

    /**
     * The token of the game instance this message sent from.
     * 
     * @return the token of the game instance this message sent from.
     */
    public String instanceToken() {
        return messageBean.getInstanceToken();
    }

    /**
     * The time this message sent.
     * 
     * @return the time the message sent.
     */
    public Date timestamp() {
        return messageBean.getTimestamp();
    }

    /**
     * Set the message body.
     * 
     * @param body
     *            a short description of this message, for logging purpose.
     */
    public void setBody(String body) {
        messageBean.setBody(body);
    }

    /**
     * Set the command of this message.
     * 
     * @param command
     *            the name to categorised the message.
     */
    public void setCommand(String command) {
        messageBean.setCommand(command);
    }

    /**
     * Set the attribute-value pairs.
     * 
     * @param attributeValuePairs
     *            the attribute-value pairs.
     */
    public void setAttributeValuePairs(Map<String, String> attributeValuePairs) {
        messageBean.setAttributeValuePairs(attributeValuePairs);
    }

    /**
     * Add an attribute.
     * 
     * @param attribute
     *            the attribute.
     * @param value
     *            the value.
     */
    public void addAttribute(String attribute, String value) {
        if (messageBean.getAttributeValuePairs() == null)
            messageBean.setAttributeValuePairs(new HashMap<String, String>());
        messageBean.getAttributeValuePairs().put(attribute, value);
    }

    /**
     * Set the attribute-list pairs.
     * 
     * @param attributeListPairs
     *            the attribute-list pairs.
     */
    public void setAttributeListPairs(
            Map<String, List<String>> attributeListPairs) {
        messageBean.setAttributeListPairs(attributeListPairs);
    }

    /**
     * Add an attribute-list.
     * 
     * @param attribute
     *            the attribute.
     * @param list
     *            the list.
     */
    public void addList(String attribute, List<String> list) {
        if (messageBean.getAttributeListPairs() == null)
            messageBean
                    .setAttributeListPairs(new HashMap<String, List<String>>());
        messageBean.getAttributeListPairs().put(attribute, list);
    }

    protected void setSender(String sender) {
        messageBean.setSender(sender);
    }

    private void setReceivers(Set<String> receivers) {
        messageBean.setReceivers(receivers);
    }

    private void addReceiver(String receiver) {
        if (this.receivers() == null)
            this.setReceivers(new HashSet<String>());
        messageBean.getReceivers().add(receiver);
    }

    private void setTimestamp() {
        messageBean.setTimestamp(new Date());
    }

    /**
     * Return the game information contained in this message when the server
     * command is {@link ServerCommand#NEW_GAME} and
     * {@link ServerCommand#GAME_UPDATED}
     * 
     * @return the game information.
     */
    protected GameInfo gameInfo() {
        if (serverCommand() != null) {
            switch (serverCommand()) {
            case NEW_GAME:
            case GAME_UPDATED:
                GameInfo info = new GameInfo();
                Map<String, String> attributes = getAttributeValuePairs();
                info.setName(attributes.get("name"));
                info.setPlayerCount(Integer.valueOf(attributes
                        .get("playerCount")));
                info.setSize(Integer.valueOf(attributes.get("size")));
                info.setToken(attributes.get("token"));
                return info;
            default:
                break;
            }
        }
        return null;
    }

    /**
     * Send the message to every player in the game instance.
     */
    public void send(AsyncCallback<Void> callback) {
        setTimestamp();
        channelService.sendMessage(serialiseToJson(messageBean), instanceToken,
                callback);
    }

    /**
     * Send the message to a receiver.
     * 
     * @param receiver
     *            the receiver of the message.
     * @param callback
     */
    public void send(String receiver, AsyncCallback<Void> callback) {
        HashSet<String> set = new HashSet<String>();
        addReceiver(receiver);
        set.add(receiver);
        send(set, callback);
    }

    /**
     * Send the message to multiple receivers.
     * 
     * @param receivers
     *            the receivers of the message.
     * @param callback
     */
    public void send(Set<String> receivers, AsyncCallback<Void> callback) {
        setTimestamp();
        setReceivers(receivers);
        channelService.sendMessage(serialiseToJson(messageBean), instanceToken,
                receivers, callback);
    }

    /**
     * reset all the message fields.
     */
    public void clear() {
        this.setBody(null);
        this.setCommand(null);
        this.setAttributeValuePairs(null);
        this.setAttributeListPairs(null);
        this.setReceivers(null);
    }

    private MessageBean makeMessageBean() {
        // Construct the AutoBean
        AutoBean<MessageBean> message = factory.message();

        return message.as();
    }

    private String serialiseToJson(MessageBean message) {
        AutoBean<MessageBean> bean = AutoBeanUtils.getAutoBean(message);

        return AutoBeanCodex.encode(bean).getPayload();
    }

    private MessageBean deserialiseFromJson(String json) {
        AutoBean<MessageBean> bean = AutoBeanCodex.decode(factory,
                MessageBean.class, json);
        return bean.as();
    }

    protected void setInstanceToken(String instanceToken) {
        this.instanceToken = instanceToken;
    }

    /**
     * Return global message counter in a game instance.
     * 
     * @return the global message counter in a game instance.
     */
    public int counter() {
        return this.messageBean.getCounter();
    }

}
