/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.client.ui;

import au.edu.unimelb.csse.mugle.client.Mugle;

import au.edu.unimelb.csse.mugle.shared.model.DevTeam;
import au.edu.unimelb.csse.mugle.shared.model.ModelCollection;
import au.edu.unimelb.csse.mugle.shared.model.User;
import au.edu.unimelb.csse.mugle.shared.model.Game;
import au.edu.unimelb.csse.mugle.shared.model.GameVersion;

import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.user.client.ui.*;

/**
 * Loads the private view for the Developer Team
 * @author Matt Giuca
 *
 */
public class DevTeamViewBuilder extends ViewBuilder {
    /** For the time being, games do not have different versions. Simply
     * assume the following version number. */
    private static final int ASSUME_VERSION = 0;

    public DevTeamViewBuilder(String urlFragment) {
        super(urlFragment);
    }

    @Override
    public Widget buildView() {
        this.clearPanel();
        return loadProfile(this.getDataPanel(), this.getUrlFragment());
    }

    /**
     * @param userID, ID of the user you want to load
     * @return A panel to render
     */
    public FlowPanel loadProfile(final FlowPanel panel, String urlName) {

        final ViewBuilder builder = this;

        DevTeam.getByName(urlName, new AsyncCallback<DevTeam>() {

            public void onFailure(Throwable error) {
                // This will happen only if the team doesn't exist
                // so show the Error 404 page
                panel.clear();
                panel.add(builder.getError404());
            }

            public void onSuccess(DevTeam result) {
                loadDevTeam(panel, result);
            }

        });

        return panel;

    }

    /**
     * Constructs the actual panel
     * @param result
     */
    private Widget loadDevTeam(final FlowPanel panel, final DevTeam devteam) {

        Panel nameEditPanel = new HTMLPanel("");
        panel.add(nameEditPanel);
        if (devteam.getFullName() != null) {
            nameEditPanel.add(new InlineLabel(devteam.getFullName()));
        }
        else
        {
            // No full name; display url name
            nameEditPanel.add(new InlineLabel(devteam.getUrlName()));
        }
        // Add "edit" link, if authorized
        if (devteam.userCanEdit())
        {
            nameEditPanel.add(new InlineLabel(" "));
            nameEditPanel.add(new InlineHyperlink("(edit)",
                "!/" + devteam.getUrlName() + "/+edit"));
        }

        final VerticalPanel userPanel = new VerticalPanel();
        panel.add(userPanel);
        final VerticalPanel gamePanel = new VerticalPanel();
        panel.add(gamePanel);

        devteam.getUsers(new AsyncCallback<ModelCollection<User>>() {
            @Override
            public void onFailure(Throwable caught) {
                Mugle.showError(caught.getMessage());
            }

            @Override
            public void onSuccess(ModelCollection<User> result) {
                if (result.getCount() > 0) {
                    userPanel.add(new Label("The following users are members "
                        + "of this developer team:"));
                }
                else
                {
                    userPanel.add(new Label("This team does not have any users."));
                }
                for (User u : result.getCollection()) {
                    userPanel.add(new Anchor(u.getFullName(), "#!/~" + u.getUrlName()));
                }
            }
        });

        devteam.getGames(false, new AsyncCallback<ModelCollection<Game>>() {
            @Override
            public void onFailure(Throwable caught) {
                Mugle.showError(caught.getMessage());
            }

            @Override
            public void onSuccess(ModelCollection<Game> result) {
                if (result.getCount() > 0) {
                    gamePanel.add(new Label("This team has published the "
                        + "following games:"));
                }
                else
                {
                    gamePanel.add(new Label("This team has not published "
                        + "any games."));
                }
                for (Game g : result.getCollection()) {
                    String fullname = g.getActiveVersion() == null ?
                        g.getUrlName() : g.getActiveVersion().getFullName();
                    gamePanel.add(new Anchor(fullname, "#!/"
                        + devteam.getUrlName() + "/" + g.getUrlName()));
                }
            }
        });

        // Set the window title
        this.setWindowTitle("MUGLE - " + devteam.getFullName());
        this.updateWindowTitle();

        return panel;

    }

}
