/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.client.ui;

import java.util.Arrays;
import java.util.Comparator;

import au.edu.unimelb.csse.mugle.shared.model.*;

import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.user.client.ui.*;

public class GalleryViewBuilder extends ViewBuilder {

    public GalleryViewBuilder() {
        super();
    }

    @Override
    public Widget buildView() {

        this.clearPanel();

        Game.getAllGames(0, 20, new AsyncCallback<ModelCollection<Game>>() {

            @Override
            public void onSuccess(ModelCollection<Game> result) {
                constructView(result);
            }

            @Override
            public void onFailure(Throwable caught) {
                constructErrorView(caught.getMessage());
            }

        });

        // Set the window title
        this.setWindowTitle("MUGLE - Games Gallery");


        return this.getDataPanel();

    }

    private void constructView(ModelCollection<Game> collection) {


        FlowPanel panel = this.getDataPanel();

        Label intro = new Label();
        intro.setText("Welcome to MUGLE, a platform for educational games "
            + "built by the University of Melbourne's Software Modelling and "
            + "Design students, in the third year of Computer Science and "
            + "Software Engineering.");
        panel.add(intro);
        Label label = new Label();
        panel.add(label);

        if (collection.getCount() == 0)
        {
            String text = "There are currently no games hosted on the "
                + "MUGLE platform.";
            label.setText(text);
        }
        else
        {
            String text = "We invite you to try the following games, hosted "
                + "on the MUGLE platform:";
            label.setText(text);
            FlexTable table = new FlexTable();
            panel.add(table);
            table.setText(0, 0, "Game name");
            table.setText(0, 1, "Developer");

            int row = 1;
            Game[] games = sortGames(collection);
            for (Game game : games) {

                GameVersion gv = game.getActiveVersion();
                if (gv != null) {
                    constructGameLink(table, row, game, gv);
                    row++;
                }
            }
        }

    }

    /** Take a collection of games and sort by fullname, then developer key.
     * (Sorting on developer key is arbitrary, but ensures a consistent order
     * for games with the same name.)
     * Any games without an activeVersion will be placed arbitrarily.
     */
    private Game[] sortGames(ModelCollection<Game> collection)
    {
        Game[] games = collection.getCollection().toArray(
            new Game[collection.getCount()]);
        Arrays.sort(games, new Comparator<Game>() {
            @Override
            public int compare(Game g1, Game g2)
            {
                GameVersion g1v = g1.getActiveVersion();
                GameVersion g2v = g2.getActiveVersion();
                String g1n = g1v == null ? "" : g1v.getFullName();
                String g2n = g2v == null ? "" : g2v.getFullName();
                int cmp = g1n.compareTo(g2n);
                if (cmp < 0)
                    return -1;
                else if (cmp > 0)
                    return 1;
                else
                {
                    if (g1.getDevTeamKey() < g2.getDevTeamKey())
                        return -1;
                    else if (g1.getDevTeamKey() > g2.getDevTeamKey())
                        return 1;
                    else
                        return 0;
                }
            }
        });
        return games;
    }

    private void constructGameLink(final FlexTable table, final int row,
        final Game game, final GameVersion gv)
    {
        // XXX Find a way to get the devteam without a server request
        //     for each individual game (LP: #787403).
        DevTeam.getByKey(game.getDevTeamKey(), new AsyncCallback<DevTeam>() {
            @Override
            public void onFailure(Throwable caught) {
                table.setText(row, 0, caught.getMessage());
            }
            @Override
            public void onSuccess(DevTeam dt)
            {
                table.setWidget(row, 0,
                    new Anchor(gv.getFullName(), "#!/" + dt.getUrlName()
                        + "/" + game.getUrlName()));
                table.setText(row, 1, dt.getFullName());
            }
        });
    }

    private void constructErrorView(String message) {

        FlowPanel panel = this.getDataPanel();

        Label label = new Label();
        label.setText(message);

        panel.add(label);

    }

}
