package au.edu.unimelb.csse.mugle.client.ui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;

import au.edu.unimelb.csse.mugle.client.*;
import au.edu.unimelb.csse.mugle.client.model.comparators.AchievementComparator;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.AchievementService;
import au.edu.unimelb.csse.mugle.shared.model.services.AchievementServiceAsync;

import com.google.gwt.core.client.GWT;
import com.google.gwt.event.dom.client.*;
import com.google.gwt.event.logical.shared.*;
import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.user.client.ui.*;
import com.google.gwt.user.client.ui.FlexTable;
import com.google.gwt.user.client.ui.HTMLTable.RowFormatter;
import com.google.gwt.user.client.ui.FormPanel;
import com.google.gwt.user.client.ui.FileUpload;

import com.google.gwt.user.client.Window;

/**
 * Loads the view for editing game information.
 *
 * @author Jens Pfau
 */


/**
 * @author jens
 *
 */
public class GameEditBuilder extends ViewBuilder  {

    /**
     * For the time being, games do not have different versions. Simply assume
     * the following version number.
     */
    private static final int ASSUME_VERSION = 0;
    private final String devteamName;
    private final String gameName;

    // Internal representation of the game information to display and edit.
    private Game game;
    private GameVersion gameVersion;
    private DevTeam devTeam;

    // UI elements for Adding Achievements
    private TextBox addNameBox = new TextBox();
    private TextBox addDisplayNameBox = new TextBox();
    private TextBox addDescriptionBox = new TextBox();
    private TextBox addMaxProgressBox = new TextBox();

    //UI Elements for the Achievements
    private FlowPanel achieveTable = new FlowPanel();
    private HorizontalPanel hp = new HorizontalPanel();
    private Label navLabel = new Label();
    private FlexTable achieveFlexTable = new FlexTable();
    private Button achieveButton = new Button("Create");

    //Internal representation of achievements info
    private ArrayList<Achievement> achievements = new ArrayList<Achievement>();
    private ModelCollection<Achievement> toDelete = new ModelCollection<Achievement>();

    public GameEditBuilder(String devteamName, String gameName) {
        super(gameName);
        this.devteamName = devteamName;
        this.gameName = gameName;
    }


    public Widget buildView() {

        FlowPanel panel = this.getDataPanel();
        panel.clear();

        try {

            this.loadGame(this.devteamName, this.gameName);

        } catch (Exception e) {

            this.showError("The page with URL " + this.getUrlFragment()
                    + " could not be loaded. Reason was: " + e.getMessage());

        }

        return panel;

    }


    /**
     * Method in charge of creating the UI and filling pageWidget.
     */
    private void buildUI(final FlowPanel panel) {

        final ViewBuilder builder = this;

        // Set the window title
        this.setWindowTitle("MUGLE - Editing " + gameVersion.getFullName());
        this.updateWindowTitle();

        // Create table for primitive attributes
        FlexTable flxPrims = new FlexTable();

        // Assemble container panel
        panel.add(flxPrims);

        // Assemble name row
        flxPrims.setText(0, 0, "Display Name:");
        final TextBox nameBox = new TextBox();
        //nameBox.setMaxLength(10);
        nameBox.setVisibleLength(10);
        nameBox.setValue(gameVersion.getFullName());
        flxPrims.setWidget(0, 1, nameBox);
        flxPrims.setText(0,2,"The title of the game, as displayed to users.");

        // Assemble description row
        flxPrims.setText(1, 0, "Description:");
        final TextBox descBox = new TextBox();
        //descBox.setMaxLength(10);
        descBox.setVisibleLength(20);
        descBox.setValue(gameVersion.getDescription());
        flxPrims.setWidget(1, 1, descBox);
        flxPrims.setText(1, 2, "A short summary of the game.");

        // Assemble entry point row
        flxPrims.setText(2, 0, "Entry point:");
        final TextBox entryPointBox = new TextBox();
        //entryPointBox.setMaxLength(10);
        entryPointBox.setVisibleLength(20);
        entryPointBox.setValue(gameVersion.getEntryPoint());
        flxPrims.setWidget(2, 1, entryPointBox);
        flxPrims.setText(2, 2, "Filename of the game's main HTML page "
            + "(e.g., MyGame.html).");

        // Assemble public row
        flxPrims.setText(3, 0, "Public:");
        final CheckBox publicBox = new CheckBox();
        boolean isPublic = game.getIsPublic() == null ? false
                            : game.getIsPublic();
        publicBox.setValue(isPublic);
        flxPrims.setWidget(3, 1, publicBox);
        flxPrims.setText(3, 2, "Check to make game appear in public "
            + "gallery list.");

        flxPrims.setText(4, 0, "Game token:");
        flxPrims.setText(4, 1, game.getGameToken());
        flxPrims.setText(4, 2, "Your game needs to send this token when "
            + "using the developer API.");

        Label warning = new Label("Warning: Do not give this game " +
                                    "token to anybody outside of your developer team.");
        warning = MugleUiBuilder.setWarning(warning);
        panel.add(warning);

        Button btnSave = new Button("Save changes");
        btnSave.addClickHandler(new ClickHandler() {
          public void onClick(ClickEvent event) {
            // Store currently valid properties.
            final String tmpName = gameVersion.getFullName();
            final String tmpDesc = gameVersion.getDescription();
            final String tmpEntry = gameVersion.getEntryPoint();
            final boolean tmpPublic = game.getIsPublic() == null ? false
                                        : game.getIsPublic();

            game.setIsPublic(publicBox.getValue());
            // REQUIRED or game.activeVersion will be set to null
            // See LP: #788075
            game.setActiveVersion(gameVersion);
            // Attempt to submit changes to server.
            game.write(new AsyncCallback<Game>() {
                @Override
                public void onFailure(Throwable error) {
                    // Reset value in textbox and name in internal game
                    // representation.
                    publicBox.setValue(tmpPublic);
                    game.setIsPublic(tmpPublic);
                    // Set the GameVersion textboxes back
                    nameBox.setValue(tmpName);
                    descBox.setValue(tmpDesc);
                    entryPointBox.setValue(tmpEntry);
                    Mugle.showError(error.getMessage());
                }

                @Override
                public void onSuccess(Game result) {
                    gameVersion.setFullName(nameBox.getValue());
                    gameVersion.setDescription(descBox.getValue());
                    gameVersion.setEntryPoint(entryPointBox.getValue());

                    // Attempt to submit changes to server.
                    gameVersion.write(new AsyncCallback<GameVersion>() {
                        @Override
                        public void onFailure(Throwable error) {
                            // Reset value in textbox and name in internal
                            // game version representation.
                            nameBox.setValue(tmpName);
                            gameVersion.setFullName(tmpDesc);
                            descBox.setValue(tmpDesc);
                            gameVersion.setDescription(tmpDesc);
                            entryPointBox.setValue(tmpEntry);
                            gameVersion.setEntryPoint(tmpEntry);
                            Mugle.showError(error.getMessage());
                        }

                        @Override
                        public void onSuccess(GameVersion result) {
                            Mugle.showMessage("Game updated successfully.");
                            builder.setWindowTitle("MUGLE - Editing "
                                + result.getFullName());
                            builder.updateWindowTitle();
                        }
                    });
                }
            });
          }
        });
        panel.add(btnSave);

        panel.add(new Label("Badges"));
        panel.add(assembleAchievementsTable());
        panel.add(assembleUploadBox(gameVersion));
        // Opens in a new tab
        panel.add(new Anchor("Play game", "/+games/" + devTeam.getUrlName()
            + "/" + game.getUrlName() + "/", "_blank"));
    }

    /**
     * Creates a table of achievements associated with this game. The content
     * of the table is editable and an additional row of fields allows the
     * user to create new achievements for this game.
     */
    private Widget assembleAchievementsTable() {
        clearAll();
        achieveTable.add(new Label("Create a Badge"));

        //Add the Panel for Achievement Creation
        FlexTable createPanel = new FlexTable();
        createPanel.setText(0, 0, "Name:");
        createPanel.setText(0, 1, "Display Name");
        createPanel.setText(0, 2, "Description");
        createPanel.setText(0, 3, "Maximum Progress");
        createPanel.setWidget(1, 0, addNameBox);
        createPanel.setWidget(1, 1, addDisplayNameBox);
        createPanel.setWidget(1, 2, addDescriptionBox);
        createPanel.setWidget(1, 3, addMaxProgressBox);
        createPanel.setWidget(1, 4, achieveButton);
        achieveTable.add(createPanel);

        //Now show all the achievements
        showAchievements();
        achieveTable.add(achieveFlexTable);

        // Listen for mouse events on the Add button.
        achieveButton.addClickHandler(new ClickHandler() {
          public void onClick(ClickEvent event) {
            try {
                String t = addMaxProgressBox.getText().trim();
                Integer prog;
                if (t.length() == 0)
                    prog = 0;           // Default to 0 (no progress)
                else
                    prog = Integer.parseInt(t);
                if (prog < 0) {
                    Mugle.showError("Maximum Progress must be 0 or greater.");
                    return;
                }
                addAchievement(addNameBox.getText(), addDisplayNameBox.getText(),
                        addDescriptionBox.getText(), prog);
            } catch (Error e) {
                Mugle.showError(e.getMessage());
            }
          }
        });

        return achieveTable;
    }


    /**
     * Shows the Achievements for the current page
     */
    private void showAchievements() {
        game.getAchievements(new AsyncCallback<ModelCollection<Achievement>>() {

            @Override
            public void onFailure(Throwable caught) {
                  Mugle.showError(caught.getMessage());
            }

            @Override
            public void onSuccess(ModelCollection<Achievement> result) {
                //When we get the DevTeams, create the table
                achieveFlexTable.setStyleName("entities");

                achieveFlexTable.setText(0, 1, "Name");
                achieveFlexTable.setText(0, 2, "Display name");
                achieveFlexTable.setText(0, 3, "Description");
                achieveFlexTable.setText(0, 4, "Maximum progress");
                achieveFlexTable.setText(0, 5, "Edit");

                //Format first row
                RowFormatter rf = achieveFlexTable.getRowFormatter();
                rf.addStyleName(0, "even");
                for (Achievement a : result.getCollection()) {
                    addAchievementToTable(a);
                }

                constructPageNav(result);
            }
        });
    }

    /**
     * Shows an individual Achievement in the table
     * @param a the Achievement to show
     */
    private void showAchievement(Achievement a) {
        RowFormatter rf = achieveFlexTable.getRowFormatter();
        final Achievement temp = a; // Need this for async calls
        final int row = achievements.indexOf(a)+1;

        if (row-1 == -1) {
            /* Note this _should_ never happen - the row is -1 if the DevTeam isnt
             * found in the client side list.  This should be impossible due to the
             * way addDevTeamToTable() works
             */
            Mugle.showError("Error displaying Achievement, does not exist on this page");
            return;
        }

        /* Checkbox adds/removes the achievement to the client side list of Achievements
         * waiting to be deleted
         */
        CheckBox cb = new CheckBox();
        cb.addClickHandler(new ClickHandler() {
            @Override
            public void onClick(ClickEvent event) {
                boolean checked = ((CheckBox)event.getSource()).getValue();
                if (checked) {
                    toDelete.getCollection().add(temp);
                } else {
                    toDelete.getCollection().remove(temp);
                }
            }
        });

        //Sets the data in each row
        // Removed temporarily due to LP: #786418
        //achieveFlexTable.setWidget(row, 0, cb);
        achieveFlexTable.setText(row, 1, a.getName());

        // Create text boxes for editing
        final TextBox disp = new TextBox();
        disp.setText(a.getDisplayName());

        final TextBox desc = new TextBox();
        desc.setText(a.getDescription());

        final TextBox maxProg = new TextBox();
        maxProg.setText(Integer.toString(a.getMaxProgress()));

        Button edit = new Button("Edit");

        achieveFlexTable.setWidget(row, 2, disp);
        achieveFlexTable.setWidget(row, 3, desc);
        achieveFlexTable.setWidget(row, 4, maxProg);
        achieveFlexTable.setWidget(row, 5, edit);

        edit.addStyleDependentName("edit");
        edit.addClickHandler(new ClickHandler() {
            public void onClick(ClickEvent event) {
                temp.setDisplayName(disp.getText());
                temp.setDescription(desc.getText());
                try {
                    String t = maxProg.getText().trim();
                    Integer prog;
                    if (t.length() == 0)
                        prog = 0;           // Default to 0 (no progress)
                    else
                        prog = Integer.parseInt(t);
                    if (prog < 0) {
                        Mugle.showError("Maximum Progress must be 0 or greater.");
                        return;
                    }
                    temp.setMaxProgress(prog);
                } catch (Error e) {
                    Mugle.showError(e.getMessage());
                }
                temp.write(new AsyncCallback<Achievement>() {
                    @Override
                    public void onFailure(Throwable caught) {
                        Mugle.showError(caught.getMessage());
                    }

                    @Override
                    public void onSuccess(Achievement result) {
                        Mugle.showMessage(result.getName() + " succesfully updated!");
                        achievements.remove(row-1);
                        achieveFlexTable.removeRow(row);
                        addAchievementToTable(result);
                    }

                });

            }
        });
        // format the rows
        String style = ((row%2)!=0) ? "odd" : "even";
        rf.addStyleName(row, style);
    }

    /**
     * Adds an Achievement to the table being rendered - It is inserted
     * to the correct position on the table
     * @param a the Achievement to add
     */
    private void addAchievementToTable(Achievement a) {
        achievements.add(a);
        updateNavLabel();
        // Need to sort the client side list by UrlName
        Collections.sort(achievements, new AchievementComparator());

        int row = achievements.indexOf(a);

        /* Need to update all the achievements row numbers after the
         * one inserted
         */
        for (int i = row; i < achievements.size(); i++) {
            showAchievement(achievements.get(i));
        }
    }

    /**
     * Removes all the Achievements in the client side toDelete list from
     * the table rendered.  This should only be called from removeAchievements()
     * which removes the Achievements from the datastore
     */
    private void removeAchievementsFromTable() {
        //Remove each DevTeam from the table, and the client side list
        for (Achievement a : toDelete.getCollection()) {
            int row = achievements.indexOf(a);
            achievements.remove(row);
            achieveFlexTable.removeRow(row+1);
        }
        toDelete.getCollection().clear();

        //reformat the rows
        RowFormatter rf = achieveFlexTable.getRowFormatter();
        for (int row = 0; row <= achievements.size(); row++) {
            String style = ((row%2)!=0) ? "odd" : "even";
            rf.setStyleName(row, style);
        }
        updateNavLabel();
    }

    /**
     * Creates a new Achievement and renders it in the table
     * @param text
     */
    private void addAchievement(String name, String Fullname,
            String description, Integer maxProgress) {
        Mugle.hideMessage();
        Achievement achievement = new Achievement();
        achievement.setName(name);
        achievement.setDisplayName(Fullname);
        achievement.setCreatedUser(Mugle.getAuthUser().getUser().getPrimaryKey());
        achievement.setCreatedTime(new Date());
        achievement.setGameKey(game.getPrimaryKey());
        achievement.setMaxProgress(maxProgress);
        achievement.setDescription(description);

        achievement.create(new AsyncCallback<Achievement>() {
            public void onFailure(Throwable caught) {
                Mugle.showError(caught.getMessage());
            }

            public void onSuccess(Achievement result) {
                addNameBox.setText("");
                addDisplayNameBox.setText("");
                addDescriptionBox.setText("");
                addMaxProgressBox.setText("");
                Mugle.showMessage("Achievement '" + result.getName() + "' successfully created!");
                addAchievementToTable(result);
            }
        });
    }

    /**
     * removes the Achievements in the list of Achievements generated by the checkboxes
     */
    private void removeAchievements() {
        Mugle.hideMessage();

        AchievementServiceAsync as = GWT.create(AchievementService.class);
        as.remove(toDelete, new AsyncCallback<Void>() {
            public void onFailure(Throwable caught) {
                Mugle.showError(caught.getMessage());
            }
            public void onSuccess(Void result) {
                removeAchievementsFromTable();
            }
        });
    }

    /**
     * Creates a widget allowing the user to upload a new game in a zip file.
     * @param version Game version to upload over the top of.
     */
    private Widget assembleUploadBox(GameVersion version) {
        final FormPanel form = new FormPanel();
        form.setAction("/mugle/upload?gameversion="
            + Long.toString(version.getPrimaryKey()));

        // File upload widget requires multipart/form-data encoding, and POST
        // method.
        form.setEncoding(FormPanel.ENCODING_MULTIPART);
        form.setMethod(FormPanel.METHOD_POST);

        VerticalPanel vpanel = new VerticalPanel();
        form.setWidget(vpanel);
        vpanel.add(new Label("Upload the game code. Instructions: Compile "
        + "your GWT project and then make a copy of the 'war' directory, "
        + "called 'war-client'. In war-client, delete the WEB-INF directory "
        + "-- it isn't needed and is far too big. Then zip up the war-client "
        + "directory. Your zip file must have a single top-level directory. "
        + "Upload the zip file here. This will replace any existing versions "
        + "of the game."));

        HorizontalPanel hpanel = new HorizontalPanel();
        vpanel.add(hpanel);

        // Create a FileUpload widget
        FileUpload upload = new FileUpload();
        upload.setName("game_upload");
        hpanel.add(upload);

        // Add a 'submit' button
        hpanel.add(new Button("Upload", new ClickListener() {
          public void onClick(Widget sender) {
            form.submit();
          }
        }));


        // Handle the response from the server after upload
        form.addFormHandler(new FormHandler() {
          public void onSubmit(FormSubmitEvent event) {}

          public void onSubmitComplete(FormSubmitCompleteEvent event) {
            Window.alert(event.getResults());
          }
        });

        return form;
    }

    /**
     * Constructs the Navigation bar down the bottom (with delete button)
     */
    private void constructPageNav(ModelCollection<Achievement> result) {
        // add data details

        hp.setStylePrimaryName("entities-control");
        updateNavLabel();
        // Add a remove button for the DevTeams
        /* Removed temporarily due to LP: #786418
        Button removeAchievementButton = new Button("Delete");
        removeAchievementButton.addStyleDependentName("remove");
        removeAchievementButton.addClickHandler(new ClickHandler() {
            public void onClick(ClickEvent event) {
                Mugle.showMessage("Deleting!");
                removeAchievements();
            }
        });
        */

        /* The Horizontal Panel down the bottom with the delete button
         */
        //hp.add(removeAchievementButton);
        hp.add(navLabel);

        achieveTable.add(hp);
    }

    /**
     * Updates the Label showing "Showing results X-Y" based on the
     * current page number of results
     */
    private void updateNavLabel() {
        String text = "Showing results: 1-" + Integer.toString(achievements.size());
        navLabel.setText(text);
    }

    private void clearAll() {
        achieveTable.clear();
        achievements.clear();
        toDelete.getCollection().clear();
        hp.clear();
        navLabel.setText("");
        achieveFlexTable.clear();
        achieveFlexTable.removeAllRows();
        addNameBox.setText("");
        addDisplayNameBox.setText("");
        addDescriptionBox.setText("");
        addMaxProgressBox.setText("");
    }

    /**
     * Loads the game data from the server.
     * @param urlname The unique name of the game to be loaded.
     */
    private void loadGame(final String devteamName, final String urlname) {

        final ViewBuilder builder = this;

        // Try to load game data
        Game.getByName(devteamName, urlname, new AsyncCallback<Game>() {
            public void onFailure(Throwable error) {
                Mugle.showError("The game with URL " + urlname
                        + " could not be loaded. Reason was: "
                        + error.getMessage());
            }

            public void onSuccess(final Game resultGame) {
                game = resultGame;
                // Check that the user has the privilege to edit this object
                if (resultGame.userCanEdit())
                {
                    loadGameVersion();
                }
                else
                {
                    builder.getDataPanel().add(
                        builder.getAuthorizationError());
                }
            }
        });
    }



    /**
     * Loads the most recent version of the game retrieved in
     * {@ #loadGame()}.
     */
    private void loadGameVersion() {
        // Get the GameVersion for hard-coded version ASSUME_VERSION
        GameVersion.getByVersion(game, ASSUME_VERSION,
            new AsyncCallback<GameVersion>() {
                public void onFailure(Throwable error) {
                    Mugle.showError("The version of the game " +
                            "with URL " + game.getUrlName() +
                            " could not be loaded. Reason was: "
                            + error.getMessage());
                }

                public void onSuccess(GameVersion resultVersion) {
                    gameVersion = resultVersion;
                    loadDevTeam();
                }
            });
    }

    /**
     * Loads the devteam.
     */
    private void loadDevTeam() {
        // Get the GameVersion for hard-coded version ASSUME_VERSION
        DevTeam.getByKey(game.getDevTeamKey(),
            new AsyncCallback<DevTeam>() {
                public void onFailure(Throwable error) {
                    Mugle.showError(error.getMessage());
                }

                public void onSuccess(DevTeam resultTeam) {
                    devTeam = resultTeam;
                    loadAchievements();
                }
            });
    }

    /**
     * Loads the achievements associated with the game retrieved in
     * {@ #loadGame()}.
     */
    private void loadAchievements() {

        final ViewBuilder builder = this;

        // If the game version has been retrieved successfully,
        // also retrieve the achievements associated with this game.
        this.game.getAchievements(
                new AsyncCallback<ModelCollection<Achievement>>() {

           @Override
           public void onFailure(Throwable error) {
               builder.showError("The achievements of the " +
                        "game with URL " + game.getUrlName() +
                        " could not be loaded. Reason was: "
                        + error.getMessage());
            }

            @Override
            public void onSuccess(ModelCollection<Achievement> result) {

                achievements = new ArrayList<Achievement>();
                achievements.addAll(result.getCollection());
                // When the game, its version, and the achievements have been
                // retrieved successfully, create the UI.
                buildUI(builder.getDataPanel());

            }
        });
    }

}
