/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.client.ui;

import au.edu.unimelb.csse.mugle.shared.model.Game;
import au.edu.unimelb.csse.mugle.shared.model.GameVersion;
import au.edu.unimelb.csse.mugle.shared.model.DevTeam;

import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.user.client.ui.*;

/**
 * Loads the private view for the Game Panel
 * @author Matt Giuca
 *
 */
public class GameViewBuilder extends ViewBuilder {

    /** For the time being, games do not have different versions. Simply
     * assume the following version number. */
    private static final int ASSUME_VERSION = 0;
    private final String devteamName;
    private final String gameName;

    public GameViewBuilder(String devteamName, String gameName) {
        super(gameName);
        this.devteamName = devteamName;
        this.gameName = gameName;
    }

    @Override
    public Widget buildView() {

        this.clearPanel();
        try {
            Long pKey = Long.parseLong(this.getUrlFragment());
            return loadProfile(this.getDataPanel(), pKey);
        } catch (NumberFormatException e) {
            return loadProfile(this.getDataPanel(), devteamName, gameName);
        }

    }

    /**
     * @param gameID, ID of the game you want to load
     * @return A panel to render
     */
    public Widget loadProfile(final FlowPanel panel, Long gameID) {

        final GameViewBuilder builder = this;

        Game.getByKey(gameID, new AsyncCallback<Game>() {

            public void onFailure(Throwable error) {
                builder.showError(error.getMessage());
            }

            public void onSuccess(final Game resultGame) {
                builder.loadProfile(panel, resultGame);
            }

        });

        return panel;

    }

    /**
     * @param urlname, Url name of the game you want to load
     * @return A panel to render
     */
    public Widget loadProfile(final FlowPanel panel, String devteamName,
        String urlname) {

        final GameViewBuilder builder = this;

        Game.getByName(devteamName, urlname, new AsyncCallback<Game>() {

            public void onFailure(Throwable error) {
                builder.showError(error.getMessage());
            }

            public void onSuccess(final Game resultGame) {
                builder.loadProfile(panel, resultGame);
            }

        });

        return panel;

    }

    /**
     * @param game, The game you want to load
     * @return A panel to render
     */
    public Widget loadProfile(final FlowPanel panel, final Game game)
    {
        final ViewBuilder builder = this;

        // Get the GameVersion for hard-coded version ASSUME_VERSION
        GameVersion.getByVersion(game, ASSUME_VERSION,
                                    new AsyncCallback<GameVersion>() {

            public void onFailure(Throwable error) {
                builder.showError(error.getMessage());
            }

            public void onSuccess(final GameVersion resultVersion) {
                // Get the devteam
                DevTeam.getByKey(game.getDevTeamKey(),
                    new AsyncCallback<DevTeam>() {
                    @Override
                    public void onFailure(Throwable error) {
                        builder.showError(error.getMessage());
                    }
                    @Override
                    public void onSuccess(DevTeam dt)
                    {
                        loadGame(panel, game, resultVersion, dt);
                    }
                });
            }

        });

        return panel;
    }

    /**
     * Constructs the actual panel
     * @param result
     */
    private void loadGame(FlowPanel panel, Game game, GameVersion version,
        DevTeam devteam) {
        Panel nameEditPanel = new HTMLPanel("");
        panel.add(nameEditPanel);
        nameEditPanel.add(new InlineLabel(version.getFullName()));
        // Add "edit" link, if authorized
        if (game.userCanEdit())
        {
            nameEditPanel.add(new InlineLabel(" "));
            nameEditPanel.add(new InlineHyperlink("(edit)",
                "!/" + devteam.getUrlName() + "/" + game.getUrlName()
                + "/+edit"));
        }

        Panel byTeam = new HTMLPanel("by ");
        panel.add(byTeam);
        byTeam.add(new Anchor(devteam.getFullName(),
            "#!/" + devteam.getUrlName()));
        panel.add(new Label(version.getDescription()));
        panel.add(new Label("Version: " + version.getDisplayVersion()));
        // Opens in a new tab
        panel.add(new Anchor("Play game", "/+games/" + devteam.getUrlName()
            + "/" + game.getUrlName() + "/", "_blank"));
        // Set the window title
        this.setWindowTitle("MUGLE - " + version.getFullName());
        this.updateWindowTitle();
    }

}
