/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.client.ui.admin;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import au.edu.unimelb.csse.mugle.client.Mugle;
import au.edu.unimelb.csse.mugle.client.ui.ViewBuilder;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.multiplayer.model.GameInstance;

import com.google.gwt.cell.client.*;
import com.google.gwt.core.client.GWT;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.i18n.client.DateTimeFormat;
import com.google.gwt.i18n.client.DateTimeFormat.PredefinedFormat;
import com.google.gwt.safehtml.shared.SafeHtmlBuilder;
import com.google.gwt.user.cellview.client.*;
import com.google.gwt.user.cellview.client.ColumnSortEvent.ListHandler;
import com.google.gwt.user.cellview.client.SimplePager.TextLocation;
import com.google.gwt.user.client.History;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.user.client.ui.*;
import com.google.gwt.view.client.*;

/**
 * Loads the view for game monitor
 */
public class AdminGameMonitorViewBuilder extends ViewBuilder {

    private String devteamName;
    private String gameName;
    private Game game;
    private Panel mainPanel;
    private boolean isBuilt = false;
    private CellTable<GameInstance> table;
    private DataGrid<User> playerList;
    private AsyncSortHandler<GameInstance> sortHandler;
    private HashMap<GameInstance, List<User>> playerCache;

    private final int PAGE_SIZE = Math.max(Window.getClientHeight() / 50, 15);
    private final int PRELOAD_SIZE = 5 * PAGE_SIZE;

    public AdminGameMonitorViewBuilder() {
    }

    public AdminGameMonitorViewBuilder(String devteamName, String gameName) {
        super(gameName);
        this.devteamName = devteamName;
        this.gameName = gameName;
        this.playerCache = new HashMap<GameInstance, List<User>>();
    }

    @Override
    public Widget buildView() {
        if (Mugle.getAuthUser().getUser().getRole() != Role.ADMIN) {
            this.getDataPanel().add(this.getAuthorizationError(Role.ADMIN));
            return this.getDataPanel();
        }
        this.clearPanel();
        if (this.isBuilt) {
            this.getDataPanel().add(mainPanel);
            return this.getDataPanel();
        } else {
            this.isBuilt = true;
            if (this.getUrlFragment().equals(""))
                return loadGames(this.getDataPanel());
            else
                return loadGameProfile(this.getDataPanel());
        }
    }

    private FlowPanel loadGames(final FlowPanel panel) {

        final VerticalPanel gamePanel = new VerticalPanel();
        panel.add(gamePanel);
        Game.getAllGames(0, 10, new AsyncCallback<ModelCollection<Game>>() {
            @Override
            public void onFailure(Throwable caught) {
                Mugle.showError(caught.getMessage());
            }

            @Override
            public void onSuccess(ModelCollection<Game> result) {
                if (result.getCount() == 0) {
                    gamePanel.add(new Label("No games have been published"));
                }
                for (final Game g : result.getCollection()) {
                    final String fullname = g.getActiveVersion() == null ? g
                            .getUrlName() : g.getActiveVersion().getFullName();
                    DevTeam.getByKey(g.getDevTeamKey(),
                            new AsyncCallback<DevTeam>() {
                                @Override
                                public void onSuccess(DevTeam result) {
                                    gamePanel.add(new Anchor(fullname,
                                            "#!/+admin/monitor/"
                                                    + result.getUrlName() + "/"
                                                    + g.getUrlName()));
                                }

                                @Override
                                public void onFailure(Throwable caught) {
                                    Mugle.showError(caught.getMessage());
                                }
                            });

                }
            }
        });

        // Set the window title
        this.setWindowTitle("MUGLE - Game Monitor");
        this.updateWindowTitle();

        return panel;
    }

    private FlowPanel loadGameProfile(final FlowPanel panel) {

        Game.getByName(devteamName, gameName, new AsyncCallback<Game>() {

            public void onSuccess(Game result) {
                game = result;
                GameInstance.getCount(game, new AsyncCallback<Integer>() {

                    @Override
                    public void onSuccess(Integer count) {
                        initTable(panel, count);
                        setWindowTitle("MUGLE - " + getUrlFragment()
                                + " Monitor");
                        updateWindowTitle();
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        Mugle.showError(caught.getMessage());
                    }
                });
            }

            public void onFailure(Throwable caught) {
                Mugle.showError(caught.getMessage());
            }

        });

        return panel;

    }

    private void loadPlayers(GameInstance instance,
            final LoadHandler<User> handler) {

        GameInstance.getPlayers(instance,
                new AsyncCallback<ModelCollection<User>>() {

                    @Override
                    public void onSuccess(ModelCollection<User> result) {
                        if (result != null)
                            handler.onLoad(new ArrayList<User>(result
                                    .getCollection()));
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        Mugle.showError(caught.getMessage());
                    }
                });
    }

    private static class KeyProvider implements ProvidesKey<GameInstance> {
        @Override
        public Object getKey(GameInstance item) {
            return item.getToken();
        }
    }

    private void initTable(FlowPanel panel, int rowCount) {

        this.table = new CellTable<GameInstance>(new KeyProvider());
        table.setWidth("100%", false);
        table.getElement().getStyle().setProperty("textAlign", "left");
        table.setVisibleRange(0, Math.min(PAGE_SIZE, rowCount));
        table.setRowCount(rowCount);
        AsyncDataLoader<GameInstance> dataProvider = new AsyncDataLoader<GameInstance>(
                new KeyProvider(), new Loader<GameInstance>() {

                    @Override
                    public void load(int start, int count,
                            final LoadHandler<GameInstance> handler) {
                        GameInstance.getGameInstances(game, start, count,
                                new AsyncCallback<ModelList<GameInstance>>() {
                                    @Override
                                    public void onFailure(Throwable caught) {
                                        Mugle.showError(caught.getMessage());
                                    }

                                    @Override
                                    public void onSuccess(
                                            ModelList<GameInstance> result) {
                                        handler.onLoad(result.getList());
                                    }
                                });
                    }
                }, PRELOAD_SIZE);
        dataProvider.addDataDisplay(table);

        // Attach a column sort handler to the ListDataProvider to sort the
        // list.
        this.sortHandler = new AsyncSortHandler<GameInstance>(
                dataProvider.getDataList(), table);
        table.addColumnSortHandler(sortHandler);
        table.addRangeChangeHandler(new RangeChangeEvent.Handler() {

            @Override
            public void onRangeChange(RangeChangeEvent event) {
                if (playerList != null)
                    playerList.setRowCount(0);
            }
        });

        // Add a selection model so we can select cells.
        final SelectionModel<GameInstance> selectionModel = new MultiSelectionModel<GameInstance>();
        table.setSelectionModel(selectionModel, DefaultSelectionEventManager
                .<GameInstance> createCheckboxManager());

        initColumns(selectionModel, sortHandler);

        // Create a Pager to control the table.
        SimplePager.Resources pagerResources = GWT
                .create(SimplePager.Resources.class);
        SimplePager pager = new SimplePager(TextLocation.CENTER,
                pagerResources, false, 0, false);
        pager.setDisplay(table);
        pager.setPageSize(Math.min(PAGE_SIZE, rowCount));

        this.playerList = new DataGrid<User>();
        // playerList.setStyleName("playerList");
        playerList.getElement().getStyle().setProperty("marginLeft", "10px");
        playerList.setRowCount(0, true);
        playerList.setPixelSize(160, 35 * PAGE_SIZE);
        playerList.setEmptyTableWidget(new Label(
                "Click on player count to show players"));
        playerList.addColumn(new Column<User, String>(new TextCell()) {
            @Override
            public String getValue(User player) {
                return player.getUrlName();
            }
        }, "Players");
        final SingleSelectionModel<User> playerSelection = new SingleSelectionModel<User>();
        playerSelection
                .addSelectionChangeHandler(new SelectionChangeEvent.Handler() {

                    @Override
                    public void onSelectionChange(SelectionChangeEvent event) {
                        String urlName = playerSelection.getSelectedObject()
                                .getUrlName();
                        History.newItem("!/~" + urlName);
                    }
                });
        playerList.setSelectionModel(playerSelection);

        VerticalPanel vPanel = new VerticalPanel();
        vPanel.setHorizontalAlignment(HasHorizontalAlignment.ALIGN_CENTER);
        vPanel.add(table);
        vPanel.add(pager);
        mainPanel = new HorizontalPanel();
        mainPanel.add(vPanel);
        mainPanel.add(playerList);
        panel.add(mainPanel);
    }

    private void initColumns(final SelectionModel<GameInstance> selectionModel,
            ListHandler<GameInstance> sortHandler) {

        // Check box.
        String title = this.getUrlFragment().length() > 10 ? this
                .getUrlFragment().substring(0, 9) + "..." : this
                .getUrlFragment();
        Column<GameInstance, Boolean> checkColumn = addColumn(new CheckboxCell(
                true, false), title, new GetValue<Boolean>() {

            @Override
            public Boolean getValue(GameInstance instance) {
                return selectionModel.isSelected(instance);
            }
        }, false, null);
        table.setColumnWidth(checkColumn, 8, Unit.EM);
        checkColumn.setHorizontalAlignment(HasHorizontalAlignment.ALIGN_CENTER);

        // Name.
        Column<GameInstance, String> nameColumn = addColumn(new TextCell(),
                "Name", new GetValue<String>() {

                    @Override
                    public String getValue(GameInstance instance) {
                        return instance.getName();
                    }
                }, true, null);
        table.setColumnWidth(nameColumn, 12, Unit.EM);

        // Started Time.
        DateCell dateCell = new DateCell(
                DateTimeFormat.getFormat(PredefinedFormat.DATE_TIME_MEDIUM));

        Column<GameInstance, Date> startColumn = addColumn(dateCell,
                "Started Time", new GetValue<Date>() {

                    @Override
                    public Date getValue(GameInstance instance) {
                        return instance.getCreatedTime();
                    }
                }, true, null);
        table.setColumnWidth(startColumn, 14, Unit.EM);

        // Ended Time.
        Column<GameInstance, Date> endColumn = addColumn(dateCell,
                "Ended Time", new GetValue<Date>() {

                    @Override
                    public Date getValue(GameInstance instance) {
                        return instance.getEndedTime();
                    }
                }, true, null);
        table.setColumnWidth(endColumn, 14, Unit.EM);

        // Log.
        final ViewBuilder thisPanel = this;
        Column<GameInstance, GameInstance> logColumn = addColumn(
                new ActionCell<GameInstance>("view",
                        new ActionCell.Delegate<GameInstance>() {

                            @Override
                            public void execute(GameInstance instance) {
                                // String url = History.getToken() + "/"
                                // + instance.getName() + "/"
                                // + instance.getToken();
                                // History.newItem(url);
                                getDataPanel().clear();
                                ViewBuilder logView = new AdminGameLogViewBuilder(
                                        instance, thisPanel);
                                getDataPanel().add(logView.buildView());
                                logView.updateWindowTitle();
                            }
                        }) {
                    @Override
                    public void render(Cell.Context context, GameInstance gi,
                            SafeHtmlBuilder sb) {
                        if (gi.hasLog())
                            super.render(context, gi, sb);
                    }
                }, "Log", new GetValue<GameInstance>() {

                    @Override
                    public GameInstance getValue(GameInstance instance) {
                        return instance;
                    }
                }, false, null);
        logColumn.setSortable(true);
        sortHandler.setComparator(logColumn, new Comparator<GameInstance>() {

            @Override
            public int compare(GameInstance o1, GameInstance o2) {
                return ((Boolean) o2.hasLog()).compareTo(o1.hasLog());
            }
        });
        table.setColumnWidth(logColumn, 5, Unit.EM);

        // Player Count.
        FieldUpdater<GameInstance, String> updater = new FieldUpdater<GameInstance, String>() {
            @Override
            public void update(int index, final GameInstance instance,
                    String value) {
                if (playerCache.containsKey(instance)) {
                    List<User> players = playerCache.get(instance);
                    showPlayers(players);
                    return;
                }
                loadPlayers(instance, new LoadHandler<User>() {
                    @Override
                    public void onLoad(List<User> players) {
                        playerCache.put(instance, players);
                        showPlayers(players);
                    }
                });
            }

            private void showPlayers(List<User> players) {
                playerList.setRowData(players);
            }
        };
        Column<GameInstance, String> playerCountColumn = addColumn(
                new ClickableTextCell(), "Player Count",
                new GetValue<String>() {

                    @Override
                    public String getValue(GameInstance instance) {
                        return String.valueOf(instance.getPlayerCount());
                    }
                }, true, updater);
        playerCountColumn
                .setHorizontalAlignment(HasHorizontalAlignment.ALIGN_CENTER);
        table.setColumnWidth(playerCountColumn, 9, Unit.EM);
    }

    /**
     * Add a column with a header.
     * 
     * @param <C>
     *            the cell type
     * @param cell
     *            the cell used to render the column
     * @param header
     *            the header string
     * @param getter
     *            the value getter for the cell
     * @param sortable
     *            true if the column is sortable
     */
    public <C> Column<GameInstance, C> addColumn(Cell<C> cell, String header,
            final GetValue<C> getter, boolean sortable,
            FieldUpdater<GameInstance, C> updater) {

        Column<GameInstance, C> column = new Column<GameInstance, C>(cell) {
            @Override
            public C getValue(GameInstance object) {
                return getter.getValue(object);
            }
        };
        if (sortable) {
            column.setSortable(true);
            sortHandler.setComparator(column, new Comparator<GameInstance>() {
                @SuppressWarnings("unchecked")
                public int compare(GameInstance o1, GameInstance o2) {
                    if (getter.getValue(o1) == null)
                        return -1;
                    else if (getter.getValue(o2) == null)
                        return 1;
                    else
                        return ((Comparable<C>) getter.getValue(o1))
                                .compareTo(getter.getValue(o2));
                }
            });
        }
        column.setFieldUpdater(updater);
        table.addColumn(column, header);
        return column;
    }

    /**
     * Get a cell value from a record.
     * 
     * @param <C>
     *            the cell type
     */
    private interface GetValue<C> {
        public C getValue(GameInstance instance);
    }

    public interface LoadHandler<T> {
        public void onLoad(List<T> list);
    }

    public interface Loader<T> {
        public void load(int start, int count, LoadHandler<T> handler);
    }

    public static class AsyncSortHandler<T> extends ListHandler<T> {

        private HasData<T> display;

        public AsyncSortHandler(List<T> list, HasData<T> display) {
            super(list);
            this.display = display;
        }

        @Override
        public void onColumnSort(ColumnSortEvent event) {
            super.onColumnSort(event);
            display.setVisibleRangeAndClearData(display.getVisibleRange(), true);
        }

    }

    /**
     * Asynchronous data loader.
     * 
     * @param <T>
     *            the loader data type
     */
    public static class AsyncDataLoader<T> extends AsyncDataProvider<T> {

        private Timer loadTimer;
        private Loader<T> loader;
        private int loadSize;
        private ArrayList<T> dataList = new ArrayList<T>();

        public AsyncDataLoader(Loader<T> loader, int loadSize) {
            this.loader = loader;
        }

        public AsyncDataLoader(ProvidesKey<T> keyProvider, Loader<T> loader,
                int loadSize) {
            super(keyProvider);
            this.loader = loader;
            this.loadSize = loadSize;
        }

        public ArrayList<T> getDataList() {
            return this.dataList;
        }

        @Override
        protected void onRangeChanged(HasData<T> display) {
            if (loadTimer != null) {
                loadTimer.cancel();
                loadTimer = null;
            }
            Range range = display.getVisibleRange();
            final int dataSize = dataList.size();
            final int start = range.getStart();
            final int end = Math.min(start + range.getLength(),
                    display.getRowCount());
            final int loadStart = start == 0 ? start : end;
            if (end >= dataSize && dataSize != display.getRowCount())
                loader.load(loadStart, loadSize, new LoadHandler<T>() {
                    @Override
                    public void onLoad(final List<T> instances) {
                        dataList.addAll(instances);
                    }
                });
            if (dataSize < end) {
                new Timer() {
                    @Override
                    public void run() {
                        loadTimer = this;
                        if (dataList.size() >= end) {
                            updateRowData(start, dataList.subList(start, end));
                            loadTimer = null;

                            this.cancel();
                        }
                    }
                }.scheduleRepeating(100);
            } else
                updateRowData(start, dataList.subList(start, end));
        }

    }

}
