/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.client.ui.admin;

import java.util.Comparator;
import java.util.Date;

import au.edu.unimelb.csse.mugle.client.Mugle;
import au.edu.unimelb.csse.mugle.client.ui.ViewBuilder;
import au.edu.unimelb.csse.mugle.client.ui.admin.AdminGameMonitorViewBuilder.*;
import au.edu.unimelb.csse.mugle.shared.model.ModelList;
import au.edu.unimelb.csse.mugle.shared.multiplayer.model.*;

import com.google.gwt.cell.client.*;
import com.google.gwt.core.client.GWT;
import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.i18n.client.DateTimeFormat;
import com.google.gwt.i18n.client.DateTimeFormat.PredefinedFormat;
import com.google.gwt.safehtml.shared.SafeHtmlBuilder;
import com.google.gwt.safehtml.shared.SafeHtmlUtils;
import com.google.gwt.user.cellview.client.*;
import com.google.gwt.user.cellview.client.SimplePager.TextLocation;
import com.google.gwt.user.client.Window;
import com.google.gwt.user.client.rpc.AsyncCallback;
import com.google.gwt.user.client.ui.*;

/**
 * Loads the view for log message
 */
public class AdminLogMessageViewBuilder extends ViewBuilder {

    private CellTable<LogMessage> table;
    private Panel mainPanel;
    private ViewBuilder parent;
    private String gameName;

    private final int PAGE_SIZE = Math.max(Window.getClientHeight() / 50, 15);
    private final int PRELOAD_SIZE = 5 * PAGE_SIZE;

    public AdminLogMessageViewBuilder(String gameName, String logToken,
            ViewBuilder parent) {
        super(logToken);
        this.gameName = gameName;
        this.parent = parent;
        this.mainPanel = new HorizontalPanel();
        this.mainPanel.setPixelSize(-1, 35 * PAGE_SIZE);
    }

    @Override
    public Widget buildView() {
        return loadLogProfile(this.getDataPanel());
    }

    private FlowPanel loadLogProfile(final FlowPanel panel) {
        GameLog.getLogMessageCount(this.getUrlFragment(),
                new AsyncCallback<Integer>() {

                    @Override
                    public void onSuccess(Integer result) {
                        initTable(panel, result);
                    }

                    @Override
                    public void onFailure(Throwable caught) {
                        Mugle.showError(caught.getMessage());
                    }
                });

        setWindowTitle("MUGLE - " + gameName + " Log Messages");
        updateWindowTitle();

        return panel;
    }

    private void initTable(FlowPanel panel, int rowCount) {

        this.table = new CellTable<LogMessage>();
        table.setWidth("100%", false);
        table.getElement().getStyle().setProperty("textAlign", "left");
        table.setVisibleRange(0, Math.min(PAGE_SIZE, rowCount));
        table.setRowCount(rowCount);

        AsyncDataLoader<LogMessage> dataProvider = new AsyncDataLoader<LogMessage>(
                new AdminGameMonitorViewBuilder.Loader<LogMessage>() {

                    @Override
                    public void load(int start, int count,
                            final LoadHandler<LogMessage> handler) {

                        GameLog.getLogMessages(getUrlFragment(), start, count,
                                new AsyncCallback<ModelList<LogMessage>>() {

                                    @Override
                                    public void onSuccess(
                                            ModelList<LogMessage> result) {
                                        handler.onLoad(result.getList());
                                    }

                                    @Override
                                    public void onFailure(Throwable caught) {
                                        Mugle.showError(caught.getMessage());
                                    }
                                });
                    }
                }, PRELOAD_SIZE);
        dataProvider.addDataDisplay(table);

        AsyncSortHandler<LogMessage> sortHandler = new AsyncSortHandler<LogMessage>(
                dataProvider.getDataList(), table);
        table.addColumnSortHandler(sortHandler);

        initColumns(sortHandler);

        // Create a Pager to control the table.
        SimplePager.Resources pagerResources = GWT
                .create(SimplePager.Resources.class);
        SimplePager pager = new SimplePager(TextLocation.CENTER,
                pagerResources, false, 0, false);
        pager.setDisplay(table);
        pager.setPageSize(Math.min(PAGE_SIZE, rowCount));

        VerticalPanel vPanel = new VerticalPanel();
        vPanel.add(table);
        HorizontalPanel hPanel = new HorizontalPanel();
        Button button = new Button("Back to Log");
        button.getElement().getStyle().setProperty("marginRight", "100px");
        button.addClickHandler(new ClickHandler() {

            @Override
            public void onClick(ClickEvent event) {
                parent.buildView();
            }
        });
        hPanel.add(button);
        hPanel.add(pager);
        vPanel.add(hPanel);
        mainPanel.add(vPanel);
        panel.add(mainPanel);
    }

    private void initColumns(AsyncSortHandler<LogMessage> sortHandler) {

        // Game Name.
        Column<LogMessage, Number> nameColumn = new Column<LogMessage, Number>(
                new NumberCell()) {

            @Override
            public Number getValue(LogMessage lm) {
                return null;
            }

            @Override
            public void render(Cell.Context context, LogMessage lm,
                    SafeHtmlBuilder sb) {
                sb.append(SafeHtmlUtils.fromTrustedString(String
                        .valueOf(context.getIndex() + 1)));
            }
        };
        String title = gameName.length() > 20 ? gameName.substring(0, 19)
                + "..." : gameName;
        nameColumn.setHorizontalAlignment(HasHorizontalAlignment.ALIGN_CENTER);
        table.addColumn(nameColumn, title);

        // Body.
        Column<LogMessage, String> bodyColumn = new Column<LogMessage, String>(
                new TextCell()) {

            @Override
            public String getValue(LogMessage lm) {
                return lm.getBody();
            }
        };
        table.addColumn(bodyColumn, "Body");

        // Sender.
        Column<LogMessage, String> senderColumn = new Column<LogMessage, String>(
                new TextCell()) {

            @Override
            public String getValue(LogMessage lm) {
                return lm.getSender();
            }
        };
        senderColumn.setSortable(true);
        sortHandler.setComparator(senderColumn, new Comparator<LogMessage>() {
            public int compare(LogMessage o1, LogMessage o2) {
                return o1.getSender().compareTo(o2.getSender());
            }
        });
        table.addColumn(senderColumn, "Sender");

        // Time.
        DateCell dateCell = new DateCell(
                DateTimeFormat.getFormat(PredefinedFormat.DATE_TIME_MEDIUM));
        Column<LogMessage, Date> timeColumn = new Column<LogMessage, Date>(
                dateCell) {

            @Override
            public Date getValue(LogMessage lm) {
                return lm.getCreatedTime();
            }
        };
        timeColumn.setSortable(true);
        sortHandler.setComparator(timeColumn, new Comparator<LogMessage>() {
            public int compare(LogMessage o1, LogMessage o2) {
                return o1.getCreatedTime().compareTo(o2.getCreatedTime());
            }
        });
        table.addColumn(timeColumn, "Timestamp");

    }

}
