/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server;

import java.io.InputStream;
import java.io.PrintWriter;
import java.io.IOException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.jdo.PersistenceManager;

import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;

import org.apache.commons.fileupload.FileItemIterator;
import org.apache.commons.fileupload.FileItemStream;
import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.apache.commons.fileupload.FileUploadException;

import com.google.appengine.api.datastore.Key;

import java.util.Vector;

import au.edu.unimelb.csse.mugle.server.model.UserData;
import au.edu.unimelb.csse.mugle.server.model.UserGetter;
import au.edu.unimelb.csse.mugle.server.model.GameData;
import au.edu.unimelb.csse.mugle.server.model.GameGetter;
import au.edu.unimelb.csse.mugle.server.model.GameFileData;
import au.edu.unimelb.csse.mugle.server.model.GameFileContents;
import au.edu.unimelb.csse.mugle.server.model.GameFileGetter;
import au.edu.unimelb.csse.mugle.server.model.GameVersionData;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.shared.model.Role;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.GameNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.GameFileNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.GameFileContentsNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.UserPrivilegeException;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.ServerException;

public class UploadService extends HttpServlet
{
    public void doPost(HttpServletRequest request,
        HttpServletResponse response) throws IOException
    {
        long gameversionID = Long.parseLong(
                                request.getParameter("gameversion"));
        GameVersionData gameversion;
        boolean got_zip = false;
        int numfiles = 0;
        Vector<String> filenames = new Vector<String>();

        // Read the POST data fields
        ServletFileUpload upload = new ServletFileUpload();
        PersistenceManager pm = PMF.getManager();
        try
        {
            // Get the GameVersion object
            gameversion =
                pm.getObjectById(GameVersionData.class, gameversionID);
            // Ensure the current user has permission to write to this game
            checkPermissions(pm, gameversion);
        }
        catch (UserPrivilegeException e)
        {
            response.sendError(403, e.toString());
            return;
        }
        catch (ServerException e)
        {
            response.sendError(500, e.toString());
            return;
        }
        finally
        {
            pm.close();
        }
        try
        {
            // For each field
            FileItemIterator iter = upload.getItemIterator(request);
            while (iter.hasNext())
            {
                FileItemStream item = iter.next();
                // Only care about the game_upload field
                if (!item.getFieldName().equals("game_upload"))
                    continue;
                got_zip = true;

                // Assume it is a ZIP file. Read each file contained inside.
                InputStream stream = item.openStream();
                ZipInputStream zstream = new ZipInputStream(stream);
                ZipEntry entry;
                while ((entry = zstream.getNextEntry()) != null)
                {
                    if (entry.isDirectory())
                    {
                        // Ignore directories (since we are not building a
                        // tree; just remembering the path to each file)
                        continue;
                    }
                    String filename = entry.getName();
                    // Remove the first path segment from the filename (since
                    // ZIP files tend to have a top-level directory, and we
                    // want to explode out of that).
                    String[] filename_split = filename.split("/", 2);
                    if (filename_split.length > 1)
                        filename = filename_split[1];
                    if (filename.startsWith("WEB-INF/"))
                    {
                        // Ignore any file in the WEB-INF directory (for GWT
                        // projects, this represents the server-side and
                        // library code, which is HUGE, and unnecessary in a
                        // MUGLE game, which is client-only).
                        continue;
                    }
                    long size = entry.getSize();
                    // Create a new GameFile object
                    createGameFile(gameversion, filename, zstream, size);
                    filenames.add(filename);
                    numfiles++;
                }
            }
        }
        catch (FileUploadException e)
        {
            response.sendError(400, e.toString());
            return;
        }
        catch (ZipException e)
        {
            response.sendError(400, e.toString());
            return;
        }
        if (!got_zip)
        {
            response.sendError(400, "Missing POST field game_upload");
            return;
        }

        // Set HTML content type, even though returning text
        // The result is shown in an alert, and if it's text, Firefox will add
        // "<pre>".
        response.setContentType("text/html");
        PrintWriter out = response.getWriter();
        out.println("Successfully uploaded a ZIP file with "
            + numfiles + " files:");
        for (String s : filenames)
            out.println("- " + s);
    }

    /** Create a new GameFile and store it in the database, from a part of a
     * Zip file.
     * @param gameVersionKey The game version to upload this file to.
     * @param filePath Path to the file relative to this game version.
     * @param data Binary stream to read file contents from.
     * @param dataSize Number of bytes to read from data.
     */
    public static void createGameFile(GameVersionData gameVersion,
        String filePath, InputStream data, long dataSize)
        throws IOException
    {
        PersistenceManager pm = PMF.getManager();
        try
        {
            // Check if a file already exists at that path
            GameFileData file;
            GameFileContents contents;
            try
            {
                file = GameFileGetter.getGameFile(pm, gameVersion, filePath);
            }
            catch (GameFileNotExists e)
            {
                // First, store the contents as a blob in the database
                contents = new GameFileContents(data, dataSize);
                pm.makePersistent(contents);
                // Then, create a GameFile which points at that blob
                file = new GameFileData(gameVersion, filePath, contents);
                pm.makePersistent(file);
                return;
            }
            // Already exists; just update the file and contents
            // (Note that if we were uploading a new version, we wouldn't
            // overwrite the file or contents, but since we're updating an
            // existing version, we can replace the contents).
            // XXX Note that if contents are shared across versions this will
            // clobber ALL versions, so probably we should create a new
            // content object each time.
            try
            {
                contents = GameFileGetter.getGameFileContents(file);
                contents.setContents(data, dataSize);
                pm.makePersistent(contents);
            }
            catch (GameFileContentsNotExists e)
            {
                contents = new GameFileContents(data, dataSize);
                pm.makePersistent(contents);
                file.setContents(contents.getServerKey());
            }
            // Also set the MIME type based on the filename (so the result is
            // consistent with what would have happened if the file didn't
            // exist).
            file.setMimeTypeFromFilename();
        }
        finally
        {
            pm.close();
        }
    }

    /** Check if the currently logged in user is allowed to write to this
     * game.
     * @return The current user's role.
     */
    private static Role checkPermissions(PersistenceManager pm,
        GameVersionData gv) throws UserPrivilegeException, ServerException
    {
        // Check if the currently logged in user is part of the game's devTeam
        GameData g;
        try {
            g = GameGetter.getGame(gv.getGame().getId());
        } catch (GameNotExists e) {
            /*  IF this happens it means the server has screwed
             *  up entity relations somewhere
             */
            throw new ServerException("Server Entity Relationship " +
                    "Management Error: " + e.getMessage());
        }
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.equals(g.getDevTeam())) {
                isDev = true;
            }
        }
        // Only an admin or person in the DevTeam can modify the game
        if (!curUser.isActive() || (curUser.getRole()!=Role.ADMIN && !isDev))
        {
            String msg = "The current user does not have the the right "
                         + "privilege to modify the specified Game object.";
            throw new UserPrivilegeException(msg);
        }

        return curUser.getRole();
    }
}
