/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.api;

import java.util.ArrayList;
import java.util.List;

import au.edu.unimelb.csse.mugle.client.api.BadgeService;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.AchievementData;
import au.edu.unimelb.csse.mugle.server.model.AchievementGetter;
import au.edu.unimelb.csse.mugle.server.model.GameData;
import au.edu.unimelb.csse.mugle.server.model.GameGetter;
import au.edu.unimelb.csse.mugle.server.model.UserAchievementData;
import au.edu.unimelb.csse.mugle.server.model.UserAchievementGetter;
import au.edu.unimelb.csse.mugle.shared.api.Badge;
import au.edu.unimelb.csse.mugle.shared.api.BadgeError;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;

import au.edu.unimelb.csse.mugle.shared.multiplayer.api.InstanceTokenError;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;

/**
 * The server side implementation of KeyValueService.
 */
@SuppressWarnings("serial")
public class BadgeServiceImpl extends RemoteServiceServlet implements
        BadgeService {
    private UserAchievementData getUserAchievement(String gameToken, String name)
            throws GameTokenError, BadgeError {
        PersistenceManager pm = PMF.getManager();
        try {
            return this.getUserAchievement(pm, gameToken, name);
        } finally {
            pm.close();
        }
    }

    private UserAchievementData getUserAchievement(PersistenceManager pm,
            String gameToken, String name) throws GameTokenError, BadgeError {
        try {
            return UserAchievementGetter
                    .getUserAchievement(pm, name, gameToken);
        } catch (AchievementNotExists e) {
            throw new BadgeError(e.getMessage());
        }
    }

    @SuppressWarnings("unchecked")
    @Override
    public String[] getBadgeNames(String gameToken) throws GameTokenError {

        ArrayList<String> toReturn = new ArrayList<String>();

        GameData curGame = GameGetter.getCurrentGame(gameToken);

        PersistenceManager pm = PMF.getManager();
        try {
            Query q = pm.newQuery(AchievementData.class, "game == g");
            q.declareParameters("com.google.appengine.api.datastore.Key g");
            List<AchievementData> achievements = (List<AchievementData>) q
                    .execute(curGame.getServerKey());
            for (AchievementData a : achievements) {
                toReturn.add(a.getName());
            }

        } finally {
            pm.close();
        }
        return (String[]) toReturn.toArray(new String[toReturn.size()]);
    }

    @Override
    public Badge getBadgeInfo(String gameToken, String name)
            throws GameTokenError, BadgeError {

        PersistenceManager pm = PMF.getManager();
        AchievementData a;
        try {
            a = AchievementGetter.getAchievement(pm, name, gameToken);
            ;
        } catch (AchievementNotExists e) {
            throw new BadgeError(e.getMessage());
        } finally {
            pm.close();
        }
        // XXX The badge API requires both a short name and a display name
        // Currently use the same name for both; requires changes to
        // Achievement class.
        return new Badge(a.getName(), a.getDisplayName(), a.getDescription(),
                a.getMaxProgress());
    }

    @Override
    public boolean isAchieved(String gameToken, String name)
            throws GameTokenError, BadgeError {
        return this.getUserAchievement(gameToken, name).getAchieved();
    }

    @Override
    public int getProgress(String gameToken, String name)
            throws GameTokenError, BadgeError {
        return this.getUserAchievement(gameToken, name).getProgress();
    }

    @Override
    public boolean incrementProgress(String gameToken, String name, int amount)
            throws GameTokenError, BadgeError {

        PersistenceManager pm = PMF.getManager();

        int maxProgress = 0;

        if (amount < 0)
            throw new BadgeError(
                    "incrementProgress called with negative amount");

        try {
            UserAchievementData ua = this.getUserAchievement(pm, gameToken,
                    name);

            try {
                AchievementData achievement = AchievementGetter
                        .getAchievement(ua.getAchievement().getId());
                maxProgress = achievement.getMaxProgress();
            } catch (AchievementNotExists e) {
                /*
                 * Fatal server exception - Achievement Key stored in the User
                 * Achievement doesnt exist Try to correct by looking up the
                 * Achievement independently
                 */
                try {
                    AchievementData achievement = AchievementGetter
                            .getAchievement(name, gameToken);
                    ua.setAchievement(achievement); // correct for key being
                                                    // wrong
                    maxProgress = achievement.getMaxProgress();
                } catch (AchievementNotExists f) {
                    throw new BadgeError(f.getMessage());
                }
            }

            if (maxProgress == 0) {
                throw new BadgeError(
                        "incrementProgress called on non-progress badge "
                                + name);
            }
            int prog = ua.getProgress() + amount;

            if (prog >= maxProgress) {
                ua.setProgress(maxProgress);
                ua.setAchieved(true);
                return true;
            } else {
                ua.setProgress(prog);
                ua.setAchieved(false); // If for some reason, they change the
                                       // max progress to be less.
                return false;
            }
        } finally {
            pm.close();
        }
    }

    @Override
    public boolean incrementProgress(String gameToken, String name)
            throws GameTokenError, BadgeError {
        return incrementProgress(gameToken, name, 1);
    }

    @Override
    public void setAchieved(String gameToken, String name)
            throws GameTokenError, BadgeError {
        PersistenceManager pm = PMF.getManager();

        try {
            UserAchievementData ua = this.getUserAchievement(pm, gameToken,
                    name);
            ua.setAchieved(true);

            // also need to set the progress to the max progress
            try {
                AchievementData achievement = AchievementGetter
                        .getAchievement(ua.getAchievement().getId());
                ua.setProgress(achievement.getMaxProgress());
            } catch (AchievementNotExists e) {
                /*
                 * Fatal server exception - Achievement Key stored in the User
                 * Achievement doesnt exist Try to correct by looking up the
                 * Achievement independently
                 */
                try {
                    AchievementData achievement = AchievementGetter
                            .getAchievement(name, gameToken);
                    ua.setAchievement(achievement); // correct for key being
                                                    // wrong
                    ua.setProgress(achievement.getMaxProgress());
                } catch (AchievementNotExists f) {
                    throw new BadgeError(f.getMessage());
                }
            }

        } finally {
            pm.close();
        }
    }

    private UserAchievementData getInstanceUserAchievement(
            PersistenceManager pm, String gameToken, String name,
            String instanceToken) throws GameTokenError, BadgeError,
            InstanceTokenError {
        try {
            return UserAchievementGetter.getInstanceUserAchievement(pm, name,
                    gameToken, instanceToken);
        } catch (AchievementNotExists e) {
            throw new BadgeError(e.getMessage());
        }
    }

    @Override
    public void setAchieved(String gameToken, String name, String instanceToken)
            throws GameTokenError, BadgeError, InstanceTokenError {
        PersistenceManager pm = PMF.getManager();

        try {
            UserAchievementData ua = null;
            if (instanceToken == null)
                ua = this.getUserAchievement(pm, gameToken, name);
            else
                ua = this.getInstanceUserAchievement(pm, gameToken, name,
                        instanceToken);
            ua.setAchieved(true);

            // also need to set the progress to the max progress
            try {
                AchievementData achievement = AchievementGetter
                        .getAchievement(ua.getAchievement().getId());
                ua.setProgress(achievement.getMaxProgress());
            } catch (AchievementNotExists e) {
                /*
                 * Fatal server exception - Achievement Key stored in the User
                 * Achievement doesnt exist Try to correct by looking up the
                 * Achievement independently
                 */
                try {
                    AchievementData achievement = AchievementGetter
                            .getAchievement(name, gameToken);
                    ua.setAchievement(achievement); // correct for key being
                                                    // wrong
                    ua.setProgress(achievement.getMaxProgress());
                } catch (AchievementNotExists f) {
                    throw new BadgeError(f.getMessage());
                }
            }

        } finally {
            pm.close();
        }
    }

    @Override
    public boolean incrementProgress(String gameToken, String name, int amount,
            String instanceToken) throws GameTokenError, BadgeError,
            InstanceTokenError {
        PersistenceManager pm = PMF.getManager();

        int maxProgress = 0;

        if (amount < 0)
            throw new BadgeError(
                    "incrementProgress called with negative amount");

        try {
            UserAchievementData ua = null;
            if (instanceToken == null)
                ua = this.getUserAchievement(pm, gameToken, name);
            else
                ua = this.getInstanceUserAchievement(pm, gameToken, name,
                        instanceToken);

            try {
                AchievementData achievement = AchievementGetter
                        .getAchievement(ua.getAchievement().getId());
                maxProgress = achievement.getMaxProgress();
            } catch (AchievementNotExists e) {
                /*
                 * Fatal server exception - Achievement Key stored in the User
                 * Achievement doesnt exist Try to correct by looking up the
                 * Achievement independently
                 */
                try {
                    AchievementData achievement = AchievementGetter
                            .getAchievement(name, gameToken);
                    ua.setAchievement(achievement); // correct for key being
                                                    // wrong
                    maxProgress = achievement.getMaxProgress();
                } catch (AchievementNotExists f) {
                    throw new BadgeError(f.getMessage());
                }
            }

            if (maxProgress == 0) {
                throw new BadgeError(
                        "incrementProgress called on non-progress badge "
                                + name);
            }
            int prog = ua.getProgress() + amount;

            if (prog >= maxProgress) {
                ua.setProgress(maxProgress);
                ua.setAchieved(true);
                return true;
            } else {
                ua.setProgress(prog);
                ua.setAchieved(false); // If for some reason, they change the
                                       // max progress to be less.
                return false;
            }
        } finally {
            pm.close();
        }
    }

    @Override
    public boolean isAchieved(String gameToken, String name,
            String instanceToken) throws GameTokenError, BadgeError,
            InstanceTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            if (instanceToken == null)
                return this.getUserAchievement(pm, gameToken, name)
                        .getAchieved();
            else
                return this.getInstanceUserAchievement(pm, gameToken, name,
                        instanceToken).getAchieved();
        } finally {
            pm.close();
        }

    }

    @Override
    public int getProgress(String gameToken, String name, String instanceToken)
            throws GameTokenError, BadgeError, InstanceTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            if (instanceToken == null)
                return this.getUserAchievement(pm, gameToken, name)
                        .getProgress();
            else
                return this.getInstanceUserAchievement(pm, gameToken, name,
                        instanceToken).getProgress();
        } finally {
            pm.close();
        }
    }

}
