/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.List;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.AchievementNotExists;

public class AchievementGetter {

    /**
     * Gets the Achievement of the current game based on its name - READ ONLY
     * @param name Name of the achievement you're looking for
     * @param gameToken the secret game token for the current game, provided by the client
     * @return The requested Achievement - READ ONLY
     * @throws GameTokenError
     * @throws AchievementNotExists
     */
    public static AchievementData getAchievement(String name, String gameToken) 
            throws GameTokenError, AchievementNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getAchievement(pm, name, gameToken);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the Achievement of the current game based on its name, for editing object in datstore
     * The PersistenceManager must be handled by the caller
     * @param pm The Persistence Manager
     * @param name Name of the achievement you're looking for
     * @param gameToken the secret game token for the current game, provided by the client
     * @return The requested Achievement
     * @throws GameTokenError
     * @throws AchievementNotExists
     */
    @SuppressWarnings("unchecked")
    public static AchievementData getAchievement(PersistenceManager pm, String name, String gameToken) 
            throws GameTokenError, AchievementNotExists {
        
        GameData curGame = GameGetter.getCurrentGame(pm, gameToken);
        
        Query q = pm.newQuery(AchievementData.class, "game == g && name == n");
        q.declareParameters("com.google.appengine.api.datastore.Key g, String n");
        List<AchievementData> results = (List<AchievementData>) q.execute(curGame.getServerKey(), name);

        if (results.isEmpty()) {
            throw new AchievementNotExists(curGame.getUrlName(), name);
        }
        
        return results.get(0);
    }

    /**
     * Gets the Achievement of the current game based on its Primary Key - READ ONLY
     * @param primaryKey the primary key of the Achievement
     * @return The requested Achievement - READ ONLY
     * @throws AchievementNotExists
     */
    public static AchievementData getAchievement(Long primaryKey) 
            throws AchievementNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getAchievement(pm, primaryKey);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the Achievement of the current game based on its Primary Key, for editing object in datstore
     * The PersistenceManager must be handled by the caller
     * @param pm The Persistence Manager
     * @param primaryKey the primary key of the Achievement
     * @return The requested Achievement
     * @throws GameTokenError
     * @throws AchievementNotExists
     */
    public static AchievementData getAchievement(PersistenceManager pm, Long primaryKey) 
            throws AchievementNotExists {
        try {
            AchievementData achievement = pm.getObjectById(AchievementData.class, primaryKey);
            return achievement;
        } catch (JDOObjectNotFoundException e) {
            throw new AchievementNotExists(primaryKey);
        }
    }

    /**
     * Gets the Achievement of the current game based on its Primary Key - READ ONLY
     * @param primaryKey the primary key of the Achievement
     * @return The requested Achievement - READ ONLY
     * @throws AchievementNotExists
     */
    public static AchievementData getAchievement(Key primaryKey)
            throws AchievementNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getAchievement(pm, primaryKey);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the Achievement associated with the given AchievementData primary key.
     * @param pm The PersistenceManager
     * @param key The server-side key (ModelDataClass.getServerKey())
     * @return The AchievementData object directly off the datastore.
     * @throws AchievementNotExists 
     */
    public static AchievementData getAchievement(PersistenceManager pm, Key key)
            throws AchievementNotExists {
        try {
            AchievementData achievement = pm.getObjectById(AchievementData.class, key);
            return achievement;
        } catch (JDOObjectNotFoundException e) {
            throw new AchievementNotExists(key.getId());
        }
    }
    
}
