/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.jdo.annotations.IdGeneratorStrategy;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;
import javax.jdo.annotations.PrimaryKey;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.server.model.annotations.*;

@MugleData
@PersistenceCapable
public class DevTeamData extends ModelDataClass<Long, DevTeam, DevTeamData> {


    // Fields

    @PrimaryKey
    @Persistent(valueStrategy = IdGeneratorStrategy.IDENTITY)
    private Key id;

    @Persistent
    private String urlName; //Must be unique

    @Persistent
    private String fullName;

    @Persistent
    private Set<Key> games;

    // Many-to-many
    @Persistent
    private Set<Key> users;

    @Persistent
    private Key createdUser;

    @Persistent
    private Date createdTime;

    @Persistent
    private Key updatedUser;

    @Persistent
    private Date updatedTime;


    // Constructors

    public DevTeamData() {
        this.games = new HashSet<Key>();
        this.users = new HashSet<Key>();
        this.urlName = null;
        this.fullName = null;
    }

    public DevTeamData(String urlName, String fullName) {
        this.games = new HashSet<Key>();
        this.users = new HashSet<Key>();
        this.urlName = urlName;
        this.fullName = fullName;
    }

    // Getters

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setPrimaryKey")
    public Long getPrimaryKey() {
        if (this.id == null) {
            return null;
        }
        return this.id.getId();
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setFullName")
    public String getFullName() {
        return this.fullName;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, list=true)
    public Set<Key> getGames() {
        return this.games;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUrlName")
    public String getUrlName() {
        return this.urlName;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, list=true)
    public Set<Key> getUsers() {
        return this.users;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedUser", overrideBy="keyToCreated")
    public Key getCreatedUser() {
        return this.createdUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedTime")
    public Date getCreatedTime() {
        return this.createdTime;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedUser", overrideBy="keyToUpdated")
    public Key getUpdatedUser() {
        return this.updatedUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedTime")
    public Date getUpdatedTime() {
        return this.updatedTime;
    }


    // Setters

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedUser(Key createdUser) {
        this.createdUser = createdUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedTime(Date createdTime) {
        this.createdTime = createdTime;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedUser(Key updatedUser) {
        this.updatedUser = updatedUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedTime(Date updatedTime) {
        this.updatedTime = updatedTime;
    }

    @SetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedBy="getFullName")
    public void setFullName(String fullName) {
        this.fullName = fullName;
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getUrlName")
    public void setUrlName(String urlName) {
        this.urlName = urlName;
    }


    // Misc methods

    public void addUser(UserData user)
    {
        // many-to-many relationship
        this.users.add(user.getServerKey());
        user.getDevTeams().add(this.id);
    }

    public void removeUser(UserData user)
    {
        // many-to-many relationship
        this.users.remove(user.getServerKey());
        user.getDevTeams().remove(this.id);
    }

    public void addGame(GameData game) {
        game.setDevTeam(this);
    }

    public void removeGame(GameData game) {
        this.games.remove(game.getServerKey());
    }

    @Override
    public Key getServerKey() {
        return this.id;
    }

    @Override
    public void setServerKey(Key key) {
        this.id = key;
    }

    @Override
    public void ClearFields() {
        this.setFullName(null);
        this.setServerKey(null);
        this.setUrlName(null);
    }

    @Override
    public Class<DevTeam> getClientClassType() {
        return DevTeam.class;
    }

    public Long getLongKey(Key key) {
        if (key == null) { return null; }
        return key.getId();
    }

    public Long keyToCreated() {
        if (this.createdUser == null) { return null; }
        return this.createdUser.getId();
    }

    public Long keyToUpdated() {
        if (this.updatedUser == null) { return null; }
        return this.updatedUser.getId();
    }

}
