/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

import javax.jdo.PersistenceManager;
import javax.jdo.annotations.IdGeneratorStrategy;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;
import javax.jdo.annotations.PrimaryKey;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.annotations.*;

@MugleData
@PersistenceCapable
public class GameData extends ModelDataClass<Long, Game, GameData> {


    // Fields

    @PrimaryKey
    @Persistent(valueStrategy = IdGeneratorStrategy.IDENTITY)
    private Key id;

    @Persistent
    private String urlName; //Must be unique

    @Persistent
    private String gameToken;

    @Persistent
    private Boolean isPublic;

    @Persistent
    private Integer userRating;

    @Persistent
    private Integer totalRatings; // number of users who have rated this game

    @Persistent
    private Key devTeam;

    @Persistent
    private Set<Key> achievements;

    @Persistent
    private Set<Key> gameVersions;

    // Avoiding multiple parents
    @Persistent
    private Set<Key> ugps;

    @Persistent
    private Key activeVersion;

    @Persistent
    private Key createdUser;

    @Persistent
    private Date createdTime;

    @Persistent
    private Key updatedUser;

    @Persistent
    private Date updatedTime;


    // Constructors

    public GameData() {
        this.gameVersions = new HashSet<Key>();
        this.ugps = new HashSet<Key>();
        this.achievements = new HashSet<Key>();
        this.gameToken = UUID.randomUUID().toString();
        this.urlName = null;
        this.isPublic = false;
        this.userRating = 0;
        this.totalRatings = 0;
        this.devTeam = null;
        this.activeVersion = null;
    }

    // Getters

    public GameData(String gameName) {
        this.gameVersions = new HashSet<Key>();
        this.ugps = new HashSet<Key>();
        this.achievements = new HashSet<Key>();
        this.gameToken = UUID.randomUUID().toString();
        this.urlName = gameName;
        this.isPublic = false;
        this.userRating = 0;
        this.totalRatings = 0;
        this.devTeam = null;
        this.activeVersion = null;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setPrimaryKey")
    public Long getPrimaryKey() {
        if (this.id == null) {
            return null;
        }
        return this.id.getId();
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUrlName")
    public String getUrlName() {
        return this.urlName;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setIsPublic")
    public Boolean getIsPublic() {
        return this.isPublic;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUserRating")
    public Integer getUserRating() {
        return this.userRating;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedTo="setTotalRatings")
    public Integer getTotalRatings() {
        return this.totalRatings;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setDevTeamKey", overrideBy="keyToDevTeam")
    public Key getDevTeam() {
        return this.devTeam;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, list=true)
    public Set<Key> getAchievements() {
        return this.achievements;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, list=true)
    public Set<Key> getGameVersions() {
        return this.gameVersions;
    }

    @GetterUserLevel(privateView=Role.NONE, publicView=Role.NONE, list=true)
    public Set<Key> getUserGameProfiles() {
        return this.ugps;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedTo="setGameToken")
    public String getGameToken() {
        return this.gameToken;
    }

    // This field should be manually accessed as the wrapper has an object and not a key
    //@GetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public Key getActiveVersion() {
        return this.activeVersion;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedUser", overrideBy="keyToCreated")
    public Key getCreatedUser() {
        return this.createdUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedTime")
    public Date getCreatedTime() {
        return this.createdTime;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedUser", overrideBy="keyToUpdated")
    public Key getUpdatedUser() {
        return this.updatedUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedTime")
    public Date getUpdatedTime() {
        return this.updatedTime;
    }


    // Setters

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedUser(Key createdUser) {
        this.createdUser = createdUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedTime(Date createdTime) {
        this.createdTime = createdTime;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedUser(Key updatedUser) {
        this.updatedUser = updatedUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedTime(Date updatedTime) {
        this.updatedTime = updatedTime;
    }

    @SetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedBy="getUrlName")
    public void setUrlName(String newName) {
        this.urlName = newName;
    }

    @SetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedBy="getIsPublic")
    public void setIsPublic(Boolean newIsPublic) {
        this.isPublic = newIsPublic;
    }

    @SetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedBy="getDevTeamKey", overrideBy="devTeamFromKey")
    public void setDevTeam(DevTeamData devTeam) {
        this.devTeam = devTeam.getServerKey();
        devTeam.getGames().add(this.id);
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getTotalRatings")
    public void setTotalRatings(Integer totalRatings) {
        this.totalRatings = totalRatings;
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getUserRating")
    public void setUserRating(Integer userRating) {
        this.userRating = userRating;
    }

    @SetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedBy="getActiveVersion", overrideBy="versionToKey")
    public void setActiveVersion(Key version) {
        this.activeVersion = version;
    }


    // Misc methods

    /* Should only be able to add a new rating from the page when logged in,
     * not available to students programs to modify
     */
    public void updateUserRating(int newRating) throws InvalidRatingException {
        // Ratings can only be between 0 and MAX_RATING
        if (newRating < 0) {
            throw new InvalidRatingException(newRating);
        } else if (newRating > Game.MAX_RATING) {
            throw new InvalidRatingException(newRating);
        }
        //TODO: need to make sure each user can only have 1 rating
        // Calculates via average.
        Long temp;
        temp = (long)(this.userRating * this.totalRatings);
        //TODO: code for temp long overflow
        temp += newRating;
        //TODO: code for totalRatings integer overflow
        this.totalRatings += 1;
        this.userRating = (int)(temp / this.totalRatings);
    }

    public void addAchievement(AchievementData achievement) {
        achievement.setGame(this);
    }

    public void removeAchievement(AchievementData achievement) {
        this.achievements.remove(achievement.getServerKey());
    }

    public void addGameVersion(GameVersionData gameversion) {
        gameversion.setGame(this);
    }

    public void removeGameVersion(GameVersionData gameversion) {
        this.gameVersions.remove(gameversion.getServerKey());
    }

    public void addUserGameProfile(UserGameProfileData ugp) {
        ugp.setGame(this);
    }

    public void removeUserGameProfile(UserGameProfileData ugp) {
        this.ugps.remove(ugp.getServerKey());
    }

    @Override
    public Key getServerKey() {
        return this.id;
    }

    @Override
    public void setServerKey(Key key) {
        this.id = key;
    }

    @Override
    public void ClearFields() {
        this.setDevTeam(null);
        this.setServerKey(null);
        this.setTotalRatings(null);
        this.setUrlName(null);
        this.setUserRating(null);
    }

    @Override
    public Class<Game> getClientClassType() {
        return Game.class;
    }

    public Long keyToDevTeam() {
        if (this.devTeam == null) {
            return null;
        }
        return this.devTeam.getId();
    }

    public DevTeamData devTeamFromKey(Long key) {
        if (key == null) { return null; }
        PersistenceManager pm = PMF.getManager();
        DevTeamData result = null;
        try {
            result = pm.getObjectById(DevTeamData.class, key);
        } finally {
            pm.close();
        }
        return result;
    }

    public Key versionToKey(GameVersion version) {
        if (version == null) { return null; }
        PersistenceManager pm = PMF.getManager();
        GameVersionData result = null;
        try {
            result = pm.getObjectById(GameVersionData.class, version.getPrimaryKey());
        } finally {
            pm.close();
        }
        if (result == null) { return null; }
        return result.getServerKey();
    }

    public Long keyToCreated() {
        if (this.createdUser == null) { return null; }
        return this.createdUser.getId();
    }

    public Long keyToUpdated() {
        if (this.updatedUser == null) { return null; }
        return this.updatedUser.getId();
    }

}
