/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.Date;

import javax.jdo.PersistenceManager;
import javax.jdo.annotations.IdGeneratorStrategy;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;
import javax.jdo.annotations.PrimaryKey;
import javax.activation.MimetypesFileTypeMap;

import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.Blob;

import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.annotations.*;

@MugleData
@PersistenceCapable
public class GameFileData extends ModelDataClass<Long, GameFile, GameFileData>  {


    // Fields

    @PrimaryKey //Assuming different games can't share files
    @Persistent(valueStrategy = IdGeneratorStrategy.IDENTITY)
    private Key id;

    @Persistent
    private String path; //path to the file relative to the Game URL

    @Persistent
    private String mimeType; //mime type of the file

    @Persistent
    private Key contents; //key to the GameFileContents

    @Persistent
    private Key version; //Game version this Gamefile belongs to

    @Persistent
    private Key createdUser;

    @Persistent
    private Date createdTime;

    @Persistent
    private Key updatedUser;

    @Persistent
    private Date updatedTime;


    // Constructors
    public GameFileData() {
        this.version = null;
        this.setMimeType(null);
        this.path = null;
    }

    /** Create a new GameFileData object.
     * Guesses the file's MIME type based on the file extension.
     * @param version The game version the file belongs to.
     * @param path Path to the file relative to this gameversion.
     * @param contents Blob containing the file's contents.
     */
    public GameFileData(GameVersionData version, String path,
        GameFileContents contents)
    {
        this.version = version.getServerKey();
        this.path = path;
        this.setMimeTypeFromFilename();
        this.contents = contents.getServerKey();
    }

    /** Create a new GameFileData object.
     * @param version The game version the file belongs to.
     * @param path Path to the file relative to this gameversion.
     * @param mimeType MIME type of the file.
     * @param contents Blob containing the file's contents.
     */
    public GameFileData(GameVersionData version, String path, String mimeType,
        GameFileContents contents)
    {
        this.version = version.getServerKey();
        this.path = path;
        this.mimeType = mimeType;
        this.contents = contents.getServerKey();
    }

    /* When a file is moved, or renamed - The existing GameFile entry should be deleted,
     * in which case, we may want to initialise with multiple Game versions.
     */
/**  TODO: Pending multiple versions of games
    public GameFile (String path, Set<String> versions, InputStream contents) {
        this.path = path;
        this.versions = versions;
        this.setContents(contents);
    }
*/

    // Getters

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setPrimaryKey")
    @Override
    public Long getPrimaryKey() {
        if (this.id == null) {
            return null;
        }
        return this.id.getId();
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setPath")
    public String getPath() {
        return this.path;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setGameVersionKey", overrideBy="keyToVersion")
    public Key getGameVersions() {
        return this.version;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setMimeType")
    public String getMimeType() {
        return mimeType;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedUser", overrideBy="keyToCreated")
    public Key getCreatedUser() {
        return this.createdUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedTime")
    public Date getCreatedTime() {
        return this.createdTime;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedUser", overrideBy="keyToUpdated")
    public Key getUpdatedUser() {
        return this.updatedUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedTime")
    public Date getUpdatedTime() {
        return this.updatedTime;
    }


    // Setters

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedUser(Key createdUser) {
        this.createdUser = createdUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedTime(Date createdTime) {
        this.createdTime = createdTime;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedUser(Key updatedUser) {
        this.updatedUser = updatedUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedTime(Date updatedTime) {
        this.updatedTime = updatedTime;
    }

    public void setPrimaryKey(String path) {
        this.setPath(path);
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getPath")
    public void setPath(String path) {
        this.path = path;
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getGameVersionKey", overrideBy="versionFromKey")
    public void setGameVersion(GameVersionData version) {
        if (version == null) {
            this.version = null;
            return;
        }
        this.version = version.getServerKey();
        version.getGameFiles().add(this.id);
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getMimeType")
    public void setMimeType(String mimeType) {
        this.mimeType = mimeType;
    }

    // Misc methods

    /** Get the content blob of this file.
     * @return Key of the GameFileContents object.
     */
    public Key getContents() {
        return this.contents;
    }

    /** Set the content blob of this file.
     * @param contents Key of the GameFileContents object.
     */
    public void setContents(Key contents) {
        this.contents = contents;
    }

    /** Set the MIME type based on the file extension.
     */
    public void setMimeTypeFromFilename() {
        String f = this.path.toLowerCase();
        // This implicitly looks at the MIME mapping table in
        // src/META-INF/mime.types.
        this.mimeType = new MimetypesFileTypeMap().getContentType(this.path);
    }

    @Override
    public Key getServerKey() {
        return this.id;
    }

    @Override
    public void setServerKey(Key key) {
        this.id = key;
    }

    @Override
    public void ClearFields() {
        this.setPath(null);
        this.setPrimaryKey(null);
    }

    @Override
    public Class<GameFile> getClientClassType() {
        return GameFile.class;
    }

    public Long keyToVersion() {
        if (this.version == null) {
            return null;
        }
        return this.version.getId();
    }

    public GameVersionData versionFromKey(Long key) {
        if (key == null) { return null; }
        PersistenceManager pm = PMF.getManager();
        GameVersionData result = null;
        try {
            result = pm.getObjectById(GameVersionData.class, key);
        } finally {
            pm.close();
        }
        return result;
    }

    public Long keyToCreated() {
        if (this.createdUser == null) { return null; }
        return this.createdUser.getId();
    }

    public Long keyToUpdated() {
        if (this.updatedUser == null) { return null; }
        return this.updatedUser.getId();
    }

}
