/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.List;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.GameFileNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.GameFileContentsNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.GameNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.GameVersionNotExists;

public class GameFileGetter {
    /**
     * Gets the GameFile by its Primary Key - READ ONLY
     * @param primaryKey
     * @return the GameFile - read only
     * @throws GameFileNotExists 
     */
    public static GameFileData getGameFile(Long primaryKey) throws GameFileNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameFile(pm, primaryKey);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the GameFile by its Primary Key, for editing object in datastore
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param primaryKey
     * @return the GameFile
     * @throws GameFileNotExists 
     */
    public static GameFileData getGameFile(PersistenceManager pm, Long primaryKey) throws GameFileNotExists{
        try {
            GameFileData gameFile = pm.getObjectById(GameFileData.class, primaryKey);
            return gameFile;
        } catch (JDOObjectNotFoundException e) {
            throw new GameFileNotExists(primaryKey);
        }
    }
    
    /**
     * Gets the GameFile by its path, version and game - READ ONLY
     * @param gameName the name of the game
     * @param gameVersion the version of the game
     * @param path the path of the file
     * @return the GameFile - read only
     * @throws GameNotExists
     * @throws GameVersionNotExists
     * @throws GameFileNotExists 
     */
    public static GameFileData getGameFile(String devteamName,String gameName,
            String gameVersion, String path)
        throws GameNotExists, GameVersionNotExists, GameFileNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameFile(pm, devteamName, gameName, gameVersion, path);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the GameFile by its path, version and game, for editing object in datastore
     * PersistenceManager must be handled by the caller
     * @para pm The Persistence Manager
     * @param gameName the name of the game
     * @param gameVersion the version of the game
     * @param path the path of the file
     * @return the GameFile - read only
     * @throws GameNotExists
     * @throws GameVersionNotExists
     * @throws GameFileNotExists 
     */
    @SuppressWarnings("unchecked")
    public static GameFileData getGameFile(PersistenceManager pm,
        String devteamName, String gameName, String gameVersion, String path)
        throws GameFileNotExists, GameNotExists, GameVersionNotExists {
        
        GameVersionData gv = GameVersionGetter.getGameVersion(pm, devteamName,
            gameName, gameVersion);
                
        Query q = pm.newQuery(GameFileData.class, "path == p && version == v");
        q.declareParameters("String p, com.google.appengine.api.datastore.Key v");
        List<GameFileData> results = (List<GameFileData>) q.execute(path, gv.getServerKey());

        if (results.isEmpty()) {
            throw new GameFileNotExists(gameName, gameVersion, path);
        }
        
        return results.get(0);
    }
    
    /**
     * Gets the GameFile by its path, version and game, for editing object in datastore
     * @param gameName the name of the game
     * @param gameVersion the version of the game
     * @param path the path of the file
     * @return the GameFile - read only
     * @throws GameNotExists
     * @throws GameVersionNotExists
     * @throws GameFileNotExists 
     */
    public static GameFileData getGameFile(GameData game, GameVersionData gameVersion, String path)
        throws GameFileNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameFile(pm, game, gameVersion, path);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the GameFile by its path, version and game, for editing object in datastore
     * PersistenceManager must be handled by the caller
     * @para pm The Persistence Manager
     * @param gameName the name of the game
     * @param gameVersion the version of the game
     * @param path the path of the file
     * @return the GameFile - read only
     * @throws GameNotExists
     * @throws GameVersionNotExists
     * @throws GameFileNotExists 
     */
    @SuppressWarnings("unchecked")
    public static GameFileData getGameFile(PersistenceManager pm, GameData game, GameVersionData gameVersion, String path)
        throws GameFileNotExists {

            Query q = pm.newQuery(GameFileData.class, "path == p && version == v");
            q.declareParameters("String p, com.google.appengine.api.datastore.Key v");
            List<GameFileData> results = (List<GameFileData>) q.execute(path, gameVersion.getServerKey());

            if (results.isEmpty()) {
                throw new GameFileNotExists(game.getUrlName(),
                    gameVersion.getDisplayVersion(), path);
            }
            return results.get(0);
    }

    /**
     * Gets the GameFile by its path, version and game, for editing object in datastore
     * PersistenceManager must be handled by the caller
     * @para pm The Persistence Manager
     * @param gameName the name of the game
     * @param gameVersion the version of the game
     * @param path the path of the file
     * @return the GameFile - read only
     * @throws GameNotExists
     * @throws GameVersionNotExists
     * @throws GameFileNotExists
     */
    @SuppressWarnings("unchecked")
    public static GameFileData getGameFile(PersistenceManager pm, GameData game, int gameVersion, String path) 
        throws GameFileNotExists, GameNotExists, GameVersionNotExists {
        
        GameVersionData gvd = GameVersionGetter.getGameVersion(game, gameVersion);
        Query q = pm.newQuery(GameFileData.class, "path == p && version == v");
        q.declareParameters("String p, com.google.appengine.api.datastore.Key v");
        List<GameFileData> results = (List<GameFileData>) q.execute(path, gvd.getServerKey());

        if (results.isEmpty()) {
            throw new GameFileNotExists(game.getUrlName(),
                Integer.toString(gameVersion), path);
        }
        
        return results.get(0);
    }
    
    /**
     * Gets the GameFile by its path and version, for editing object in
     * datastore.
     * PersistenceManager must be handled by the caller
     * @para pm The Persistence Manager
     * @param gameVersion the version of the game
     * @param path the path of the file
     * @return the GameFile - read only
     * @throws GameFileNotExists
     */
    @SuppressWarnings("unchecked")
    public static GameFileData getGameFile(PersistenceManager pm,
        GameVersionData gameVersion, String path)
        throws GameFileNotExists {

        Query q = pm.newQuery(GameFileData.class, "path == p && version == v");
        q.declareParameters("String p, com.google.appengine.api.datastore.Key v");
        List<GameFileData> results = (List<GameFileData>) q.execute(path,
            gameVersion.getServerKey());

        if (results.isEmpty()) {
            throw new GameFileNotExists(gameVersion.getDisplayVersion(), path);
        }

        return results.get(0);
    }

    /**
     * Gets the GameFile associated with the given GameFileData primary key.
     * @param pm The PersistenceManager
     * @param key The server-side key (ModelDataClass.getServerKey())
     * @return The GameFileData object directly off the datastore.
     * @throws GameFileNotExists 
     */
    public static GameFileData getGameFile(PersistenceManager pm, Key key)
            throws GameFileNotExists {
        try {
            GameFileData gameFile = pm.getObjectById(GameFileData.class, key);
            return gameFile;
        } catch (JDOObjectNotFoundException e) {
            throw new GameFileNotExists(key.getId());
        }
    }

    /**
     * Gets the GameFileContents associated with the given GameFile.
     * @param file The file to retrieve contents for.
     * @return The GameFileContents object directly off the datastore.
     * @throws GameFileContentsNotExists
     */
    public static GameFileContents getGameFileContents(GameFileData file)
        throws GameFileContentsNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameFileContents(pm, file);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the GameFileContents associated with the given GameFile.
     * @param pm The PersistenceManager
     * @param file The file to retrieve contents for.
     * @return The GameFileContents object directly off the datastore.
     * @throws GameFileContentsNotExists
     */
    public static GameFileContents getGameFileContents(PersistenceManager pm,
        GameFileData file) throws GameFileContentsNotExists {
        Key key = file.getContents();

        try {
            GameFileContents result = pm.getObjectById(GameFileContents.class,
                                                   key);
            return result;
        } catch (JDOObjectNotFoundException e) {
            throw new GameFileContentsNotExists(key.getId());
        }
    }
}
