/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.List;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

public class GameGetter {
    /**
     * Gets the Game by its Primary Key - READ ONLY
     * @param primaryKey
     * @return the Game - read only
     * @throws GameNotExists 
     */
    public static GameData getGame(Long primaryKey) throws GameNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGame(pm, primaryKey);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the Game by its Primary Key, for editing object in datstore
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param primaryKey
     * @return the Game
     * @throws GameNotExists 
     */
    public static GameData getGame(PersistenceManager pm, Long primaryKey) throws GameNotExists{
        try {
            GameData game = pm.getObjectById(GameData.class, primaryKey);
            return game;
        } catch (JDOObjectNotFoundException e) {
            throw new GameNotExists(primaryKey);
        }
    }
    
    /**
     * Gets the Game by its name - READ ONLY
     * @param name
     * @return the Game - read only
     * @throws GameNotExists 
     */
    public static GameData getGameByName(String devteamName, String name) throws GameNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameByName(pm, devteamName, name);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the Game by name, for editing object in datstore
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param name
     * @return the Game
     * @throws GameNotExists 
     */
    @SuppressWarnings("unchecked")
    public static GameData getGameByName(PersistenceManager pm,
            String devteamName, String name) throws GameNotExists{
        Query q = pm.newQuery(DevTeamData.class, "urlName == u");
        q.declareParameters("String u");
        List<DevTeamData> devteams= (List<DevTeamData>)q.execute(devteamName);
        DevTeamData devteam = devteams.get(0);

        q = pm.newQuery(GameData.class, "devTeam == d && urlName == u");
        q.declareParameters("com.google.appengine.api.datastore.Key d, String u");
        List<GameData> results = (List<GameData>)
            q.execute(devteam.getServerKey(), name);
        
        if(results.isEmpty()) {
            throw new GameNotExists(name);
        }
        
        return results.get(0);
    }
    
    /**
     * Gets the Game from a provided gameToken - READ ONLY
     * @param gameToken the secret game token for the current game, provided by the client
     * @return the Game - read only
     * @throws GameTokenError
     */
    public static GameData getCurrentGame(String gameToken) throws GameTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            return getCurrentGame(pm, gameToken);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the Game from a provided gameToken, for editing object in datstore
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param gameToken the secret game token for the current game, provided by the client
     * @return the Game
     * @throws GameTokenError
     */
    public static GameData getCurrentGame(PersistenceManager pm, String gameToken) throws GameTokenError {
        return getGameByToken(pm, gameToken);
    }
    
    /**
     * Gets the Game from a provided gameToken - READ ONLY
     * @param gameToken the secret game token for the current game, provided by the client
     * @return the Game - read only
     * @throws GameTokenError
     */
    public static GameData getGameByToken(String gameToken) throws GameTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameByToken(pm, gameToken);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the Game from a provided gameToken, for editing object in datastore
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param gameToken the secret game token for the current game, provided by the client
     * @return the Game
     * @throws GameTokenError
     */
    @SuppressWarnings("unchecked")
    public static GameData getGameByToken(PersistenceManager pm, String gameToken) throws GameTokenError {
        Query q = pm.newQuery(GameData.class, "gameToken == g");
        q.declareParameters("String g");
        List<GameData> results = (List<GameData>) q.execute(gameToken);
        
        if(results.isEmpty()) {
            throw new GameTokenError(gameToken);
        }
        
        return results.get(0);
    }
    /**
     * Gets the Game by its Primary Key - READ ONLY
     * @param primaryKey
     * @return the Game - read only
     * @throws GameNotExists
     */
    public static GameData getGame(Key primaryKey) throws GameNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGame(pm, primaryKey);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the Game associated with the given GameData primary key.
     * @param pm The PersistenceManager
     * @param key The server-side key (ModelDataClass.getServerKey())
     * @return The GameData object directly off the datastore.
     * @throws GameNotExists 
     */
    public static GameData getGame(PersistenceManager pm, Key key)
            throws GameNotExists {
        try {
            GameData game = pm.getObjectById(GameData.class, key);
            return game;
        } catch (JDOObjectNotFoundException e) {
            throw new GameNotExists(key.getId());
        }
    }
   
    /**
     * Gets the Game by the name of the DevTeam it belongs to - READ ONLY
     * @param devTeamName name of the DevTeam
     * @param gameName name of the Game
     * @return the Game - read only
     * @throws GameNotExists 
     * @throws DevTeamNotExists 
     */
    public static GameData getGameByDevTeam(String devTeamName,
        String gameName) throws GameNotExists, DevTeamNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameByDevTeam(pm, devTeamName, gameName);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the Game by the name of the DevTeam it belongs to, for editing object in datstore
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param name name of the DevTeam
     * @return the Game
     * @throws GameNotExists 
     * @throws DevTeamNotExists 
     */
    @SuppressWarnings("unchecked")
    public static GameData getGameByDevTeam(PersistenceManager pm,
        String devTeamName, String gameName)
        throws GameNotExists, DevTeamNotExists{
        DevTeamData dtd = DevTeamGetter.getDevTeam(devTeamName);
        Query q = pm.newQuery(GameData.class, "devTeam == d && urlName == g");
        q.declareParameters("com.google.appengine.api.datastore.Key d, String g");
        List<GameData> results = (List<GameData>)
            q.execute(dtd.getServerKey(), gameName);
        
        if(results.isEmpty()) {
            throw new GameNotExists(devTeamName, gameName);
        }
        
        return results.get(0);
    }
}
