/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.List;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

public class GameVersionGetter {
    /**
     * Gets the GameVersion by its Primary Key - READ ONLY
     * @param primaryKey
     * @return the GameVersion - read only
     * @throws GameVersionNotExists 
     */
    public static GameVersionData getGameVersion(Long primaryKey) throws GameVersionNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameVersion(pm, primaryKey);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the GameVersion by its Primary Key, for editing object in datstore
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param primaryKey
     * @return the GameVersion
     * @throws GameVersionNotExists 
     */
    public static GameVersionData getGameVersion(PersistenceManager pm, Long primaryKey) throws GameVersionNotExists{
        try {
            GameVersionData gameversion = pm.getObjectById(GameVersionData.class, primaryKey);
            return gameversion;
        } catch (JDOObjectNotFoundException e) {
            throw new GameVersionNotExists(primaryKey);
        }   
    }
    
    /**
     * Gets the GameVersion by its display Version - READ ONLY
     * @param gameName the name of the Game that it belongs to
     * @param dispVersion the displayed version number
     * @return the GameVersion - read only
     * @throws GameVersionNotExists 
     */
    public static GameVersionData getGameVersion(String devteamName,
        String gameName, String dispVersion)
        throws GameVersionNotExists, GameNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameVersion(pm, devteamName, gameName, dispVersion);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the GameVersion by its display name, for editing object in datastore
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param gameName the name of the Game that it belongs to
     * @param dispVersion the displayed version number
     * @return the GameVersion
     * @throws GameVersionNotExists
     * @throws GameNotExists 
     */
    @SuppressWarnings("unchecked")
    public static GameVersionData getGameVersion(PersistenceManager pm,
        String devteamName, String gameName, String dispVersion)
        throws GameVersionNotExists, GameNotExists{

        GameData game = GameGetter.getGameByName(pm, devteamName, gameName);
        
        Query q = pm.newQuery(GameVersionData.class, "dispVersion == d && game == g");
        q.declareParameters("String d, com.google.appengine.api.datastore.Key g");
        List<GameVersionData> results = (List<GameVersionData>)
            q.execute(dispVersion, game.getServerKey());
        
        if(results.isEmpty()) {
            throw new GameVersionNotExists(dispVersion, game.getUrlName());
        }
        
        return results.get(0);
    }

    /**
     * Gets the GameVersion by its version number
     * @param game the Game that this version belongs to
     * @param version the version number
     * @return the GameVersion
     * @throws GameVersionNotExists
     * @throws GameNotExists 
     */
    public static GameVersionData getGameVersion(GameData game, int version) 
        throws GameNotExists, GameVersionNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameVersion(pm, game, version);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the GameVersion by its version number
     * PersistenceManager must be handled by the caller
     * @param pm the PersistenceManager
     * @param game the Game that this version belongs to
     * @param version the version number
     * @return the GameVersion
     * @throws GameVersionNotExists
     * @throws GameNotExists 
     */
    @SuppressWarnings("unchecked")
    public static GameVersionData getGameVersion(PersistenceManager pm,
        GameData game, int version) 
        throws GameNotExists, GameVersionNotExists {

        Query q = pm.newQuery(GameVersionData.class, "version == v && game == g");
        q.declareParameters("Integer v, com.google.appengine.api.datastore.Key g");
        List<GameVersionData> results = (List<GameVersionData>)
            q.execute((Integer)version, game.getServerKey());
        
        if(results.isEmpty()) {
            throw new GameVersionNotExists(version, game.getUrlName());
        }
        
        return results.get(0);
    }
    
    /**
     * Gets the GameVersion by its Primary Key - READ ONLY
     * @param primaryKey
     * @return the GameVersion - read only
     * @throws GameVersionNotExists 
     */
    public static GameVersionData getGameVersion(Key primaryKey) throws GameVersionNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getGameVersion(pm, primaryKey);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the GameVersion associated with the given GameVersionData primary key.
     * @param pm The PersistenceManager
     * @param key The server-side key (ModelDataClass.getServerKey())
     * @return The GameVersionData object directly off the datastore.
     * @throws GameVersionNotExists 
     */
    public static GameVersionData getGameVersion(PersistenceManager pm, Key key)
            throws GameVersionNotExists {
        try {
            GameVersionData gameversion = pm.getObjectById(GameVersionData.class, key);
            return gameversion;
        } catch (JDOObjectNotFoundException e) {
            throw new GameVersionNotExists(key.getId());
        }
    }
    
}

