/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Date;

import javax.jdo.PersistenceManager;
import javax.jdo.annotations.IdGeneratorStrategy;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;
import javax.jdo.annotations.PrimaryKey;

import com.google.appengine.api.datastore.Blob;
import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.annotations.*;

@MugleData
@PersistenceCapable
public class KeyValuePairData extends ModelDataClass<Long, KeyValuePair, KeyValuePairData> {


    // Fields

    @PrimaryKey
    @Persistent(valueStrategy = IdGeneratorStrategy.IDENTITY)
    private Key id;

    @Persistent
    private Key ugp;

    // The actual key
    @Persistent
    private String key;

    @Persistent(serialized = "true")
    private Blob value;

    @Persistent
    private Key createdUser;

    @Persistent
    private Date createdTime;

    @Persistent
    private Key updatedUser;

    @Persistent
    private Date updatedTime;


    // Constructors

    public KeyValuePairData() {

        this.id = null;
        this.key = null;
        this.ugp = null;
        this.value = null;

    }

    // ALL methods are protected since they can only be accessed by the appropriate User+Game
    public KeyValuePairData(UserGameProfileData ugp, String key, Serializable value) {
        //TODO: Check the User hasn't used up all their space

        this.key = key;
        this.ugp = ugp.getServerKey();
        ugp.getKeyValuePairs().add(this.id);
        this.setValue(value);

    }


    // Getters

    /**
     * Returns the primary key made up of the UGP ID and the Key used for key value store
     * separated by a /
     * @return the primary key
     */
    @GetterUserLevel(privateView=Role.NONE, publicView=Role.NONE, mappedTo="setPrimaryKey")
    public Long getPrimaryKey() {
        if (this.id == null) {
            return null;
        }
        return this.id.getId();
    }

    // Gets the Key used for storage
    @GetterUserLevel(privateView=Role.NONE, publicView=Role.NONE, mappedTo="setKey")
    public String getKey() {
        return this.key;
    }

    @GetterUserLevel(privateView=Role.NONE, publicView=Role.NONE, mappedTo="setUserGameProfileKey", overrideBy="keyToUGP")
    public Key getUserGameProfile() {
        return this.ugp;
    }

    @GetterUserLevel(privateView=Role.NONE, publicView=Role.NONE, mappedTo="setValue")
    public Serializable getValue() {
        ByteArrayInputStream byteStream = new ByteArrayInputStream(this.value.getBytes());
        try {
            Object obj = (new ObjectInputStream(byteStream)).readObject();
            return (Serializable)obj;
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return null;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedUser", overrideBy="keyToCreated")
    public Key getCreatedUser() {
        return this.createdUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedTime")
    public Date getCreatedTime() {
        return this.createdTime;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedUser", overrideBy="keyToUpdated")
    public Key getUpdatedUser() {
        return this.updatedUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedTime")
    public Date getUpdatedTime() {
        return this.updatedTime;
    }


    // Setters

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedUser(Key createdUser) {
        this.createdUser = createdUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedTime(Date createdTime) {
        this.createdTime = createdTime;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedUser(Key updatedUser) {
        this.updatedUser = updatedUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedTime(Date updatedTime) {
        this.updatedTime = updatedTime;
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getKey")
    public void setKey(String key) {
        this.key = key;
    }

    public Boolean setValue(Serializable value) {
        ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        ObjectOutput out;
        try {
            out = new ObjectOutputStream(byteStream);
            out.writeObject(value);
            this.value = new Blob(byteStream.toByteArray());
            return true;
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return false;
    }

    @SetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedBy="getUserGameProfileKey", overrideBy="ugpFromKey")
    public void setUserGameProfile(UserGameProfileData ugp) {
        if (ugp == null) {
            this.ugp = null;
            return;
        }
        this.ugp = ugp.getServerKey();
        ugp.getKeyValuePairs().add(this.id);
    }

    @Override
    public Key getServerKey() {
        return this.id;
    }

    @Override
    public void setServerKey(Key key) {
        this.id = key;
    }

    @Override
    public void ClearFields() {
        this.setKey(null);
        this.setServerKey(null);
        this.setValue(null);
    }


    @Override
    public Class<KeyValuePair> getClientClassType() {
        return KeyValuePair.class;
    }

    public Long keyToUGP() {
        if (this.ugp == null) {
            return null;
        }
        return this.ugp.getId();
    }

    public UserGameProfileData ugpFromKey(Long key) {
        if (key == null) { return null; }
        PersistenceManager pm = PMF.getManager();
        UserGameProfileData result = null;
        try {
            result = pm.getObjectById(UserGameProfileData.class, key);
        } finally {
            pm.close();
        }
        return result;
    }

    public Long keyToCreated() {
        if (this.createdUser == null) { return null; }
        return this.createdUser.getId();
    }

    public Long keyToUpdated() {
        if (this.updatedUser == null) { return null; }
        return this.updatedUser.getId();
    }

}
