/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model;

import java.util.List;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.GameInstanceData;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.GameInstanceGetter;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.api.KeyError;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.InstanceTokenError;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.KeyValuePairNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.UserGameProfileNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.UserNotExists;

public class KeyValuePairGetter {

    /**
     * Gets the KeyValuePair for the key provided from the datastore for the
     * player currently playing - READ ONLY
     * 
     * @param key
     *            the key to look up of the KeyValuePair
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @return The request KeyValuePair - READ ONLY
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws KeyError
     */
    public static KeyValuePairData getKeyValuePair(String key, String gameToken)
            throws GameTokenError, KeyError {
        PersistenceManager pm = PMF.getManager();
        try {
            return getKeyValuePair(pm, key, gameToken);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the KeyValuePair for the key provided from the datastore for the
     * player currently playing, for editing object in datastore The Persistence
     * Manager must be handled by the caller
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            the key to look up of the KeyValuePair
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @return The request KeyValuePair
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws KeyError
     */
    public static KeyValuePairData getKeyValuePair(PersistenceManager pm,
            String key, String gameToken) throws GameTokenError, KeyError {
        return getKeyValuePair(pm, key, gameToken, false);
    }

    /**
     * Gets the KeyValuePair for the key provided from the datastore for the
     * player currently playing, for editing object in datastore - if the
     * KeyValuePair isnt found the parameter createIfNotFound determines whether
     * an exception will be thrown, or if a new KeyValuePair is created and
     * associated to that User's UserGameProfile The Persistence Manager must be
     * handled by the caller
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            the key to look up of the KeyValuePair
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @param createIfNotFound
     *            boolean to determine whether a new KeyValuePair is created if
     *            its not found.s
     * @return The request KeyValuePair
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws KeyError
     */
    @SuppressWarnings("unchecked")
    public static KeyValuePairData getKeyValuePair(PersistenceManager pm,
            String key, String gameToken, boolean createIfNotFound)
            throws GameTokenError, KeyError {

        UserGameProfileData ugp = UserGameProfileGetter
                .getCurrentUserGameProfile(pm, gameToken);

        Query q = pm.newQuery(KeyValuePairData.class, "key == k && ugp == u");
        q.declareParameters("String k, com.google.appengine.api.datastore.Key u");
        List<KeyValuePairData> results = (List<KeyValuePairData>) q.execute(
                key, ugp.getServerKey());

        /*
         * If the key value pair isn't found - we want to create one if
         * createIfNotFound is set to true
         */
        if (results.isEmpty()) {
            if (createIfNotFound) {
                KeyValuePairData kvp = new KeyValuePairData();

                pm.makePersistent(kvp);
                // MUST happen after write to datastore
                kvp.setUserGameProfile(ugp);
                kvp.setKey(key);

                return kvp;
            } else {
                throw new KeyError(key);
            }
        }

        return results.get(0);
    }

    /**
     * Gets the KeyValuePair associated with the given KeyValuePairData primary
     * key.
     * 
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The KeyValuePairData object directly off the datastore.
     * @throws KeyValuePairNotExists
     */
    public static KeyValuePairData getKeyValuePair(Key key)
            throws KeyValuePairNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getKeyValuePair(pm, key);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the KeyValuePair associated with the given KeyValuePairData primary
     * key.
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The UserAchievementData object directly off the datastore.
     * @throws KeyValuePairNotExists
     */
    public static KeyValuePairData getKeyValuePair(PersistenceManager pm,
            Key key) throws KeyValuePairNotExists {
        try {
            KeyValuePairData kvp = pm
                    .getObjectById(KeyValuePairData.class, key);
            return kvp;
        } catch (JDOObjectNotFoundException e) {
            throw new KeyValuePairNotExists(key.getId());
        }
    }

    /**
     * Gets the KeyValuePair associated with the given KeyValuePairData primary
     * key.
     * 
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The KeyValuePairData object directly off the datastore.
     * @throws KeyValuePairNotExists
     */
    public static KeyValuePairData getKeyValuePair(Long key)
            throws KeyValuePairNotExists {

        PersistenceManager pm = PMF.getManager();
        try {
            return getKeyValuePair(pm, key);
        } finally {
            pm.close();
        }

    }

    /**
     * Gets the KeyValuePair associated with the given KeyValuePairData primary
     * key.
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The UserAchievementData object directly off the datastore.
     * @throws KeyValuePairNotExists
     */
    public static KeyValuePairData getKeyValuePair(PersistenceManager pm,
            Long key) throws KeyValuePairNotExists {
        try {
            KeyValuePairData kvp = pm
                    .getObjectById(KeyValuePairData.class, key);
            return kvp;
        } catch (JDOObjectNotFoundException e) {
            throw new KeyValuePairNotExists(key);
        }
    }

    /**
     * Gets the KeyValuePair for the key provided from the datastore for a user
     * , for editing object in datastore - if the KeyValuePair isnt found the
     * parameter createIfNotFound determines whether an exception will be
     * thrown, or if a new KeyValuePair is created and associated to that User's
     * UserGameProfile The Persistence Manager must be handled by the caller
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            the key to look up of the KeyValuePair
     * @param instanceToken
     *            The secret token for the current game instance.
     * @param createIfNotFound
     *            boolean to determine whether a new KeyValuePair is created if
     *            its not found.
     * @return The request InstanceKeyValuePair
     * @throws KeyError
     * @throws InstanceTokenError
     * @throws UserNotExists
     * @throws UserGameProfileNotExists
     * @throws GameTokenError
     */
    @SuppressWarnings("unchecked")
    public static KeyValuePairData getUserKeyValuePair(PersistenceManager pm,
            String userName, String key, String gameToken,
            boolean createIfNotFound) throws KeyError, UserNotExists,
            GameTokenError {
        GameData gd = GameGetter.getCurrentGame(gameToken);
        UserData user = UserGetter.getUserByName(userName);

        UserGameProfileData ugp = UserGameProfileGetter.getUserGameProfile(pm,
                user, gd);

        Query q = pm.newQuery(KeyValuePairData.class, "key == k && ugp == u");
        q.declareParameters("String k, com.google.appengine.api.datastore.Key u");
        List<KeyValuePairData> results = (List<KeyValuePairData>) q.execute(
                key, ugp.getServerKey());

        if (results.isEmpty()) {
            if (createIfNotFound) {
                KeyValuePairData kvp = new KeyValuePairData();

                pm.makePersistent(kvp);
                // MUST happen after write to datastore
                kvp.setUserGameProfile(ugp);
                kvp.setKey(key);

                return kvp;
            } else {
                throw new KeyError(key);
            }
        }

        return results.get(0);
    }

    /**
     * Gets the KeyValuePair for the key provided from the datastore for the
     * game instance player currently playing, for editing object in datastore -
     * if the KeyValuePair isnt found the parameter createIfNotFound determines
     * whether an exception will be thrown, or if a new KeyValuePair is created
     * and associated to that User's UserGameProfile The Persistence Manager
     * must be handled by the caller
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            the key to look up of the KeyValuePair
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @param createIfNotFound
     *            boolean to determine whether a new KeyValuePair is created if
     *            its not found.
     * @param instanceToken
     *            The secret token for the current game instance.
     * @return The request InstanceKeyValuePair
     * @throws KeyError
     * @throws InstanceTokenError
     */
    public static KeyValuePairData getInstanceKeyValuePair(String key,
            String instanceToken, boolean createIfNotFound) throws KeyError,
            InstanceTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            return getInstanceKeyValuePair(pm, key, instanceToken,
                    createIfNotFound);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the KeyValuePair for the key provided from the datastore for the
     * game instance player currently playing, for editing object in datastore -
     * if the KeyValuePair isnt found the parameter createIfNotFound determines
     * whether an exception will be thrown, or if a new KeyValuePair is created
     * and associated to that User's UserGameProfile The Persistence Manager
     * must be handled by the caller
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            the key to look up of the KeyValuePair
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @param createIfNotFound
     *            boolean to determine whether a new KeyValuePair is created if
     *            its not found.
     * @param instanceToken
     *            The secret token for the current game instance.
     * @return The request InstanceKeyValuePair
     * @throws KeyError
     * @throws InstanceTokenError
     */
    @SuppressWarnings("unchecked")
    public static KeyValuePairData getInstanceKeyValuePair(
            PersistenceManager pm, String key, String instanceToken,
            boolean createIfNotFound) throws KeyError, InstanceTokenError {

        UserGameProfileData ugp = UserGameProfileGetter
                .getCurrentUserGameInstanceProfile(pm, instanceToken);

        Query q = pm.newQuery(KeyValuePairData.class, "key == k && ugp == u");
        q.declareParameters("String k, com.google.appengine.api.datastore.Key u");
        List<KeyValuePairData> results = (List<KeyValuePairData>) q.execute(
                key, ugp.getServerKey());

        /*
         * If the key value pair isn't found - we want to create one if
         * createIfNotFound is set to true
         */
        if (results.isEmpty()) {
            if (createIfNotFound) {
                KeyValuePairData kvp = new KeyValuePairData();

                pm.makePersistent(kvp);
                // MUST happen after write to datastore
                kvp.setUserGameProfile(ugp);
                kvp.setKey(key);

                return kvp;
            } else {
                throw new KeyError(key);
            }
        }

        return results.get(0);
    }

    /**
     * Gets the KeyValuePair for the key provided from the datastore for a user,
     * for editing object in datastore - if the KeyValuePair isnt found the
     * parameter createIfNotFound determines whether an exception will be
     * thrown, or if a new KeyValuePair is created and associated to that User's
     * UserGameProfile The Persistence Manager must be handled by the caller
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            the key to look up of the KeyValuePair
     * @param instanceToken
     *            The secret token for the current game instance.
     * @param createIfNotFound
     *            boolean to determine whether a new KeyValuePair is created if
     *            its not found.
     * @return The request InstanceKeyValuePair
     * @throws KeyError
     * @throws InstanceTokenError
     * @throws UserNotExists
     */
    @SuppressWarnings("unchecked")
    public static KeyValuePairData getUserInstanceKeyValuePair(
            PersistenceManager pm, String userName, String key,
            String instanceToken, boolean createIfNotFound) throws KeyError,
            InstanceTokenError, UserNotExists {
        GameInstanceData gi = GameInstanceGetter
                .getInstanceByToken(instanceToken);
        UserData user = UserGetter.getUserByName(userName);

        UserGameProfileData ugp = UserGameProfileGetter
                .getUserGameInstanceProfile(pm, user, gi);

        Query q = pm.newQuery(KeyValuePairData.class, "key == k && ugp == u");
        q.declareParameters("String k, com.google.appengine.api.datastore.Key u");
        List<KeyValuePairData> results = (List<KeyValuePairData>) q.execute(
                key, ugp.getServerKey());

        if (results.isEmpty()) {
            if (createIfNotFound) {
                KeyValuePairData kvp = new KeyValuePairData();

                pm.makePersistent(kvp);
                // MUST happen after write to datastore
                kvp.setUserGameProfile(ugp);
                kvp.setKey(key);

                return kvp;
            } else {
                throw new KeyError(key);
            }
        }

        return results.get(0);
    }

}
