/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.lang.reflect.*;

import javax.jdo.PersistenceManager;

import au.edu.unimelb.csse.mugle.server.model.annotations.*;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.DataObjectNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.ServerException;

public final class ModelWrapper {

    // Empty private constructor to treat class as static
    private ModelWrapper() { }


    // Privilege Checks

    /**
     * Check whether a given user level has privileges
     * @param userLevel User level for a given method
     * @param role The role of the user to check against
     * @return true if the user has privileges, else false
     */
    protected static boolean isAllowed(Role userLevel, Role role) {

        switch (userLevel) {

            case ADMIN:
                return (role == Role.ADMIN);

            case DEVELOPER:
                return (role == Role.ADMIN || role == Role.DEVELOPER);

            case GUEST:
                return true;

            default:
                return false;

        }

    }

    // Wrapper - class conversions

    /**
     * Converts a ModelDataClass instance to its counter part ModelClass,
     * using the given privilege information.
     * @param <T> The wrapper class type (of type ModelClass)
     * @param <U> The datastore class type (of type ModelDataClass)
     * @param obj The Data Model object to be converted.
     * @param role The role of the user requesting this conversion.
     * @param cview The client view of the user requesting this conversion.
     * @return The stripped down clone, 'null' if 'obj' is 'null'.
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws NoSuchFieldException
     * @throws SecurityException
     * @throws InvocationTargetException
     * @throws IllegalArgumentException
     * @throws NoSuchMethodException
     */
    @SuppressWarnings("unchecked") // Unchecked, safe enough since clients are not allowed to change anything regarding the Model classes
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
        T getClientClone(U obj, Role role, ClientView cview)
            throws  InstantiationException, IllegalAccessException, SecurityException,
                    NoSuchFieldException, NoSuchMethodException, IllegalArgumentException,
                    InvocationTargetException {

        if (obj == null) { return null; }

        T stripped = obj.getClientClassType().newInstance();

        Method[] methods = obj.getClass().getMethods();

        // If we have a private view on this object, or the user is an admin,
        // tell the client that the user is allowed to edit it
        if (isAllowed(Role.ADMIN, role) || cview == ClientView.PRIVATE)
            stripped.setUserCanEdit();

        for (Method m : methods) {

            GetterUserLevel ul = m.getAnnotation(GetterUserLevel.class);
            if (ul == null) { continue; }
            Role r = (cview == ClientView.PRIVATE) ? ul.privateView() : ul.publicView();
            if (ul.list() || !isAllowed(r, role)) { continue; }

            Method sm = getMethodByName(stripped.getClass(), ul.mappedTo());

            Method invk = (ul.overrideBy().length() > 0) ? getDataMethodByName(obj.getClass(), ul.overrideBy()) : m;

            sm.invoke(stripped, invk.invoke(obj));

        }

        return stripped;

    }

    /**
     * Similar to getClienClone() but suppresses different types
     * of exceptions (throws single type of exception)
     * @param <T> The wrapper class type (of type ModelClass)
     * @param <U> The datastore class type (of type ModelDataClass)
     * @param obj The Data Model object to be converted.
     * @param roleAndView The role and client view of the user requesting this conversion.
     * @return The stripped down clone, 'null' if 'obj' is 'null'.
     * @throws ServerException
     */
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
        T getClientCloneSafe(U obj, RoleAndView roleAndView)
            throws ServerException {

        return getClientCloneSafe(obj, roleAndView.getRole(), roleAndView.getClientView());

    }

    /**
     * Similar to getClienClone() but suppresses different types
     * of exceptions (throws single type of exception)
     * @param <T> The wrapper class type (of type ModelClass)
     * @param <U> The datastore class type (of type ModelDataClass)
     * @param obj The Data Model object to be converted.
     * @param role The role of the user requesting this conversion.
     * @param cview The client view of the user requesting this conversion.
     * @return The stripped down clone, 'null' if 'obj' is 'null'.
     * @throws ServerException
     */
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
        T getClientCloneSafe(U obj, Role role, ClientView cview)
            throws ServerException {

        try {
            return getClientClone(obj, role, cview);
        } catch (SecurityException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (InstantiationException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (IllegalAccessException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (NoSuchFieldException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (InvocationTargetException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        }

    }

    /**
     * Gets a method by name for a given class type
     * @param <T> Any wrapper class (of type ModelClass)
     * @param cls The class type
     * @param name Name of method
     * @return
     * @throws NoSuchMethodException
     */
    private static <T extends ModelClientClass<?,T>>
        Method getMethodByName(Class<T> cls, String name)
            throws NoSuchMethodException {

        Method[] methods = cls.getMethods();
        for (Method m : methods) {
            if (m.getName().equals(name)) { return m; }
        }
        String error = (name.length() == 0) ? "Invalid method name (empty name)." :
            "Method '" + name + "' in class '" + cls.getName() + "' not found!";
        throw new NoSuchMethodException(error);
    }

    /**
     * Gets a method by name for a given class type
     * @param <T> Any wrapper class (of type ModelClass)
     * @param cls The class type
     * @param name Name of method
     * @return
     * @throws NoSuchMethodException
     */
    private static <U extends ModelDataClass<?,?,U>>
        Method getDataMethodByName(Class<U> cls, String name)
            throws NoSuchMethodException {

        Method[] methods = cls.getMethods();
        for (Method m : methods) {
            if (m.getName().equals(name)) { return m; }
        }
        String error = (name.length() == 0) ? "Invalid method name (empty name)." :
            "Method '" + name + "' in class '" + cls.getName() + "' not found!";
        throw new NoSuchMethodException(error);
    }

    /**
     * Converts a ModelClass instance to its counter part ModelDataClass,
     * using the given privilege information.
     * @param obj The Client Model object to be converted.
     * @param dataType Type of the Data Model class
     * @param role The role of the user requesting this conversion.
     * @param cview The client view of the user requesting this conversion.
     * @return The converted Data Model instance, 'null' if either of 'clientObj' or 'emptyData' is 'null'.
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws NoSuchMethodException
     * @throws SecurityException
     * @throws NoSuchFieldException
     * @throws InvocationTargetException
     * @throws IllegalArgumentException
     * @throws ServerException
     */
    @SuppressWarnings("unchecked")
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
        U getDataClone(T obj, Class<U> dataType, Role role, ClientView cview)
            throws InstantiationException, IllegalAccessException,
                    NoSuchMethodException, IllegalArgumentException,
                    InvocationTargetException, DataObjectNotExists {

        if (obj == null) { return null; }

        U dataObj = null;

        // Get original, if it doesn't exist then create a new instance
        PersistenceManager pm = PMF.getManager();
        try {

            //Query q = pm.newQuery(dataType, "id == i");
            //q.declareParameters("Long i");
            //List<U> results = (List<U>) q.execute(obj.getPrimaryKey());
            //dataObj = (results.isEmpty()) ? createNewDataObject(dataType) : results.get(0);

            if (obj.getPrimaryKey() != null) {
                dataObj = pm.getObjectById(dataType, obj.getPrimaryKey());
                if (dataObj == null) {
                    throw new DataObjectNotExists();
                }
            } else {
                throw new DataObjectNotExists();
            }

        } finally {
            pm.close();
        }

        Method[] methods = dataType.getMethods();
        for (Method m : methods) {

            SetterUserLevel ul = m.getAnnotation(SetterUserLevel.class);
            if (ul == null) { continue; }
            Role r = (cview == ClientView.PRIVATE) ? ul.privateView() : ul.publicView();
            if (!isAllowed(r, role)) { continue; }

            Method cm = getMethodByName(obj.getClass(), ul.mappedBy());

            Object value = cm.invoke(obj);
            if (ul.overrideBy().length() > 0) {
                value = getDataMethodByName(dataType, ul.overrideBy()).invoke(dataObj, value);
            }
            m.invoke(dataObj, value);

        }

        return dataObj;

    }

    /**
     * Converts a ModelClass instance to its counter part ModelDataClass,
     * using the given privilege information.
     * @param obj The Client Model object to be converted.
     * @param dataType Type of the Data Model class
     * @param roleAndView The role and client view of the user requesting this conversion.
     * @return The converted Data Model instance, 'null' if either of 'clientObj' or 'emptyData' is 'null'.
     * @throws ServerException
     */
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
    U getDataCloneSafe(T obj, Class<U> dataType, RoleAndView roleAndView)
        throws ServerException, DataObjectNotExists {

        return getDataCloneSafe(obj, dataType, roleAndView.getRole(), roleAndView.getClientView());

    }

    /**
     * Converts a ModelClass instance to its counter part ModelDataClass,
     * using the given privilege information.
     * @param obj The Client Model object to be converted.
     * @param dataType Type of the Data Model class
     * @param role The role of the user requesting this conversion.
     * @param cview The client view of the user requesting this conversion.
     * @return The converted Data Model instance, 'null' if either of 'clientObj' or 'emptyData' is 'null'.
     * @throws ServerException
     */
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
        U getDataCloneSafe(T obj, Class<U> dataType, Role role, ClientView cview)
            throws ServerException, DataObjectNotExists {

        try {
            return getDataClone(obj, dataType, role, cview);
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (InstantiationException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (IllegalAccessException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (InvocationTargetException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        }

    }

    /**
     * Creates a new instance of a ModelDataClass by converting a ModelClientClass
     * instance to its counter part ModelDataClass, using the given privilege information.
     * This function assumes all uniqueness in fields have been tested before hand.
     * @param obj The Client Model object to be converted.
     * @param dataType Type of the Data Model class
     * @param role The role of the user requesting this conversion.
     * @param cview The client view of the user requesting this conversion.
     * @return The converted Data Model instance, 'null' if either of 'clientObj' or 'emptyData' is 'null'.
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws NoSuchMethodException
     * @throws SecurityException
     * @throws NoSuchFieldException
     * @throws InvocationTargetException
     * @throws IllegalArgumentException
     * @throws ServerException
     */
    @SuppressWarnings("unchecked")
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
        U createDataClone(T obj, Class<U> dataType, Role role, ClientView cview)
            throws InstantiationException, IllegalAccessException,
                    NoSuchMethodException, IllegalArgumentException,
                    InvocationTargetException, ServerException {

        if (obj == null) { return null; }

        U dataObj = createNewDataObject(dataType);

        PersistenceManager pm = PMF.getManager();
        try {
            dataObj = pm.makePersistent(dataObj);
        } finally {
            pm.close();
        }
        if (!pm.isClosed()) { pm.close(); }

        Method[] methods = dataType.getMethods();
        for (Method m : methods) {

            SetterUserLevel ul = m.getAnnotation(SetterUserLevel.class);
            if (ul == null) { continue; }
            Role r = (cview == ClientView.PRIVATE) ? ul.privateView() : ul.publicView();
            if (!isAllowed(r, role)) { continue; }

            try {
                Method cm = getMethodByName(obj.getClass(), ul.mappedBy());
                Object value = cm.invoke(obj);
                if (ul.overrideBy().length() > 0) {
                    value = getDataMethodByName(dataType, ul.overrideBy()).invoke(dataObj, value);
                }
                m.invoke(dataObj, value);
            } catch (IllegalArgumentException e) {
                deleteObject(dataObj, e);
            } catch (IllegalAccessException e) {
                deleteObject(dataObj, e);
            } catch (InvocationTargetException e) {
                deleteObject(dataObj, e);
            } catch (NoSuchMethodException e) {
                deleteObject(dataObj, e);
            }

        }

        return dataObj;

    }

    private static <E extends Exception, U extends ModelDataClass<?,?,U>>
                    void deleteObject(U object, E exception)
                        throws E {

        PersistenceManager pm = PMF.getManager();
        try {
            pm.deletePersistent(object);
        } finally {
            pm.close();
        }
        throw exception;

    }

    /**
     * Creates a new instance of a ModelDataClass by converting a ModelClientClass
     * instance to its counter part ModelDataClass, using the given privilege information.
     * This function assumes all uniqueness in fields have been tested before hand.
     * @param obj The Client Model object to be converted.
     * @param dataType Type of the Data Model class
     * @param roleAndView The role and client view of the user requesting this conversion.
     * @return The converted Data Model instance, 'null' if either of 'clientObj' or 'emptyData' is 'null'.
     * @throws ServerException
     */
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
    U createDataCloneSafe(T obj, Class<U> dataType, RoleAndView roleAndView)
        throws ServerException {

        return createDataCloneSafe(obj, dataType, roleAndView.getRole(), roleAndView.getClientView());

    }

    /**
     * Creates a new instance of a ModelDataClass by converting a ModelClientClass
     * instance to its counter part ModelDataClass, using the given privilege information.
     * This function assumes all uniqueness in fields have been tested before hand.
     * @param obj The Client Model object to be converted.
     * @param dataType Type of the Data Model class
     * @param role The role of the user requesting this conversion.
     * @param cview The client view of the user requesting this conversion.
     * @return The converted Data Model instance, 'null' if either of 'clientObj' or 'emptyData' is 'null'.
     * @throws ServerException
     */
    public static <T extends ModelClientClass<?,T>, U extends ModelDataClass<?,T,U>>
        U createDataCloneSafe(T obj, Class<U> dataType, Role role, ClientView cview)
            throws ServerException {

        try {
            return createDataClone(obj, dataType, role, cview);
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (InstantiationException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (IllegalAccessException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        } catch (InvocationTargetException e) {
            e.printStackTrace();
            throw new ServerException(e.getMessage());
        }

    }

    /**
     * Create an empty instance of a ModelDataClass
     * @param <U> Type of data class
     * @param dataType Type
     * @return A newly generated instance
     * @throws InstantiationException
     * @throws IllegalAccessException
     */
    private static <U extends ModelDataClass<?,?,U>>
        U createNewDataObject(Class<U> dataType)
            throws InstantiationException, IllegalAccessException {

        U obj = dataType.newInstance();

        // operate on the new object if required.

        return obj;

    }

}
