/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.List;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.InstanceTokenError;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

public class UserAchievementGetter {

    /**
     * Gets the UserAchievement for the requested achievement of the user
     * currently playing the current game - READ ONLY
     * 
     * @param achievementName
     *            the name of the achievement requested
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @return The UserAchievement - READ ONLY
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws AchievementNotExists
     */
    public static UserAchievementData getUserAchievement(
            String achievementName, String gameToken) throws GameTokenError,
            AchievementNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUserAchievement(pm, achievementName, gameToken);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the UserAchievement for the requested achievement of the user
     * currently playing the current game, for editing object in datastore The
     * PersistenceManager must be handled by the caller
     * 
     * @param pm
     *            the Persistence Manager
     * @param achievementName
     *            the name of the achievement requested
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @return The UserAchievement
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws AchievementNotExists
     */
    @SuppressWarnings("unchecked")
    public static UserAchievementData getUserAchievement(PersistenceManager pm,
            String achievementName, String gameToken) throws GameTokenError,
            AchievementNotExists {

        UserGameProfileData ugp = UserGameProfileGetter
                .getCurrentUserGameProfile(pm, gameToken);
        AchievementData achievement = AchievementGetter.getAchievement(pm,
                achievementName, gameToken);

        Query q = pm.newQuery(UserAchievementData.class,
                "ugp == u && achievement == a");
        q.declareParameters("com.google.appengine.api.datastore.Key u, com.google.appengine.api.datastore.Key a");
        List<UserAchievementData> results = (List<UserAchievementData>) q
                .execute(ugp.getServerKey(), achievement.getServerKey());

        /*
         * If we can't find the UserAchievement it makes sense to create one if
         * the UserGameProfile and the Achievement exist.
         */
        if (results.isEmpty()) {
            UserAchievementData ua = new UserAchievementData();

            pm.makePersistent(ua);
            ua.setAchievement(achievement);
            ua.setUserGameProfile(ugp);

            return ua;
        }

        return results.get(0);
    }

    /**
     * Gets the UserAchievement for the requested achievement of the user
     * currently playing the current game, for editing object in datastore The
     * PersistenceManager must be handled by the caller
     * 
     * @param pm
     *            the Persistence Manager
     * @param achievementName
     *            the name of the achievement requested
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @return The UserAchievement
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws AchievementNotExists
     * @throws UserGameProfileNotExists
     */
    @SuppressWarnings("unchecked")
    public static UserAchievementData getUserAchievement(PersistenceManager pm,
            Long ugp, Long achievement) throws UserAchievementNotExists,
            AchievementNotExists, UserGameProfileNotExists {

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm,
                ugp);
        AchievementData ad = AchievementGetter.getAchievement(pm, achievement);

        Query q = pm.newQuery(UserAchievementData.class,
                "ugp == u && achievement == a");
        q.declareParameters("com.google.appengine.api.datastore.Key u, com.google.appengine.api.datastore.Key a");
        List<UserAchievementData> results = (List<UserAchievementData>) q
                .execute(ugpd.getServerKey(), ad.getServerKey());

        if (results.isEmpty()) {
            throw new UserAchievementNotExists();
        }

        return results.get(0);

    }

    /**
     * Gets the UserAchievement associated with the given UserAchievementData
     * primary key.
     * 
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The UserAchievementData object directly off the datastore.
     * @throws UserAchievementNotExists
     */
    public static UserAchievementData getUserAchievement(Long key)
            throws UserAchievementNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUserAchievement(pm, key);
        } finally {
            pm.close();
        }

    }

    /**
     * Gets the UserAchievement associated with the given UserAchievementData
     * primary key.
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The UserAchievementData object directly off the datastore.
     * @throws UserAchievementNotExists
     */
    public static UserAchievementData getUserAchievement(PersistenceManager pm,
            Long key) throws UserAchievementNotExists {
        try {
            UserAchievementData ua = pm.getObjectById(
                    UserAchievementData.class, key);
            return ua;
        } catch (JDOObjectNotFoundException e) {
            throw new UserAchievementNotExists(key);
        }
    }

    /**
     * Gets the UserAchievement associated with the given UserAchievementData
     * primary key.
     * 
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The UserAchievementData object directly off the datastore.
     * @throws UserAchievementNotExists
     */
    public static UserAchievementData getUserAchievement(Key key)
            throws UserAchievementNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUserAchievement(pm, key);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the UserAchievement associated with the given UserAchievementData
     * primary key.
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The UserAchievementData object directly off the datastore.
     * @throws UserAchievementNotExists
     */
    public static UserAchievementData getUserAchievement(PersistenceManager pm,
            Key key) throws UserAchievementNotExists {
        try {
            UserAchievementData ua = pm.getObjectById(
                    UserAchievementData.class, key);
            return ua;
        } catch (JDOObjectNotFoundException e) {
            throw new UserAchievementNotExists(key.getId());
        }
    }

    /**
     * Gets the UserAchievement for the requested achievement of the user
     * currently playing the current game - READ ONLY
     * 
     * @param achievementName
     *            the name of the achievement requested
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @param instanceToken
     *            the secret token for the current game instance, provided by
     *            the client
     * @return The UserAchievement - READ ONLY
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws AchievementNotExists
     * @throws InstanceTokenError
     */
    public static UserAchievementData getUserInstanceAchievement(
            String achievementName, String gameToken, String instanceToken)
            throws GameTokenError, AchievementNotExists, InstanceTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUserInstanceAchievement(pm, achievementName, gameToken,
                    instanceToken);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the UserAchievement for the requested achievement of the user
     * currently playing the current game, for editing object in datastore The
     * PersistenceManager must be handled by the caller
     * 
     * @param pm
     *            the Persistence Manager
     * @param achievementName
     *            the name of the achievement requested
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @param instanceToken
     *            the secret token for the current game instance, provided by
     *            the client
     * @return The UserAchievement
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws AchievementNotExists
     * @throws InstanceTokenError
     */
    @SuppressWarnings("unchecked")
    public static UserAchievementData getUserInstanceAchievement(
            PersistenceManager pm, String achievementName, String gameToken,
            String instanceToken) throws GameTokenError, AchievementNotExists,
            InstanceTokenError {

        UserGameProfileData ugp = UserGameProfileGetter
                .getCurrentUserGameInstanceProfile(pm, instanceToken);
        AchievementData achievement = AchievementGetter.getAchievement(pm,
                achievementName, gameToken);

        Query q = pm.newQuery(UserAchievementData.class,
                "ugp == u && achievement == a");
        q.declareParameters("com.google.appengine.api.datastore.Key u, com.google.appengine.api.datastore.Key a");
        List<UserAchievementData> results = (List<UserAchievementData>) q
                .execute(ugp.getServerKey(), achievement.getServerKey());

        /*
         * If we can't find the UserAchievement it makes sense to create one if
         * the UserGameProfile and the Achievement exist.
         */
        if (results.isEmpty()) {
            UserAchievementData ua = new UserAchievementData();

            pm.makePersistent(ua);
            ua.setAchievement(achievement);
            ua.setUserGameProfile(ugp);

            return ua;
        }

        return results.get(0);
    }

    /**
     * Gets the UserAchievement for the requested achievement of the user
     * currently playing the current game - READ ONLY
     * 
     * @param achievementName
     *            the name of the achievement requested
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @param instanceToken
     *            the secret game instance token for the current game instance
     * @return The UserAchievement - READ ONLY
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws AchievementNotExists
     * @throws InstanceTokenError
     */
    public static UserAchievementData getInstanceUserAchievement(
            String achievementName, String gameToken, String instanceToken)
            throws GameTokenError, AchievementNotExists, InstanceTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            return getInstanceUserAchievement(pm, achievementName, gameToken,
                    instanceToken);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the UserAchievement for the requested achievement of the user
     * currently playing the current game, for editing object in datastore The
     * PersistenceManager must be handled by the caller
     * 
     * @param pm
     *            the Persistence Manager
     * @param achievementName
     *            the name of the achievement requested
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @param instanceToken
     *            the secret game instance token for the current game instance
     * @return The UserAchievement
     * @throws UserNotExists
     * @throws GameTokenError
     * @throws AchievementNotExists
     * @throws InstanceTokenError
     */
    @SuppressWarnings("unchecked")
    public static UserAchievementData getInstanceUserAchievement(
            PersistenceManager pm, String achievementName, String gameToken,
            String instanceToken) throws GameTokenError, AchievementNotExists,
            InstanceTokenError {

        UserGameProfileData ugp = UserGameProfileGetter
                .getCurrentUserGameInstanceProfile(pm, instanceToken);
        AchievementData achievement = AchievementGetter.getAchievement(pm,
                achievementName, gameToken);

        Query q = pm.newQuery(UserAchievementData.class,
                "ugp == u && achievement == a");
        q.declareParameters("com.google.appengine.api.datastore.Key u, com.google.appengine.api.datastore.Key a");
        List<UserAchievementData> results = (List<UserAchievementData>) q
                .execute(ugp.getServerKey(), achievement.getServerKey());

        /*
         * If we can't find the UserAchievement it makes sense to create one if
         * the UserGameProfile and the Achievement exist.
         */
        if (results.isEmpty()) {
            UserAchievementData ua = new UserAchievementData();

            pm.makePersistent(ua);
            ua.setAchievement(achievement);
            ua.setUserGameProfile(ugp);

            return ua;
        }

        return results.get(0);
    }
}
