/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.jdo.annotations.IdGeneratorStrategy;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;
import javax.jdo.annotations.PrimaryKey;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.server.model.annotations.*;

@MugleData
@PersistenceCapable
public class UserData extends ModelDataClass<Long, User, UserData> {


    // Fields

    @PrimaryKey
    @Persistent(valueStrategy = IdGeneratorStrategy.IDENTITY)
    private Key id;

    @Persistent
    private String googleID; //Unique userID provided by google

    @Persistent
    private String urlName; //Must be unique

    @Persistent
    private String fullName;

    @Persistent
    private String email;

    @Persistent
    private Role role;

    @Persistent
    private Boolean active;

    // Avoiding multiple parents
    @Persistent
    private Set<Key> ugps;

    // Many-to-many
    @Persistent
    private Set<Key> devTeams;

    @Persistent
    private Key createdUser;

    @Persistent
    private Date createdTime;

    @Persistent
    private Key updatedUser;

    @Persistent
    private Date updatedTime;

    // Constructors

    public UserData() {

        this.devTeams = new HashSet<Key>();
        this.ugps = new HashSet<Key>();
        this.googleID = null;
        this.urlName = null;
        this.fullName = null;
        this.email = null;
        this.role = null;
        this.active = true;

    }

    public UserData(String urlName, String fullName, String email, Role role) {
        this.devTeams = new HashSet<Key>();
        this.ugps = new HashSet<Key>();
        this.googleID = null;
        this.urlName = urlName;
        this.fullName = fullName;
        this.email = email;
        this.role = role;
        this.active = true;
    }

    // Getters

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setPrimaryKey")
    public Long getPrimaryKey() {
        if (this.id == null) {
            return null;
        }
        return this.id.getId();
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUrlName")
    public String getUrlName() {
        return this.urlName;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedTo="setGoogleID")
    public String getGoogleID() {
        return googleID;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setFullName")
    public String getFullName() {
        return this.fullName;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedTo="setEmail")
    public String getEmail() {
        return this.email;
    }

    @GetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedTo="setActive")
    public Boolean isActive() {
        return active;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setRole")
    public Role getRole() {
        return this.role;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, list=true)
    public Set<Key> getDevTeams() {
        return this.devTeams;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, list=true)
    public Set<Key> getUserGameProfiles() {
        return this.ugps;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedUser", overrideBy="keyToCreated")
    public Key getCreatedUser() {
        return this.createdUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedTime")
    public Date getCreatedTime() {
        return this.createdTime;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedUser", overrideBy="keyToUpdated")
    public Key getUpdatedUser() {
        return this.updatedUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedTime")
    public Date getUpdatedTime() {
        return this.updatedTime;
    }


    // Setters

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedUser(Key createdUser) {
        this.createdUser = createdUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedTime(Date createdTime) {
        this.createdTime = createdTime;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedUser(Key updatedUser) {
        this.updatedUser = updatedUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedTime(Date updatedTime) {
        this.updatedTime = updatedTime;
    }

    /* Users need to be able to change their own URL name for privacy reasons.
     * (Otherwise it defaults to their email address, which shouldn't be
     * exposed.)
     */
    @SetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedBy="getUrlName")
    public void setUrlName(String newName) {
        //TODO: check that newName is unique.
        this.urlName = newName;
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getGoogleID")
    public void setGoogleID(String googleID) {
        this.googleID = googleID;
    }

    /* This can only be changed by an Admin, again we need an Admin control panel */
    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getRole")
    public void setRole(Role newRole) {
        this.role = newRole;
    }

    @SetterUserLevel(privateView=Role.GUEST, publicView=Role.ADMIN, mappedBy="getFullName")
    public void setFullName(String newName) {
        this.fullName = newName;
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getEmail")
    public void setEmail(String newEmail) {
        //TODO: check that email is unique and valid
        this.email = newEmail;
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="isActive")
    public void setActive(Boolean active) {
        this.active = active;
    }
        // Misc methods
    public void addUserGameProfile(UserGameProfileData ugp) {
        ugp.setUser(this);
    }

    public void removeUserGameProfile(UserGameProfileData ugp) {
        this.ugps.remove(ugp.getServerKey());
    }

    public void addDevTeam(DevTeamData devteam)
    {
        // many-to-many relationship
        this.devTeams.add(devteam.getServerKey());
        devteam.getUsers().add(this.id);
    }

    public void removeDevTeam(DevTeamData devteam)
    {
        // many-to-many relationship
        this.devTeams.remove(devteam.getServerKey());
        devteam.getUsers().remove(this.id);
    }

    @Override
    public Key getServerKey() {
        return this.id;
    }

    @Override
    public void setServerKey(Key key) {
        this.id = key;
    }

    @Override
    public void ClearFields() {
        this.setActive(null);
        this.setEmail(null);
        this.setFullName(null);
        this.setRole(null);
        this.setUrlName(null);
        this.setServerKey(null);
    }

    @Override
    public Class<User> getClientClassType() {
        return User.class;
    }

    public Long keyToCreated() {
        if (this.createdUser == null) { return null; }
        return this.createdUser.getId();
    }

    public Long keyToUpdated() {
        if (this.updatedUser == null) { return null; }
        return this.updatedUser.getId();
    }

}
