/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.jdo.PersistenceManager;
import javax.jdo.annotations.IdGeneratorStrategy;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;
import javax.jdo.annotations.PrimaryKey;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.annotations.*;

@MugleData
@PersistenceCapable
public class UserGameProfileData extends ModelDataClass<Long, UserGameProfile, UserGameProfileData> {


    // Fields

    @PrimaryKey
    @Persistent(valueStrategy = IdGeneratorStrategy.IDENTITY)
    private Key id;

    @Persistent
    private Integer highscore;

    @Persistent
    private Key user;

    @Persistent
    private Key game;
    
    @Persistent
    private Key gameInstance;

    @Persistent(mappedBy="ugp")
    private Set<Key> userAchievements;

    @Persistent(mappedBy="ugp")
    private Set<Key> keyValuePairs;

    @Persistent
    private Key createdUser;

    @Persistent
    private Date createdTime;

    @Persistent
    private Key updatedUser;

    @Persistent
    private Date updatedTime;


    // Constructors

    public UserGameProfileData() {

        this.userAchievements = new HashSet<Key>();
        this.keyValuePairs = new HashSet<Key>();
        this.highscore = 0;
        this.user = null;
        this.game = null;

    }

    // Getters

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setPrimaryKey")
    public Long getPrimaryKey() {
        if (this.id == null) {
            return null;
        }
        return this.id.getId();
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUser", overrideBy="keyToUser")
    public Key getUser() {
        return this.user;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setGame", overrideBy="keyToGame")
    public Key getGame() {
        return this.game;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setHighscore")
    public Integer getHighscore() {
        return this.highscore;
    }

    @GetterUserLevel(privateView=Role.NONE, publicView=Role.NONE, list=true)
    public Set<Key> getKeyValuePairs() {
        return this.keyValuePairs;
    }

    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, list=true)
    public Set<Key> getUserAchievements() {
        return this.userAchievements;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedUser", overrideBy="keyToCreated")
    public Key getCreatedUser() {
        return this.createdUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setCreatedTime")
    public Date getCreatedTime() {
        return this.createdTime;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedUser", overrideBy="keyToUpdated")
    public Key getUpdatedUser() {
        return this.updatedUser;
    }

    @Override
    @GetterUserLevel(privateView=Role.GUEST, publicView=Role.GUEST, mappedTo="setUpdatedTime")
    public Date getUpdatedTime() {
        return this.updatedTime;
    }


    // Setters

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedUser(Key createdUser) {
        this.createdUser = createdUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setCreatedTime(Date createdTime) {
        this.createdTime = createdTime;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedUser(Key updatedUser) {
        this.updatedUser = updatedUser;
    }

    @Override
    @SetterUserLevel(privateView=Role.NONE, publicView=Role.NONE)
    public void setUpdatedTime(Date updatedTime) {
        this.updatedTime = updatedTime;
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getGame", overrideBy="gameFromKey")
    public void setGame(GameData game) {
        if (game == null) {
            this.game = null;
            return;
        }
        this.game = game.getServerKey();
        game.getUserGameProfiles().add(this.id);
    }

    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getUser", overrideBy="userFromKey")
    public void setUser(UserData user) {
        if (user == null) {
            this.user = null;
            return;
        }
        this.user = user.getServerKey();
        user.getUserGameProfiles().add(this.id);
    }

    /* can only be set by the current Game */
    @SetterUserLevel(privateView=Role.ADMIN, publicView=Role.ADMIN, mappedBy="getHighscore")
    public void setHighscore(Integer newScore) {
        this.highscore = newScore;
    }

    // Misc Methods

    public void addKeyValue(KeyValuePairData kvp)
    {
        kvp.setUserGameProfile(this);
    }

    public void removeKeyValue(KeyValuePairData kvp)
    {
        this.keyValuePairs.remove(kvp.getServerKey());
    }

    public void addUserAchievement(UserAchievementData ua)
    {
        ua.setUserGameProfile(this);
    }

    public void removeUserAchievement(UserAchievementData ua)
    {
        this.userAchievements.remove(ua.getServerKey());
    }

    @Override
    public Key getServerKey() {
        return this.id;
    }

    @Override
    public void setServerKey(Key key) {
        this.id = key;
    }

    @Override
    public void ClearFields() {
        this.setGame(null);
        this.setHighscore(null);
        this.setServerKey(null);
        this.setUser(null);
    }

    @Override
    public Class<UserGameProfile> getClientClassType() {
        return UserGameProfile.class;
    }

    public Long keyToUser() {
        if (this.user == null) {
            return null;
        }
        return this.user.getId();
    }

    public Long keyToGame() {
        if (this.game == null) {
            return null;
        }
        return this.game.getId();
    }

    public GameData gameFromKey(Long key) {
        if (key == null) { return null; }
        PersistenceManager pm = PMF.getManager();
        GameData result = null;
        try {
            result = pm.getObjectById(GameData.class, key);
        } finally {
            pm.close();
        }
        return result;
    }

    public UserData userFromKey(Long key) {
        if (key == null) { return null; }
        PersistenceManager pm = PMF.getManager();
        UserData result = null;
        try {
            result = pm.getObjectById(UserData.class, key);
        } finally {
            pm.close();
        }
        return result;
    }

    public Long keyToCreated() {
        if (this.createdUser == null) { return null; }
        return this.createdUser.getId();
    }

    public Long keyToUpdated() {
        if (this.updatedUser == null) { return null; }
        return this.updatedUser.getId();
    }

    public Key getGameInstance() {
        return gameInstance;
    }

    public void setGameInstance(Key gameInstance) {
        this.gameInstance = gameInstance;
    }

}

