/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import java.util.List;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.GameInstanceData;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.GameInstanceGetter;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.InstanceTokenError;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

public class UserGameProfileGetter {

    /**
     * Gets the UserGameProfile by its Primary Key - READ ONLY
     * 
     * @param primaryKey
     * @return The UserGameProfile - READ ONLY
     * @throws UserGameProfileNotExists
     */
    public static UserGameProfileData getUserGameProfile(Long primaryKey)
            throws UserGameProfileNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUserGameProfile(pm, primaryKey);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the UserGameProfile by its Primary Key, for editing the object in
     * the datastore PersistenceManager must be handled by the caller
     * 
     * @param pm
     *            The PersistenceManager
     * @param primaryKey
     * @return The UserGameProfile
     * @throws UserGameProfileNotExists
     */
    public static UserGameProfileData getUserGameProfile(PersistenceManager pm,
            Long primaryKey) throws UserGameProfileNotExists {
        try {
            UserGameProfileData ugp = pm.getObjectById(
                    UserGameProfileData.class, primaryKey);
            return ugp;
        } catch (JDOObjectNotFoundException e) {
            throw new UserGameProfileNotExists(primaryKey);
        }
    }

    /**
     * Gets the UserGameProfile of the current user based on the game they're
     * playing (read only)
     * 
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @return the UserGameProfile - READ ONLY
     * @throws UserNotExists
     * @throws GameTokenError
     */
    public static UserGameProfileData getCurrentUserGameProfile(String gameToken)
            throws GameTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            return getCurrentUserGameProfile(pm, gameToken);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the UserGameProfile of the current user based on the game they're
     * playing, for editing object in datstore The Persistence Manager must be
     * handled by the caller
     * 
     * @param pm
     *            The Persistence Manager
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @return The UserGameProfile of the current user / game.
     * @throws UserNotExists
     * @throws GameTokenError
     */
    @SuppressWarnings("unchecked")
    public static UserGameProfileData getCurrentUserGameProfile(
            PersistenceManager pm, String gameToken) throws GameTokenError {

        UserData curUser = UserGetter.getCurrentUser();
        GameData curGame = GameGetter.getCurrentGame(gameToken);

        Query q = pm.newQuery(UserGameProfileData.class,
                "user == u && game == g");
        q.declareParameters("com.google.appengine.api.datastore.Key u, com.google.appengine.api.datastore.Key g");
        List<UserGameProfileData> results = (List<UserGameProfileData>) q
                .execute(curUser.getServerKey(), curGame.getServerKey());

        /*
         * If we can't find the UserGameProfile it makes sense to create one if
         * the user is playing the game
         */
        if (results.isEmpty()) {
            UserGameProfileData ugp = new UserGameProfileData();
            pm.makePersistent(ugp);
            ugp.setGame(curGame);
            ugp.setUser(curUser);

            return ugp;
        }
        return results.get(0);

    }

    /**
     * Gets the UserGameProfile associated with the given UserGameProfileData
     * primary key.
     * 
     * @param pm
     *            The PersistenceManager
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The UserData object directly off the datastore.
     * @throws UserGameProfileNotExists
     */
    public static UserGameProfileData getUserGameProfile(PersistenceManager pm,
            Key key) throws UserGameProfileNotExists {
        try {
            UserGameProfileData ugp = pm.getObjectById(
                    UserGameProfileData.class, key);
            return ugp;
        } catch (JDOObjectNotFoundException e) {
            throw new UserGameProfileNotExists(key.getId());
        }
    }

    /**
     * Gets the UserGameProfile associated with the given UserGameProfileData
     * primary key.
     * 
     * @param key
     *            The server-side key (ModelDataClass.getServerKey())
     * @return The UserData object directly off the datastore.
     * @throws UserGameProfileNotExists
     */
    public static UserGameProfileData getUserGameProfile(Key key)
            throws UserGameProfileNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUserGameProfile(pm, key);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the UserGameProfile of the current user based on the game they're
     * playing, for editing object in datstore The Persistence Manager must be
     * handled by the caller
     * 
     * @param pm
     *            The Persistence Manager
     * @param gameToken
     *            the secret game token for the current game, provided by the
     *            client
     * @return The UserGameProfile of the current user / game.
     * @throws UserNotExists
     * @throws GameTokenError
     */
    @SuppressWarnings("unchecked")
    public static UserGameProfileData getUserGameProfile(PersistenceManager pm,
            UserData user, GameData game) {

        Query q = pm.newQuery(UserGameProfileData.class,
                "user == u && game == g");
        q.declareParameters("com.google.appengine.api.datastore.Key u, com.google.appengine.api.datastore.Key g");
        List<UserGameProfileData> results = (List<UserGameProfileData>) q
                .execute(user.getServerKey(), game.getServerKey());

        if (results.isEmpty()) {
            UserGameProfileData ugp = new UserGameProfileData();
            pm.makePersistent(ugp);
            ugp.setGame(game);
            ugp.setUser(user);

            return ugp;
        }
        return results.get(0);

    }

    public static UserGameProfileData getCurrentUserGameInstanceProfile(
            String instanceToken) throws InstanceTokenError {

        PersistenceManager pm = PMF.getManager();
        try {
            return getCurrentUserGameInstanceProfile(pm, instanceToken);
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    public static UserGameProfileData getCurrentUserGameInstanceProfile(
            PersistenceManager pm, String instanceToken)
            throws InstanceTokenError {
        UserData ud = UserGetter.getCurrentUser();
        GameInstanceData gi = GameInstanceGetter
                .getInstanceByToken(instanceToken);

        Query q = pm.newQuery(UserGameProfileData.class,
                "user == u && gameInstance == gi");
        q.declareParameters("com.google.appengine.api.datastore.Key u, com.google.appengine.api.datastore.Key gi");
        List<UserGameProfileData> results = (List<UserGameProfileData>) q
                .execute(ud.getServerKey(), gi.getServerKey());

        if (results.isEmpty()) {
            UserGameProfileData ugp = new UserGameProfileData();
            pm.makePersistent(ugp);
            ugp.setGameInstance(gi.getServerKey());
            ugp.setUser(ud);

            return ugp;
        }
        return results.get(0);
    }

    public static UserGameProfileData getUserGameInstanceProfile(UserData user,
            GameInstanceData gameInstance) {

        PersistenceManager pm = PMF.getManager();
        try {
            return getUserGameInstanceProfile(user, gameInstance);
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    public static UserGameProfileData getUserGameInstanceProfile(
            PersistenceManager pm, UserData user, GameInstanceData gameInstance) {

        Query q = pm.newQuery(UserGameProfileData.class,
                "user == u && gameInstance == gi");
        q.declareParameters("com.google.appengine.api.datastore.Key u, com.google.appengine.api.datastore.Key gi");
        List<UserGameProfileData> results = (List<UserGameProfileData>) q
                .execute(user.getServerKey(), gameInstance.getServerKey());

        if (results.isEmpty()) {
            UserGameProfileData ugp = new UserGameProfileData();
            pm.makePersistent(ugp);
            ugp.setGameInstance(gameInstance.getServerKey());
            ugp.setUser(user);

            return ugp;
        }
        return results.get(0);

    }

}
