/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.model;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;

import java.util.List;

import au.edu.unimelb.csse.mugle.client.LoginInfo;
import au.edu.unimelb.csse.mugle.server.ClientServiceImpl;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.shared.model.Role;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

public class UserGetter {

    private static UserData emptyUser = new UserData();
    
    /**
     * Gets a User based on the Primary Key
     * @param primaryKey
     * @return The User class - READ ONLY
     * @throws UserNotExists
     */
    public static UserData getUser(Long primaryKey) throws UserNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUser(pm, primaryKey);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets a User based on their Primary Key, for editing object in datstore
     * Persistence Manager must be handle by caller
     * @param pm The Persistence Manager
     * @param primaryKey
     * @return The User class
     * @throws UserNotExists
     */
    public static UserData getUser(PersistenceManager pm, Long primaryKey) throws UserNotExists {
        try {
            UserData u = pm.getObjectById(UserData.class, primaryKey);
            return u;
        } catch (JDOObjectNotFoundException e) {
            throw new UserNotExists(primaryKey);
        }
    }
    /**
     * Gets a User based on their urlName - READ ONLY
     * @param name
     * @return The User class - READ ONLY
     * @throws UserNotExists
     */
    public static UserData getUserByName(String name) throws UserNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUserByName(pm, name);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets a User based on their urlName, for editing object in datstore
     * Persistence Manager must be handle by caller
     * @param pm The Persistence Manager
     * @param name
     * @return The User class
     * @throws UserNotExists
     */
    @SuppressWarnings("unchecked")
    public static UserData getUserByName(PersistenceManager pm, String name) throws UserNotExists {
        Query q = pm.newQuery(UserData.class, "urlName == u");
        q.declareParameters("String u");
        List<UserData> results = (List<UserData>) q.execute(name);
                
        if (results.isEmpty()) {
            throw new UserNotExists(name);
        }
        
        return results.get(0);
    }
    
    /**
     * Gets a User based on their unique google ID - READ ONLY
     * @param googleID
     * @return The User class - READ ONLY
     * @throws UserNotExists
     */
    public static UserData getUser(String googleID) throws UserNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUser(pm, googleID);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets a User based on their unique google ID, for editing object in datstore
     * Persistence Manager must be handle by caller
     * @param pm The Persistence Manager
     * @param googleID
     * @return The User class
     * @throws UserNotExists
     */
    @SuppressWarnings("unchecked")
    public static UserData getUser(PersistenceManager pm, String googleID) throws UserNotExists {
        Query q = pm.newQuery(UserData.class, "googleID == u");
        q.declareParameters("String u");
        List<UserData> results = (List<UserData>) q.execute(googleID);

        if (results.isEmpty()) {
            throw new UserNotExists("googleID: " + googleID);
        }

        return results.get(0);
    }
    
    /**
     * Gets the currently logged in User - READ ONLY
     * @return the User - read only
     */
    public static UserData getCurrentUser() {
        PersistenceManager pm = PMF.getManager();
        try {
            return getCurrentUser(pm);
        } finally {
            pm.close();
        }
    }
    
    /**
     * Gets the currently logged in User, for editing object in datastore
     * Creates a new user if the currently logged in user is not found
     * Persistence Manager must be handled by the caller
     * @param pm The Persistence Manager
     * @return the current User
     */
    public static UserData getCurrentUser(PersistenceManager pm) {
        ClientServiceImpl loginService = new ClientServiceImpl();
        LoginInfo loginInfo = loginService.getCurrentUser();
        try {
            return getUser(pm, loginInfo.getUserID());
        } catch (UserNotExists e) {
            UserData newUser = new UserData();
            
            //Creates the new user using the info provided by google
            newUser.setGoogleID(loginInfo.getUserID());
            newUser.setUrlName(loginInfo.getNickname());
            newUser.setFullName(loginInfo.getNickname());
            newUser.setEmail(loginInfo.getEmailAddress());
            if (loginInfo.isAdmin()) {
                newUser.setRole(Role.ADMIN);
            } else {
                newUser.setRole(Role.GUEST);
            }
            newUser.setActive(true);
            
            //Stores the new user in the datastore
            pm.makePersistent(newUser);
            
            return newUser;
        }
        
    }
    
    /**
     * Gets the currently logged in User.
     * This should only be used for cases other than when logging in.
     * This methods does not create a new user if the user does not exist.
     * @param pm The Persistence Manager
     * @return the current User; if failed, returns and empty user object
     */
    public static UserData getCurrentUserQuick(PersistenceManager pm) {
        ClientServiceImpl loginService = new ClientServiceImpl();
        LoginInfo loginInfo = loginService.getCurrentUser();
        try {
            return getUser(pm, loginInfo.getUserID());
        } catch (UserNotExists e) {
            return emptyUser;
        }
    }
    
    /**
     * Gets the currently logged in User.
     * This should only be used for cases other than when logging in.
     * This methods does not create a new user if the user does not exist.
     * @return the current User; if failed, returns and empty user object
     */
    public static UserData getCurrentUserQuick() {
        PersistenceManager pm = PMF.getManager();
        try {
            return getCurrentUserQuick(pm);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the User associated with the given UserData primary key.
     * @param pm The PersistenceManager
     * @param key The server-side key (ModelDataClass.getServerKey())
     * @return The UserData object directly off the datastore.
     * @throws UserNotExists
     */
    public static UserData getUser(Key key)
            throws UserNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            return getUser(pm, key);
        } finally {
            pm.close();
        }
    }

    /**
     * Gets the User associated with the given UserData primary key.
     * @param pm The PersistenceManager
     * @param key The server-side key (ModelDataClass.getServerKey())
     * @return The UserData object directly off the datastore.
     * @throws UserNotExists 
     */
    public static UserData getUser(PersistenceManager pm, Key key)
            throws UserNotExists {
        try {
            UserData u = pm.getObjectById(UserData.class, key);
            return u;
        } catch (JDOObjectNotFoundException e) {
            throw new UserNotExists(key.getId());
        }
    }
}
