/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;
import java.util.HashSet;

import javax.jdo.PersistenceManager;

import com.google.appengine.api.datastore.Key;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class AchievementServiceImpl extends RemoteServiceServlet implements AchievementService {

    @Override
    public Achievement getByKey(Long key)
            throws ServerException, AchievementNotExists {

        AchievementData ad = AchievementGetter.getAchievement(key);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(ad, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public Achievement getByName(String name, Game game)
                throws ServerException, AchievementNotExists, GameTokenError {

        PersistenceManager pm = PMF.getManager();

        try {
            GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
            AchievementData ad = AchievementGetter.getAchievement(pm, name, gd.getGameToken());
            UserData curUser = UserGetter.getCurrentUser(pm);
            return ModelWrapper.getClientCloneSafe(ad, curUser.getRole(), ClientView.PRIVATE);
        } catch (GameNotExists e) {
            e.printStackTrace();
            throw new AchievementNotExists(game.getUrlName(), name);
        } finally {
            pm.close();
        }


    }

    @Override
    public Achievement getByName(String name, String gameToken)
                throws ServerException, AchievementNotExists, GameTokenError {

        AchievementData ad = AchievementGetter.getAchievement(name, gameToken);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(ad, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public Achievement create(Achievement object)
                throws ServerException, UserPrivilegeException, AchievementExists, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);
        GameData gd = null;
        // check for uniqueness
        try {
            gd = GameGetter.getGame(pm, object.getGameKey());
            if (AchievementGetter.getAchievement(pm, object.getName(), gd.getGameToken()) != null) {
                throw new AchievementExists();
            }
        } catch (AchievementNotExists e) {
            // this is excepted
        } catch (GameTokenError r) {
            // Game is gotten from server side so this is impossible
            throw new ServerException("Server Error on token creation: " + r.getMessage());
        }

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        AchievementData data = ModelWrapper.createDataCloneSafe(object, AchievementData.class, rv);
        gd.addAchievement(data); //need to create the other direction of the relationship

        // update the database
        try {
            Date timestamp = new Date();
            data.logCreated(rv.getUser().getServerKey(), timestamp);
            data.logUpdated(rv.getUser().getServerKey(), timestamp);
            data = pm.makePersistent(data);
        } finally {
            pm.close();
        }

        // Convert the updated object back to
        return ModelWrapper.getClientCloneSafe(data, rv);

    }

    @Override
    public Achievement write(Achievement object)
                throws ServerException, UserPrivilegeException, AchievementNotExists, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, AchievementData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            //throw new AchievementNotExists(object);
            throw new AchievementNotExists(object.getPrimaryKey());
        }

    }

    @Override
    public ModelCollection<UserAchievement> getUserAchievements(Achievement achievement)
            throws ServerException, AchievementNotExists {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PRIVATE);
        AchievementData ad = AchievementGetter.getAchievement(pm, achievement.getPrimaryKey());

        try {

            HashSet<UserAchievement> list = new HashSet<UserAchievement>(ad.getUserAchievements().size());
            for (Key k : ad.getUserAchievements()) {
                UserAchievementData ua;
                try {
                    ua = null;
                    ua = UserAchievementGetter.getUserAchievement(k.getId());
                    UserAchievement wrapped = ModelWrapper.getClientCloneSafe(ua, rv);
                    list.add(wrapped);
                } catch (UserAchievementNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<UserAchievement> toRet = new ModelCollection<UserAchievement>();
            toRet.setCollection(list);
            toRet.setCapacity(list.size());
            toRet.setStartIndex(0);
            toRet.setEndIndex(list.size());
            return toRet;

        } finally {
            pm.close();
        }

    }

    @Override
    public UpdatedPair<Achievement, UserAchievement>
                addUserAchievement(Achievement achievement, UserAchievement userAchievement)
                    throws ServerException, AchievementNotExists, UserAchievementNotExists, UserPrivilegeException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(achievement, pm);

        AchievementData ad = AchievementGetter.getAchievement(pm, achievement.getPrimaryKey());
        UserAchievementData uad = UserAchievementGetter.getUserAchievement(pm, userAchievement.getPrimaryKey());

        // do action
        ad.addUserAchievement(uad);
        Date timestamp = new Date();
        ad.logUpdated(rv.getUser().getServerKey(), timestamp);
        uad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ad, uad };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        achievement = ModelWrapper.getClientCloneSafe(ad, rv);
        userAchievement = ModelWrapper.getClientCloneSafe(uad, rv);

        return new UpdatedPair<Achievement, UserAchievement>(achievement, userAchievement);

    }

    @Override
    public UpdatedPair<Achievement, UserAchievement>
                removeUserAchievement(Achievement achievement, UserAchievement userAchievement)
                    throws ServerException, AchievementNotExists, UserAchievementNotExists, UserPrivilegeException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(achievement, pm);

        AchievementData ad = AchievementGetter.getAchievement(pm, achievement.getPrimaryKey());
        UserAchievementData uad = UserAchievementGetter.getUserAchievement(pm, userAchievement.getPrimaryKey());

        // do action
        ad.removeUserAchievement(uad);
        Date timestamp = new Date();
        ad.logUpdated(rv.getUser().getServerKey(), timestamp);
        uad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ad, uad };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        achievement = ModelWrapper.getClientCloneSafe(ad, rv);
        userAchievement = ModelWrapper.getClientCloneSafe(uad, rv);

        return new UpdatedPair<Achievement, UserAchievement>(achievement, userAchievement);

    }

    @Override
    public void remove(Achievement object) throws ServerException,
            UserPrivilegeException, AchievementNotExists, GameNotExists {


        PersistenceManager pm = PMF.getManager();
        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            DataServiceHelper.remove(object, AchievementData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            //throw new AchievementNotExists(object);
            throw new AchievementNotExists(object.getPrimaryKey());
        }

    }

    @Override
    public void remove(ModelCollection<Achievement> achievements)
            throws ServerException, UserPrivilegeException,
            AchievementNotExists, GameNotExists {
        for (Achievement a : achievements.getCollection()) {
            remove(a);
        }
    }

    private RoleAndView checkPermissions(Achievement object, PersistenceManager pm) throws UserPrivilegeException, GameNotExists {
        GameData gd = GameGetter.getGame(object.getGameKey());
        // get currently logged in user whether they are part of the devTeam
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.equals(gd.getDevTeam())) {
                isDev = true;
            }
        }
        // Only an admin or person in the DevTeam can create the game
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN && !isDev)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified Badge object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = isDev ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }

}
