/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;
import java.util.HashSet;
import java.util.List;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class DevTeamServiceImpl extends RemoteServiceServlet implements DevTeamService {

    @Override
    public DevTeam getByKey(Long key)
            throws ServerException, DevTeamNotExists {

        DevTeamData dt = DevTeamGetter.getDevTeam(key);
        UserData curUser = UserGetter.getCurrentUser();
        ClientView cv = getClientView(dt);
        return ModelWrapper.getClientCloneSafe(dt, curUser.getRole(), cv);

    }

    @Override
    public DevTeam getByName(String name)
            throws ServerException, DevTeamNotExists {

        DevTeamData dt = DevTeamGetter.getDevTeam(name);
        UserData curUser = UserGetter.getCurrentUser();
        ClientView cv = getClientView(dt);
        return ModelWrapper.getClientCloneSafe(dt, curUser.getRole(), cv);

    }

    @Override
    public DevTeam create(DevTeam object)
            throws ServerException, UserPrivilegeException, DevTeamExists {

        PersistenceManager pm = PMF.getManager();
        try {
            // get currently logged in user and ownership
            UserData curUser = UserGetter.getCurrentUser(pm);

            // Only an admin can create a dev team
            if (curUser.getRole() != Role.ADMIN) {
                String msg = "The current user does not have the the right privilege to"
                                    + " create the specified DevTeam object.";
                throw new UserPrivilegeException(msg);
            }

            // check for uniqueness
            try {
                if (DevTeamGetter.getDevTeam(pm, object.getUrlName()) != null) {
                    throw new DevTeamExists(object);
                }
            } catch (DevTeamNotExists e) {
                // this is excepted
            }

            // figure out the Role and ClientView
            // ClientView.PUBLIC (by definition, nobody is in the devteam yet)
            RoleAndView rv = new RoleAndView(curUser.getRole(),
                                             ClientView.PUBLIC);

            // gets the server-side objects with its properties updated
            // from the client-side object. However, this only happens
            // using permissions.
            DevTeamData data = ModelWrapper.createDataCloneSafe(object, DevTeamData.class, rv);

            // update the database
            Date timestamp = new Date();
            data.logCreated(curUser.getServerKey(), timestamp);
            data.logUpdated(curUser.getServerKey(), timestamp);
            data = pm.makePersistent(data);
            // Convert the updated object back to
            return ModelWrapper.getClientCloneSafe(data, rv);
        } finally {
            pm.close();
        }

    }

    @Override
    public DevTeam write(DevTeam object)
            throws ServerException, UserPrivilegeException, DevTeamNotExists {

        PersistenceManager pm = PMF.getManager();
        try {
            RoleAndView rv = checkPermissions(object, pm);
            return DataServiceHelper.write(object, DevTeamData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            throw new DevTeamNotExists(object);
        }

    }

    @Override
    public ModelCollection<Game> getGames(DevTeam devTeam,
            Boolean includePrivate)
            throws ServerException, DevTeamNotExists {

        PersistenceManager pm = PMF.getManager();
        try {

            UserData curUser = UserGetter.getCurrentUser(pm);
            ClientView cv = getClientView(devTeam, pm);
            RoleAndView rv = new RoleAndView(curUser.getRole(), cv);
            DevTeamData dt = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());

            HashSet<Game> list = new HashSet<Game>(dt.getGames().size());
            for (Key k : dt.getGames()) {
                GameData gd;
                try {
                    gd = null;
                    gd = GameGetter.getGame(k.getId());
                    // Only show public games, unless asked to show private
                    if (includePrivate || gd.getIsPublic())
                    {
                        Game wrapped = ModelWrapper.getClientCloneSafe(gd,rv);
                        GameVersionData gvd = null;
                        if (gd.getActiveVersion() != null) {
                            gvd = pm.getObjectById(GameVersionData.class,
                                gd.getActiveVersion());
                        }
                        if (gvd != null) {
                            GameVersion gv = ModelWrapper.getClientCloneSafe(
                                                            gvd, rv);
                            wrapped.setActiveVersion(gv);
                        }
                        list.add(wrapped);
                    }
                } catch (GameNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<Game> toRet = new ModelCollection<Game>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;

        } finally {
            pm.close();
        }

    }

    @Override
    public ModelCollection<User> getUsers(DevTeam devTeam)
            throws ServerException, DevTeamNotExists {

        PersistenceManager pm = PMF.getManager();
        try {
            UserData curUser = UserGetter.getCurrentUser(pm);
            ClientView cv = getClientView(devTeam, pm);
            RoleAndView rv = new RoleAndView(curUser.getRole(), cv);
            DevTeamData dt = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());

            HashSet<User> list = new HashSet<User>(dt.getUsers().size());
            for (Key k : dt.getUsers()) {
                UserData ud;
                try {
                    ud = null;
                    ud = UserGetter.getUser(k.getId());
                    User wrapped = ModelWrapper.getClientCloneSafe(ud, rv);
                    list.add(wrapped);
                } catch (UserNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<User> toRet = new ModelCollection<User>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;

        } finally {
            pm.close();
        }

    }

    @Override
    public UpdatedPair<DevTeam, Game> addGame(DevTeam devTeam, Game game)
            throws ServerException, DevTeamNotExists, GameNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();
        try {
            RoleAndView rv = checkPermissions(devTeam, pm);

            DevTeamData dtd = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());
            GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());

            // do action
            dtd.addGame(gd);
            Date timestamp = new Date();
            dtd.logUpdated(rv.getUser().getServerKey(), timestamp);
            gd.logUpdated(rv.getUser().getServerKey(), timestamp);

            ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { dtd, gd };
            pm.makePersistentAll(objs);

            devTeam = ModelWrapper.getClientCloneSafe(dtd, rv);
            game = ModelWrapper.getClientCloneSafe(gd, rv);

            return new UpdatedPair<DevTeam, Game>(devTeam, game);
        } finally {
            pm.close();
        }
    }

    @Override
    public UpdatedPair<DevTeam, Game> removeGame(DevTeam devTeam, Game game)
                throws ServerException, DevTeamNotExists, GameNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();
        try {
            RoleAndView rv = checkPermissions(devTeam, pm);

            DevTeamData dtd = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());
            GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());

            // do action
            dtd.removeGame(gd);
            Date timestamp = new Date();
            dtd.logUpdated(rv.getUser().getServerKey(), timestamp);
            gd.logUpdated(rv.getUser().getServerKey(), timestamp);

            ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { dtd, gd };
            pm.makePersistentAll(objs);

            devTeam = ModelWrapper.getClientCloneSafe(dtd, rv);
            game = ModelWrapper.getClientCloneSafe(gd, rv);

            return new UpdatedPair<DevTeam, Game>(devTeam, game);
        } finally {
            pm.close();
        }
    }

    @Override
    public UpdatedPair<DevTeam, User> addUser(DevTeam devTeam, User user)
                throws ServerException, DevTeamNotExists, UserNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();
        try {
            RoleAndView rv = checkPermissions(devTeam, pm);

            DevTeamData dtd = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());
            UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());

            // do action
            dtd.addUser(ud);
            Date timestamp = new Date();
            dtd.logUpdated(rv.getUser().getServerKey(), timestamp);
            ud.logUpdated(rv.getUser().getServerKey(), timestamp);

            ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { dtd, ud };
            pm.makePersistentAll(objs);

            devTeam = ModelWrapper.getClientCloneSafe(dtd, rv);
            user = ModelWrapper.getClientCloneSafe(ud, rv);

            return new UpdatedPair<DevTeam, User>(devTeam, user);
        } finally {
            pm.close();
        }
    }

    @Override
    public UpdatedPair<DevTeam, User> removeUser(DevTeam devTeam, User user)
                throws ServerException, DevTeamNotExists, UserNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();
        try {
            RoleAndView rv = checkPermissions(devTeam, pm);

            DevTeamData dtd = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());
            UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());

            // do action
            dtd.removeUser(ud);
            Date timestamp = new Date();
            dtd.logUpdated(rv.getUser().getServerKey(), timestamp);
            ud.logUpdated(rv.getUser().getServerKey(), timestamp);

            ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { dtd, ud };
            pm.makePersistentAll(objs);

            devTeam = ModelWrapper.getClientCloneSafe(dtd, rv);
            user = ModelWrapper.getClientCloneSafe(ud, rv);

            return new UpdatedPair<DevTeam, User>(devTeam, user);
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelCollection<DevTeam> getDevTeams(int startIndex, int count)
            throws UserPrivilegeException, ServerException {

        PersistenceManager pm = PMF.getManager();
        try {
            UserData curUser = UserGetter.getCurrentUser(pm);
            if (curUser.isActive() && curUser.getRole() != Role.ADMIN) {
                throw new UserPrivilegeException("Only admins can retrieve all dev teams!");
            }

            // check the range
            if (startIndex < 0 || count < 1) {
                return new ModelCollection<DevTeam>();
            }

            Query q = pm.newQuery(DevTeamData.class);
            q.setOrdering("urlName asc");
            q.setRange(startIndex, startIndex + count);

            List<DevTeamData> results = (List<DevTeamData>)q.execute();

            if (results.isEmpty()) {
                return new ModelCollection<DevTeam>();
            }

            int c = results.size();
            if (count < c) { c = count; }

            RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PUBLIC);

            ModelCollection<DevTeam> list = new ModelCollection<DevTeam>(c);
            list.setStartIndex(startIndex);
            list.setEndIndex(startIndex+c);
            list.setCapacity(results.size());
            for (int i=0; i<c; i++) {
                DevTeam dt = ModelWrapper.getClientCloneSafe(results.get(i), rv);
                list.getCollection().add(dt);
            }
            return list;

        } finally {
            pm.close();
        }

    }

    @Override
    public void remove(DevTeam object) throws ServerException,
            UserPrivilegeException, DevTeamNotExists {

        PersistenceManager pm = PMF.getManager();
        try {
            // get currently logged in user and ownership
            UserData curUser = UserGetter.getCurrentUser(pm);

            // Only an admin can remove a dev team
            if (curUser.getRole() != Role.ADMIN) {
                String msg = "The current user does not have the the right privilege to"
                                    + " create the specified DevTeam object.";
                throw new UserPrivilegeException(msg);
            }

            // figure out the Role and ClientView
            ClientView cv = getClientView(object, pm);
            RoleAndView rv = new RoleAndView(curUser.getRole(), cv);

            // gets the server-side objects with its properties updated
            // from the client-side object. However, this only happens
            // using permissions.
            try {
                DataServiceHelper.remove(object, DevTeamData.class, rv);
            } catch (DataObjectNotExists e) {
                e.printStackTrace();
                throw new DevTeamNotExists(object);
            }
        } finally {
            pm.close();
        }
    }

    @Override
    public void remove(ModelCollection<DevTeam> teams) throws ServerException,
            UserPrivilegeException, DevTeamNotExists {
        for (DevTeam dt : teams.getCollection()) {
            remove(dt);
        }
    }

    /** Determine whether he can view a game in public or private view.
     * Same as checkPermissions, but doesn't throw an exception (so it is safe
     * to use for readers, which need to know whether to present the public or
     * private view, but don't want to die if there is no edit permission).
     */
    private ClientView getClientView(DevTeam object, PersistenceManager pm) {
        // get currently logged in user whether they are part of the devTeam
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.getId() == object.getPrimaryKey()) {
                isDev = true;
            }
        }
        // figure out the Role and ClientView
        return isDev ? ClientView.PRIVATE : ClientView.PUBLIC;
    }

    /** Determine whether he can view a game in public or private view.
     */
    private ClientView getClientView(DevTeamData object) {
        PersistenceManager pm = PMF.getManager();
        try
        {
            return getClientView(object, pm);
        }
        finally
        {
            pm.close();
        }
    }

    /** Determine whether he can view a game in public or private view.
     * Takes a DevTeamData instead of a DevTeam.
     */
    private ClientView getClientView(DevTeamData object,
        PersistenceManager pm) {
        // get currently logged in user whether they are part of the devTeam
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.equals(object.getServerKey())) {
                isDev = true;
            }
        }
        // figure out the Role and ClientView
        return isDev ? ClientView.PRIVATE : ClientView.PUBLIC;
    }

    private RoleAndView checkPermissions(DevTeam object, PersistenceManager pm) throws UserPrivilegeException {
        // get currently logged in user whether they are part of the devTeam
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.getId() == object.getPrimaryKey()) {
                isDev = true;
            }
        }
        // Only an admin or person in the DevTeam can create the game
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN && !isDev)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified DevTeam object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = isDev ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }
}
