/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;

import javax.jdo.PersistenceManager;

import com.google.appengine.api.datastore.Key;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class GameFileServiceImpl extends RemoteServiceServlet implements GameFileService {

    @Override
    public GameFile getByKey(Long key)
            throws ServerException, GameFileNotExists {

        GameFileData ad = GameFileGetter.getGameFile(key);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(ad, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public GameFile getByPath(GameVersion gameVersion, String path)
                throws ServerException, GameFileNotExists, GameVersionNotExists {

        PersistenceManager pm = PMF.getManager();

        try {
            GameVersionData gvd = GameVersionGetter.getGameVersion(pm, gameVersion.getPrimaryKey());
            GameData gd = GameGetter.getGame(pm, gvd.getGame().getId());
            GameFileData ad = GameFileGetter.getGameFile(pm, gd, gvd.getVersion(), path);
            UserData curUser = UserGetter.getCurrentUser(pm);
            return ModelWrapper.getClientCloneSafe(ad, curUser.getRole(), ClientView.PRIVATE);
        } catch (GameNotExists e) {
            e.printStackTrace();
            throw new GameFileNotExists();
        } finally {
            pm.close();
        }

    }


    @Override
    public GameFile getByPath(Game game, int gameVersion, String path)
                throws ServerException, GameFileNotExists, GameVersionNotExists {

        PersistenceManager pm = PMF.getManager();

        try {
            GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
            GameFileData ad = GameFileGetter.getGameFile(pm, gd, gameVersion, path);
            UserData curUser = UserGetter.getCurrentUser(pm);
            return ModelWrapper.getClientCloneSafe(ad, curUser.getRole(), ClientView.PRIVATE);
        } catch (GameNotExists e) {
            e.printStackTrace();
            throw new GameFileNotExists();
        } finally {
            pm.close();
        }
    }

    @Override
    public GameFile getByPath(String devteamName, String gameName,
                int gameVersion, String path)
                throws ServerException, GameFileNotExists, GameVersionNotExists {

        PersistenceManager pm = PMF.getManager();

        try {
            GameData gd = GameGetter.getGameByName(pm, devteamName, gameName);
            GameFileData ad = GameFileGetter.getGameFile(pm, gd, gameVersion, path);
            UserData curUser = UserGetter.getCurrentUser(pm);
            return ModelWrapper.getClientCloneSafe(ad, curUser.getRole(), ClientView.PRIVATE);
        } catch (GameNotExists e) {
            e.printStackTrace();
            throw new GameFileNotExists();
        } finally {
            pm.close();
        }

    }

    @Override
    public GameFile create(GameFile object)
            throws ServerException, UserPrivilegeException, GameFileExists, GameVersionNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // check for uniqueness
        try {
            GameVersionData gvd = GameVersionGetter.getGameVersion(object.getGameVersionKey());
            GameData gd = GameGetter.getGame(gvd.getGame().getId());
            if (GameFileGetter.getGameFile(pm, gd, gvd, object.getPath()) != null) {
                throw new GameFileExists();
            }
        }  catch (GameNotExists e) {
            /*  IF this happens it means the server has screwed
             *  up entity relations somewhere
             */
            throw new ServerException("Server Entity Relationship " +
                    "Management Error: " + e.getMessage());
        } catch (GameVersionNotExists e) {
            /*  IF this happens it means the server has screwed
             *  up entity relations somewhere
             */
            throw new ServerException("Server Entity Relationship " +
                    "Management Error: " + e.getMessage());
        }
        catch (GameFileNotExists e) {
            // this is excepted
        }

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        GameFileData data = ModelWrapper.createDataCloneSafe(object, GameFileData.class, rv);

        // update the database
        try {
            Date timestamp = new Date();
            data.logCreated(rv.getUser().getServerKey(), timestamp);
            data.logUpdated(rv.getUser().getServerKey(), timestamp);
            data = pm.makePersistent(data);
        } finally {
            pm.close();
        }

        // Convert the updated object back to
        return ModelWrapper.getClientCloneSafe(data, rv);

    }

    @Override
    public GameFile write(GameFile object)
        throws ServerException, UserPrivilegeException, GameFileNotExists, GameVersionNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, GameFileData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            //throw new GameFileNotExists(object);
            throw new GameFileNotExists(object.getPrimaryKey());
        }

    }

    @Override
    public void remove(GameFile object) throws ServerException,
            UserPrivilegeException, GameFileNotExists, GameVersionNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            DataServiceHelper.remove(object, GameFileData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            //throw new GameFileNotExists(object);
            throw new GameFileNotExists(object.getPrimaryKey());
        }

    }

    private RoleAndView checkPermissions(GameFile object, PersistenceManager pm)
      throws UserPrivilegeException, GameVersionNotExists, ServerException {
        // get currently logged in user whether they are part of the devTeam
        GameVersionData gv = GameVersionGetter.getGameVersion(object.getGameVersionKey());
        GameData g = null;
        try {
            g = GameGetter.getGame(gv.getGame().getId());
        } catch (GameNotExists e) {
            /*  IF this happens it means the server has screwed
             *  up entity relations somewhere
             */
            throw new ServerException("Server Entity Relationship " +
                    "Management Error: " + e.getMessage());
        }
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.equals(g.getDevTeam())) {
                isDev = true;
            }
        }
        // Only an admin or person in the DevTeam can create the game
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN && !isDev)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified GameFile object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = isDev ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }

}
