/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;
import java.util.HashSet;
import java.util.List;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class GameServiceImpl extends RemoteServiceServlet implements GameService {

    @Override
    public Game getByKey(Long key)
            throws ServerException, GameNotExists {

        GameData g = GameGetter.getGame(key);
        UserData curUser = UserGetter.getCurrentUser();
        ClientView cv = getClientView(g);
        return ModelWrapper.getClientCloneSafe(g, curUser.getRole(), cv);

    }

    @Override
    public Game getByName(String devteamName, String name)
            throws ServerException, GameNotExists {

        GameData g = GameGetter.getGameByName(devteamName, name);
        UserData curUser = UserGetter.getCurrentUser();
        ClientView cv = getClientView(g);
        return ModelWrapper.getClientCloneSafe(g, curUser.getRole(), cv);

    }

    @Override
    public Game getByGameToken(String token)
                throws ServerException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        try {
            GameData g = GameGetter.getGameByToken(pm, token);
            UserData curUser = UserGetter.getCurrentUser();
            ClientView cv = getClientView(g, pm);
            return ModelWrapper.getClientCloneSafe(g, curUser.getRole(), cv);
        } catch (GameTokenError e) {
            e.printStackTrace();
            throw new GameNotExists();
        } finally {
            pm.close();
        }

    }

    @Override
    public Game create(Game object)
                throws ServerException, UserPrivilegeException,
                       DevTeamNotExists, GameExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // check for uniqueness (within devteam)
        DevTeamData devteam= DevTeamGetter.getDevTeam(object.getDevTeamKey());
        try {
            if (GameGetter.getGameByName(pm, devteam.getUrlName(),
                                         object.getUrlName()) != null) {
                throw new GameExists(object);
            }
        } catch (GameNotExists e) {
            // this is excepted
        }

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        GameData data = ModelWrapper.createDataCloneSafe(object, GameData.class, rv);

        // update the database
        try {
            Date timestamp = new Date();
            data.logCreated(rv.getUser().getServerKey(), timestamp);
            data.logUpdated(rv.getUser().getServerKey(), timestamp);
            data = pm.makePersistent(data);
        } finally {
            pm.close();
        }

        // Convert the updated object back to
        return ModelWrapper.getClientCloneSafe(data, rv);

    }

    @Override
    public Game write(Game object)
                throws ServerException, UserPrivilegeException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, GameData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            throw new GameNotExists(object);
        }

    }

    @Override
    public ModelCollection<Achievement> getAchievements(Game game)
            throws ServerException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        ClientView cv = getClientView(game, pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), cv);
        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());

        try {

            HashSet<Achievement> list = new HashSet<Achievement>(gd.getAchievements().size());
            AchievementData a;
            for (Key k : gd.getAchievements()) {
                try {
                    a = null;
                    a = AchievementGetter.getAchievement(k.getId());
                    Achievement wrapped = ModelWrapper.getClientCloneSafe(a, rv);
                    list.add(wrapped);
                } catch (AchievementNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<Achievement> toRet = new ModelCollection<Achievement>();
            toRet.setCollection(list);
            toRet.setStartIndex(0);
            toRet.setEndIndex(list.size());
            toRet.setCapacity(list.size());
            return toRet;


        } finally {
            pm.close();
        }

    }

    @Override
    public ModelCollection<GameVersion> getGameVersions(Game game)
        throws ServerException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        ClientView cv = getClientView(game, pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), cv);
        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());

        try {

            HashSet<GameVersion> list = new HashSet<GameVersion>(gd.getGameVersions().size());
            for (Key k : gd.getGameVersions()) {
                GameVersionData gv;
                try {
                    gv = null;
                    gv = GameVersionGetter.getGameVersion(k.getId());
                    GameVersion wrapped = ModelWrapper.getClientCloneSafe(gv, rv);
                    list.add(wrapped);
                } catch (GameVersionNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<GameVersion> toRet = new ModelCollection<GameVersion>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;

        } finally {
            pm.close();
        }

    }

    @Override
    public ModelCollection<UserGameProfile> getUserGameProfiles(Game game)
            throws ServerException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        ClientView cv = getClientView(game, pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), cv);
        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());

        try {

            HashSet<UserGameProfile> list = new HashSet<UserGameProfile>(gd.getUserGameProfiles().size());
            for (Key k : gd.getUserGameProfiles()) {
                UserGameProfileData ugp;
                try {
                    ugp = null;
                    ugp = UserGameProfileGetter.getUserGameProfile(k.getId());
                    UserGameProfile wrapped = ModelWrapper.getClientCloneSafe(ugp, rv);
                    list.add(wrapped);
                } catch (UserGameProfileNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<UserGameProfile> toRet = new ModelCollection<UserGameProfile>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;

        } finally {
            pm.close();
        }

    }

    @Override
    public UpdatedPair<Game, Achievement>
                addAchievement(Game game, Achievement achievement)
                    throws ServerException, GameNotExists, AchievementNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(game, pm);

        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
        AchievementData ad = AchievementGetter.getAchievement(pm, achievement.getPrimaryKey());

        // do action
        gd.addAchievement(ad);
        Date timestamp = new Date();
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);
        ad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ad, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        game = ModelWrapper.getClientCloneSafe(gd, rv);
        achievement = ModelWrapper.getClientCloneSafe(ad, rv);

        return new UpdatedPair<Game, Achievement>(game, achievement);

    }

    @Override
    public UpdatedPair<Game, Achievement>
                removeAchievement(Game game, Achievement achievement)
                        throws ServerException, GameNotExists, AchievementNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(game, pm);

        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
        AchievementData ad = AchievementGetter.getAchievement(pm, achievement.getPrimaryKey());

        // do action
        gd.removeAchievement(ad);
        Date timestamp = new Date();
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);
        ad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ad, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        game = ModelWrapper.getClientCloneSafe(gd, rv);
        achievement = ModelWrapper.getClientCloneSafe(ad, rv);

        return new UpdatedPair<Game, Achievement>(game, achievement);

    }

    @Override
    public UpdatedPair<Game, GameVersion>
                addGameVersion(Game game, GameVersion version)
                        throws ServerException, GameNotExists, GameVersionNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(game, pm);

        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
        GameVersionData gvd = GameVersionGetter.getGameVersion(pm, version.getPrimaryKey());

        // do action
        gd.addGameVersion(gvd);
        if (gd.getActiveVersion() == null) {
            gd.setActiveVersion(gvd.getServerKey());
        }
        Date timestamp = new Date();
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);
        gvd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { gvd, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        game = ModelWrapper.getClientCloneSafe(gd, rv);
        version = ModelWrapper.getClientCloneSafe(gvd, rv);

        return new UpdatedPair<Game, GameVersion>(game, version);

    }

    @Override
    public UpdatedPair<Game, GameVersion>
                removeGameVersion(Game game, GameVersion version)
                        throws ServerException, GameNotExists, GameVersionNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(game, pm);

        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
        GameVersionData gvd = GameVersionGetter.getGameVersion(pm, version.getPrimaryKey());

        // do action
        gd.addGameVersion(gvd);
        if (gd.getActiveVersion().equals(gvd.getServerKey())) {
            gd.setActiveVersion(null);
        }
        Date timestamp = new Date();
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);
        gvd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { gvd, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        game = ModelWrapper.getClientCloneSafe(gd, rv);
        version = ModelWrapper.getClientCloneSafe(gvd, rv);

        return new UpdatedPair<Game, GameVersion>(game, version);

    }

    @Override
    public UpdatedPair<Game, UserGameProfile>
                addUserGameProfile(Game game, UserGameProfile ugp)
                       throws ServerException, GameNotExists, UserGameProfileNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(game, pm);

        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        // do action
        gd.addUserGameProfile(ugpd);
        Date timestamp = new Date();
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        game = ModelWrapper.getClientCloneSafe(gd, rv);
        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);

        return new UpdatedPair<Game, UserGameProfile>(game, ugp);

    }

    @Override
    public UpdatedPair<Game, UserGameProfile>
                removeUserGameProfile(Game game, UserGameProfile ugp)
                        throws ServerException, GameNotExists, UserGameProfileNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(game, pm);

        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        // do action
        gd.removeUserGameProfile(ugpd);
        Date timestamp = new Date();
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        game = ModelWrapper.getClientCloneSafe(gd, rv);
        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);

        return new UpdatedPair<Game, UserGameProfile>(game, ugp);

    }

    @Override
    public UpdatedPair<Game, DevTeam>
            addToDevTeam(Game game, DevTeam devTeam)
                throws ServerException, GameNotExists, DevTeamNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(game, pm);

        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
        DevTeamData dtd = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());

        // do action
        dtd.addGame(gd);
        Date timestamp = new Date();
        dtd.logUpdated(rv.getUser().getServerKey(), timestamp);
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { dtd, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        game = ModelWrapper.getClientCloneSafe(gd, rv);
        devTeam = ModelWrapper.getClientCloneSafe(dtd, rv);

        return new UpdatedPair<Game, DevTeam>(game, devTeam);

    }

    @Override
    public UpdatedPair<Game, DevTeam>
                removeFromDevTeam(Game game, DevTeam devTeam)
                    throws ServerException, GameNotExists, DevTeamNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(game, pm);

        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
        DevTeamData dtd = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());

        // do action
        dtd.removeGame(gd);
        Date timestamp = new Date();
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);
        dtd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { dtd, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        game = ModelWrapper.getClientCloneSafe(gd, rv);
        devTeam = ModelWrapper.getClientCloneSafe(dtd, rv);

        return new UpdatedPair<Game, DevTeam>(game, devTeam);

    }

    @Override
    public void remove(Game object) throws ServerException,
            UserPrivilegeException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            DataServiceHelper.remove(object, GameData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            throw new GameNotExists(object);
        }

    }

    /** Determine whether he can view a game in public or private view.
     * Same as checkPermissions, but doesn't throw an exception (so it is safe
     * to use for readers, which need to know whether to present the public or
     * private view, but don't want to die if there is no edit permission).
     */
    private ClientView getClientView(Game object, PersistenceManager pm) {
        // get currently logged in user whether they are part of the devTeam
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.getId() == object.getDevTeamKey()) {
                isDev = true;
            }
        }
        // figure out the Role and ClientView
        return isDev ? ClientView.PRIVATE : ClientView.PUBLIC;
    }

    /** Determine whether he can view a game in public or private view.
     */
    private ClientView getClientView(GameData object) {
        PersistenceManager pm = PMF.getManager();
        try
        {
            return getClientView(object, pm);
        }
        finally
        {
            pm.close();
        }
    }

    /** Determine whether he can view a game in public or private view.
     * Takes a GameData instead of a Game.
     */
    private ClientView getClientView(GameData object, PersistenceManager pm) {
        // get currently logged in user whether they are part of the devTeam
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.equals(object.getDevTeam())) {
                isDev = true;
            }
        }
        // figure out the Role and ClientView
        return isDev ? ClientView.PRIVATE : ClientView.PUBLIC;
    }

    private RoleAndView checkPermissions(Game object, PersistenceManager pm) throws UserPrivilegeException {
        // get currently logged in user whether they are part of the devTeam
        UserData curUser = UserGetter.getCurrentUser(pm);
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.getId() == object.getDevTeamKey()) {
                isDev = true;
            }
        }
        // Only an admin or person in the DevTeam can create the game
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN && !isDev)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified Game object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = isDev ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelCollection<Game> getGames(int startIndex, int count)
            throws ServerException, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        if (!curUser.isActive()) {
            throw new UserPrivilegeException("Only active users can retrieve all active games!");
        }

        // check the range
        if (startIndex < 0 || count < 1) {
            return new ModelCollection<Game>();
        }

        try {

            Query q = pm.newQuery(GameData.class);
            q.setOrdering("urlName asc");
            q.setFilter("isPublic == true");
            q.setRange(startIndex, startIndex + count);

            List<GameData> results = (List<GameData>)q.execute();

            if (results.isEmpty()) {
                return new ModelCollection<Game>();
            }

            int c = results.size();
            if (count < c) { c = count; }

            RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PUBLIC);

            ModelCollection<Game> list = new ModelCollection<Game>(c);
            list.setStartIndex(startIndex);
            list.setEndIndex(startIndex+c);
            list.setCapacity(results.size());
            for (int i=0; i<c; i++) {
                GameData gd = results.get(i);
                Game g = ModelWrapper.getClientCloneSafe(gd, rv);
                GameVersionData gvd = null;
                if (g != null && gd.getActiveVersion() != null) {
                    gvd = pm.getObjectById(GameVersionData.class, gd.getActiveVersion());
                }
                if (gvd != null) {
                    GameVersion gv = ModelWrapper.getClientCloneSafe(gvd, rv);
                    g.setActiveVersion(gv);
                }
                list.getCollection().add(g);
            }
            return list;

        } finally {
            pm.close();
        }

    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelCollection<Game> getActiveGames(int startIndex, int count)
            throws ServerException, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        if (!curUser.isActive()) {
            throw new UserPrivilegeException("Only active users can retrieve all active games!");
        }

        // check the range
        if (startIndex < 0 || count < 1) {
            return new ModelCollection<Game>();
        }

        try {

            Query q = pm.newQuery(GameData.class, "activeVersion != n");
            q.setOrdering("urlName asc");
            q.declareParameters("com.google.appengine.api.datastore.Key n");
            q.setRange(startIndex, startIndex + count);

            List<GameData> results = (List<GameData>)q.execute(null);

            if (results.isEmpty()) {
                return new ModelCollection<Game>();
            }

            int c = results.size();
            if (count < c) { c = count; }

            RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PUBLIC);

            ModelCollection<Game> list = new ModelCollection<Game>(c);
            list.setStartIndex(startIndex);
            list.setEndIndex(startIndex+c);
            list.setCapacity(results.size());
            for (int i=0; i<c; i++) {
                GameData gd = results.get(i);
                Game g = ModelWrapper.getClientCloneSafe(gd, rv);
                GameVersionData gvd = null;
                if (g != null && gd.getActiveVersion() != null) {
                    gvd = pm.getObjectById(GameVersionData.class, gd.getActiveVersion());
                }
                if (gvd != null) {
                    GameVersion gv = ModelWrapper.getClientCloneSafe(gvd, rv);
                    g.setActiveVersion(gv);
                }
                list.getCollection().add(g);
            }
            return list;

        } finally {
            pm.close();
        }

    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelCollection<Game> getMyGames(int startIndex, int count)
            throws ServerException, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        if (!curUser.isActive()) {
            throw new UserPrivilegeException("The logged in user is not active.");
        }

        // check the range
        if (startIndex < 0 || count < 1) {
            return new ModelCollection<Game>();
        }

        try {

            ModelCollection<Game> list = new ModelCollection<Game>();

            for (Key dt : curUser.getDevTeams()) {

                Query q = pm.newQuery(GameData.class, "devTeam == d");
                q.setOrdering("urlName asc");
                q.declareParameters("com.google.appengine.api.datastore.Key d");
                q.setRange(startIndex, startIndex + count);

                List<GameData> results = (List<GameData>)q.execute(dt);

                if (results.isEmpty()) {
                    continue;
                }

                int c = results.size();
                if (count < c) { c = count; }

                RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PUBLIC);

                list.setStartIndex(startIndex);
                list.setEndIndex(startIndex+c);
                list.setCapacity(results.size());
                for (int i=0; i<c; i++) {
                    GameData gd = results.get(i);
                    Game g = ModelWrapper.getClientCloneSafe(gd, rv);
                    GameVersionData gvd = null;
                    if (g != null && gd.getActiveVersion() != null) {
                        gvd = pm.getObjectById(GameVersionData.class, gd.getActiveVersion());
                    }
                    if (gvd != null) {
                        GameVersion gv = ModelWrapper.getClientCloneSafe(gvd, rv);
                        g.setActiveVersion(gv);
                    }
                    list.getCollection().add(g);
                }

            }

            return list;

        } finally {
            pm.close();
        }

    }

}
