/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;
import java.util.HashSet;

import javax.jdo.PersistenceManager;

import com.google.appengine.api.datastore.Key;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class GameVersionServiceImpl extends RemoteServiceServlet implements GameVersionService {

    @Override
    public GameVersion getByKey(Long key)
            throws ServerException, GameVersionNotExists {

        GameVersionData gv = GameVersionGetter.getGameVersion(key);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(gv, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public GameVersion getByVersion(Game game, int version)
        throws ServerException, GameNotExists, GameVersionNotExists {

        GameData gamedat = GameGetter.getGame(game.getPrimaryKey());
        GameVersionData gv = GameVersionGetter.getGameVersion(gamedat, version);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(gv, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public GameVersion create(GameVersion object)
                throws ServerException, UserPrivilegeException, GameVersionExists, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        // get currently logged in user and ownership
        RoleAndView rv = checkPermissions(object, pm);

        // check for uniqueness
        try {
            GameData gd = GameGetter.getGame(pm, object.getGameKey());
            if (GameVersionGetter.getGameVersion(pm, gd, object.getVersion()) != null) {
                throw new GameVersionExists();
            }
        } catch (GameVersionNotExists e) {
            // this is excepted
        } catch (GameNotExists e) {
            /*  IF this happens it means the server has screwed
             *  up entity relations somewhere
             */
            throw new ServerException("Server Entity Relationship " +
                    "Management Error: " + e.getMessage());
        }

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        GameVersionData data = ModelWrapper.createDataCloneSafe(object, GameVersionData.class, rv);
        // update the database
        try {
            Date timestamp = new Date();
            data.logCreated(rv.getUser().getServerKey(), timestamp);
            data.logUpdated(rv.getUser().getServerKey(), timestamp);
            data = pm.makePersistent(data);
        } finally {
            pm.close();
        }

        // Convert the updated object back to
        return ModelWrapper.getClientCloneSafe(data, rv);

    }

    @Override
    public GameVersion write(GameVersion object)
                throws ServerException, UserPrivilegeException, GameVersionNotExists, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        // get currently logged in user and ownership
        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, GameVersionData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            throw new GameVersionNotExists(object);
        }

    }

    @Override
    public ModelCollection<GameFile> getGameFiles(GameVersion version)
            throws ServerException, GameVersionNotExists {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PRIVATE);
        GameVersionData gvd = GameVersionGetter.getGameVersion(pm, version.getPrimaryKey());

        try {

            HashSet<GameFile> list = new HashSet<GameFile>(gvd.getGameFiles().size());
            for (Key k : gvd.getGameFiles()) {
                GameFileData gf;
                try {
                    gf = null;
                    gf = GameFileGetter.getGameFile(k.getId());
                    GameFile wrapped = ModelWrapper.getClientCloneSafe(gf, rv);
                    list.add(wrapped);
                } catch (GameFileNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<GameFile> toRet = new ModelCollection<GameFile>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;


        } finally {
            pm.close();
        }

    }

    @Override
    public UpdatedPair<GameVersion, GameFile>
                addGameFile(GameVersion version, GameFile gameFile)
                    throws ServerException, GameVersionNotExists, GameFileNotExists, UserPrivilegeException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(version, pm);

        GameVersionData gvd = GameVersionGetter.getGameVersion(pm, version.getPrimaryKey());
        GameFileData gfd = GameFileGetter.getGameFile(pm, gameFile.getPrimaryKey());

        // do action
        gvd.addGameFile(gfd);
        Date timestamp = new Date();
        gvd.logUpdated(rv.getUser().getServerKey(), timestamp);
        gfd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { gvd, gfd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        version = ModelWrapper.getClientCloneSafe(gvd, rv);
        gameFile = ModelWrapper.getClientCloneSafe(gfd, rv);

        return new UpdatedPair<GameVersion, GameFile>(version, gameFile);

    }

    @Override
    public UpdatedPair<GameVersion, GameFile>
                removeGameFile(GameVersion version, GameFile gameFile)
                    throws ServerException, GameVersionNotExists, GameFileNotExists, UserPrivilegeException, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(version, pm);

        GameVersionData gvd = GameVersionGetter.getGameVersion(pm, version.getPrimaryKey());
        GameFileData gfd = GameFileGetter.getGameFile(pm, gameFile.getPrimaryKey());

        // do action
        gvd.removeGameFile(gfd);
        Date timestamp = new Date();
        gvd.logUpdated(rv.getUser().getServerKey(), timestamp);
        gfd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { gvd, gfd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        version = ModelWrapper.getClientCloneSafe(gvd, rv);
        gameFile = ModelWrapper.getClientCloneSafe(gfd, rv);

        return new UpdatedPair<GameVersion, GameFile>(version, gameFile);

    }

    @Override
    public void remove(GameVersion object) throws ServerException,
            UserPrivilegeException, GameVersionNotExists, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        // get currently logged in user and ownership
        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            DataServiceHelper.remove(object, GameVersionData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            throw new GameVersionNotExists(object);
        }

    }

    private RoleAndView checkPermissions(GameVersion object, PersistenceManager pm)
            throws UserPrivilegeException, GameNotExists {
        // get currently logged in user whether they are part of the devTeam
        UserData curUser = UserGetter.getCurrentUser(pm);
        GameData gd = GameGetter.getGame(object.getGameKey());
        boolean isDev = false;
        for (Key k : curUser.getDevTeams()) {
            if (k.equals(gd.getDevTeam())) {
                isDev = true;
            }
        }
        // Only an admin or person in the DevTeam can create the game
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN && !isDev)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified GameVersion object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = isDev ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }

}
