/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;

import javax.jdo.PersistenceManager;

import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.api.KeyError;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class KeyValuePairServiceImpl extends RemoteServiceServlet implements KeyValuePairService {

    @Override
    public KeyValuePair getByKey(Long key)
            throws ServerException, KeyValuePairNotExists, UserPrivilegeException {

        KeyValuePairData kvpd = KeyValuePairGetter.getKeyValuePair(key);
        UserData curUser = UserGetter.getCurrentUser();
        // get currently logged in user whether they are the user Specified in the UGP
        UserGameProfileData ugp;
        try {
            ugp = UserGameProfileGetter.getUserGameProfile(kvpd.getPrimaryKey());
        } catch (UserGameProfileNotExists e) {
            throw new UserPrivilegeException("No User Game Profile for " +
                    "current User for the game this key belongs to");
        }
        // Only an admin or person in the DevTeam can create the game
        if (curUser.isActive() && (curUser.getRole() != Role.ADMIN &&
            !ugp.getUser().equals(curUser.getServerKey()))) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified KeyValuePair object.";
            throw new UserPrivilegeException(msg);
        }
        return ModelWrapper.getClientCloneSafe(kvpd, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public KeyValuePair getByPairKey(String key, String gameToken)
                throws ServerException, KeyValuePairNotExists, GameTokenError {
        try {
            KeyValuePairData kvpd = KeyValuePairGetter.getKeyValuePair(key, gameToken);
            UserData curUser = UserGetter.getCurrentUser();
            return ModelWrapper.getClientCloneSafe(kvpd, curUser.getRole(), ClientView.PRIVATE);
        } catch (KeyError e) {
            e.printStackTrace();
            throw new KeyValuePairNotExists();
        }

    }

    @Override
    public KeyValuePair write(KeyValuePair object)
                throws ServerException, UserPrivilegeException, KeyValuePairNotExists, UserGameProfileNotExists {

        PersistenceManager pm = PMF.getManager();

        // get currently logged in user and ownership
        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, KeyValuePairData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            //throw new UserAchievementNotExists(object);
            throw new KeyValuePairNotExists(object.getPrimaryKey());
        }

    }

    @Override
    public UpdatedPair<KeyValuePair, UserGameProfile>
                addToUserGameProfile(KeyValuePair kvp, UserGameProfile ugp)
                    throws ServerException, KeyValuePairNotExists, UserGameProfileNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(kvp, pm);

        KeyValuePairData kvpd = KeyValuePairGetter.getKeyValuePair(pm, kvp.getPrimaryKey());
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        // do action
        ugpd.addKeyValue(kvpd);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        kvpd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { kvpd, ugpd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        kvp = ModelWrapper.getClientCloneSafe(kvpd, rv);
        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);

        return new UpdatedPair<KeyValuePair, UserGameProfile>(kvp, ugp);

    }

    @Override
    public UpdatedPair<KeyValuePair, UserGameProfile>
                removeFromUserGameProfile(KeyValuePair kvp, UserGameProfile ugp)
                    throws ServerException, KeyValuePairNotExists, UserGameProfileNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(kvp, pm);

        KeyValuePairData kvpd = KeyValuePairGetter.getKeyValuePair(pm, kvp.getPrimaryKey());
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        // do action
        ugpd.removeKeyValue(kvpd);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        kvpd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { kvpd, ugpd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        kvp = ModelWrapper.getClientCloneSafe(kvpd, rv);
        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);

        return new UpdatedPair<KeyValuePair, UserGameProfile>(kvp, ugp);

    }

    private RoleAndView checkPermissions(KeyValuePair object, PersistenceManager pm)
            throws UserPrivilegeException, UserGameProfileNotExists {
        // get currently logged in user whether they are the user Specified in the UGP
        UserData curUser = UserGetter.getCurrentUser(pm);
        UserGameProfileData ugp = UserGameProfileGetter.getUserGameProfile(object.getUserGameProfile());
        // Only an admin or person in the DevTeam can create the game
        boolean curUserOwned = ugp.getUser().equals(curUser.getServerKey());
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN && curUserOwned)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified KeyValuePair object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = curUserOwned ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }

}
