/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;

import javax.jdo.PersistenceManager;

import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class PromotedGameServiceImpl extends RemoteServiceServlet implements PromotedGameService {

    @Override
    public PromotedGame getByKey(Long key)
            throws ServerException, PromotedGameNotExists {

        PromotedGameData pg = PromotedGameGetter.getPromotedGame(key);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(pg, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public PromotedGame getByName(String name)
                throws ServerException, PromotedGameNotExists {

        PromotedGameData pg = PromotedGameGetter.getPromotedGame(name);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(pg, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public PromotedGame create(PromotedGame object)
                throws ServerException, UserPrivilegeException, PromotedGameExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // check for uniqueness
        try {
            if (PromotedGameGetter.getPromotedGame(pm, object.getUrlName()) != null) {
                throw new PromotedGameExists();
            }
        } catch (PromotedGameNotExists e) {
            // this is excepted
        }


        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        PromotedGameData data = ModelWrapper.createDataCloneSafe(object, PromotedGameData.class, rv);

        // update the database
        try {
            Date timestamp = new Date();
            data.logCreated(rv.getUser().getServerKey(), timestamp);
            data.logUpdated(rv.getUser().getServerKey(), timestamp);
            data = pm.makePersistent(data);
        } finally {
            pm.close();
        }

        // Convert the updated object back to
        return ModelWrapper.getClientCloneSafe(data, rv);

    }

    @Override
    public PromotedGame write(PromotedGame object)
                throws ServerException, UserPrivilegeException, PromotedGameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, PromotedGameData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            throw new PromotedGameNotExists(object);
        }

    }

    @Override
    public void remove(PromotedGame object) throws ServerException,
            UserPrivilegeException, PromotedGameNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            DataServiceHelper.remove(object, PromotedGameData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            throw new PromotedGameNotExists(object);
        }

    }

    private RoleAndView checkPermissions(PromotedGame object, PersistenceManager pm)
        throws UserPrivilegeException{
        // get currently logged in user whether they are the user Specified in the UGP
        UserData curUser = UserGetter.getCurrentUser(pm);
        // Only an admin or person in the DevTeam can create the game
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified KeyValuePair object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        // ClientView.PUBLIC: Only admins have privilege over PromotedGames
        ClientView cv = ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }

}
