/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;

import javax.jdo.PersistenceManager;

import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class UserAchievementServiceImpl extends RemoteServiceServlet implements UserAchievementService {

    @Override
    public UserAchievement getByKey(Long key)
            throws ServerException, UserAchievementNotExists {

        UserAchievementData ua = UserAchievementGetter.getUserAchievement(key);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(ua, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public UserAchievement getByUserGameProfile(UserGameProfile ugp, Achievement achievement)
                throws ServerException, UserAchievementNotExists {

        PersistenceManager pm = PMF.getManager();

        try {
            UserAchievementData ua = UserAchievementGetter.getUserAchievement(pm, ugp.getPrimaryKey(), achievement.getPrimaryKey());
            UserData curUser = UserGetter.getCurrentUser();
            return ModelWrapper.getClientCloneSafe(ua, curUser.getRole(), ClientView.PRIVATE);
        } catch (AchievementNotExists e) {
            e.printStackTrace();
            throw new UserAchievementNotExists();
        } catch (UserGameProfileNotExists e) {
            e.printStackTrace();
            throw new UserAchievementNotExists();
        } finally {
            pm.close();
        }
    }

    @Override
    public UserAchievement write(UserAchievement object)
                throws ServerException, UserPrivilegeException, UserAchievementNotExists, UserGameProfileNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, UserAchievementData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            //throw new UserAchievementNotExists(object);
            throw new UserAchievementNotExists(object.getPrimaryKey());
        }

    }

    @Override
    public UpdatedPair<UserAchievement, UserGameProfile>
                addToUserGameProfile(UserAchievement userAchievement, UserGameProfile ugp)
                    throws ServerException, UserAchievementNotExists, UserGameProfileNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(userAchievement, pm);

        UserAchievementData uad = UserAchievementGetter.getUserAchievement(pm, userAchievement.getPrimaryKey());
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        // do action
        ugpd.addUserAchievement(uad);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        uad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, uad };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);
        userAchievement = ModelWrapper.getClientCloneSafe(uad, rv);

        return new UpdatedPair<UserAchievement, UserGameProfile>(userAchievement, ugp);

    }

    @Override
    public UpdatedPair<UserAchievement, UserGameProfile>
                removeFromUserGameProfile(UserAchievement userAchievement, UserGameProfile ugp)
                    throws ServerException, UserAchievementNotExists, UserGameProfileNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(userAchievement, pm);

        UserAchievementData uad = UserAchievementGetter.getUserAchievement(pm, userAchievement.getPrimaryKey());
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        // do action
        ugpd.removeUserAchievement(uad);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        uad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, uad };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);
        userAchievement = ModelWrapper.getClientCloneSafe(uad, rv);

        return new UpdatedPair<UserAchievement, UserGameProfile>(userAchievement, ugp);

    }

    @Override
    public UpdatedPair<UserAchievement, Achievement>
                addToAchievement(UserAchievement userAchievement, Achievement achievement)
                    throws ServerException, UserAchievementNotExists, AchievementNotExists, UserPrivilegeException, UserGameProfileNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(userAchievement, pm);

        UserAchievementData uad = UserAchievementGetter.getUserAchievement(pm, userAchievement.getPrimaryKey());
        AchievementData ad = AchievementGetter.getAchievement(pm, achievement.getPrimaryKey());

        // do action
        ad.addUserAchievement(uad);
        Date timestamp = new Date();
        ad.logUpdated(rv.getUser().getServerKey(), timestamp);
        uad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ad, uad };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        achievement = ModelWrapper.getClientCloneSafe(ad, rv);
        userAchievement = ModelWrapper.getClientCloneSafe(uad, rv);

        return new UpdatedPair<UserAchievement, Achievement>(userAchievement, achievement);

    }

    @Override
    public UpdatedPair<UserAchievement, Achievement>
            removeFromAchievement(UserAchievement userAchievement, Achievement achievement)
                throws ServerException, UserAchievementNotExists, AchievementNotExists, UserPrivilegeException, UserGameProfileNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(userAchievement, pm);

        UserAchievementData uad = UserAchievementGetter.getUserAchievement(pm, userAchievement.getPrimaryKey());
        AchievementData ad = AchievementGetter.getAchievement(pm, achievement.getPrimaryKey());

        // do action
        ad.removeUserAchievement(uad);
        Date timestamp = new Date();
        ad.logUpdated(rv.getUser().getServerKey(), timestamp);
        uad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ad, uad };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        achievement = ModelWrapper.getClientCloneSafe(ad, rv);
        userAchievement = ModelWrapper.getClientCloneSafe(uad, rv);

        return new UpdatedPair<UserAchievement, Achievement>(userAchievement, achievement);

    }

    private RoleAndView checkPermissions(UserAchievement object, PersistenceManager pm)
            throws UserPrivilegeException, UserGameProfileNotExists {
        // get currently logged in user whether they are the user Specified in the UGP
        UserData curUser = UserGetter.getCurrentUser(pm);
        UserGameProfileData ugp = UserGameProfileGetter.getUserGameProfile(object.getUserGameProfileKey());
        // Only an admin or person in the DevTeam can create the game
        boolean curUserOwned = ugp.getUser().equals(curUser.getServerKey());
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN && !curUserOwned)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified UserAchievement object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = curUserOwned ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }

}
