/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;
import java.util.HashSet;

import javax.jdo.PersistenceManager;

import com.google.appengine.api.datastore.Key;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class UserGameProfileServiceImpl extends RemoteServiceServlet implements UserGameProfileService {

    @Override
    public UserGameProfile getByKey(Long key)
        throws ServerException, UserGameProfileNotExists {

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(key);
        UserData curUser = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(ugpd, curUser.getRole(), ClientView.PRIVATE);

    }

    @Override
    public UserGameProfile getByUserAndGame(User user, Game game)
            throws ServerException, UserGameProfileNotExists, UserNotExists, GameNotExists {

        PersistenceManager pm = PMF.getManager();

        try {
            UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());
            GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());
            UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ud, gd);
            UserData curUser = UserGetter.getCurrentUser();
            return ModelWrapper.getClientCloneSafe(ugpd, curUser.getRole(), ClientView.PRIVATE);
        } finally {
            pm.close();
        }
    }

    @Override
    public UserGameProfile write(UserGameProfile object)
                throws ServerException, UserPrivilegeException, UserGameProfileNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, UserGameProfileData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            //throw new UserGameProfileNotExists(object);
            throw new UserGameProfileNotExists(object.getPrimaryKey());
        }

    }

    @Override
    public ModelCollection<KeyValuePair> getKeyValuePairs(UserGameProfile ugp)
            throws ServerException, UserGameProfileNotExists {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PRIVATE);
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        try {

            HashSet<KeyValuePair> list = new HashSet<KeyValuePair>(ugpd.getKeyValuePairs().size());
            for (Key k : ugpd.getKeyValuePairs()) {
                KeyValuePairData kvp;
                try {
                    kvp = null;
                    kvp = KeyValuePairGetter.getKeyValuePair(k.getId());
                    KeyValuePair wrapped = ModelWrapper.getClientCloneSafe(kvp, rv);
                    list.add(wrapped);
                } catch (KeyValuePairNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<KeyValuePair> toRet = new ModelCollection<KeyValuePair>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;

        } finally {
            pm.close();
        }

    }

    @Override
    public ModelCollection<UserAchievement> getUserAchievements(UserGameProfile ugp)
            throws ServerException, UserGameProfileNotExists {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PRIVATE);
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        try {

            HashSet<UserAchievement> list = new HashSet<UserAchievement>(ugpd.getUserAchievements().size());
            for (Key k : ugpd.getUserAchievements()) {
                UserAchievementData ua;
                try {
                    ua = null;
                    ua = UserAchievementGetter.getUserAchievement(k.getId());
                    UserAchievement wrapped = ModelWrapper.getClientCloneSafe(ua, rv);
                    list.add(wrapped);
                } catch (UserAchievementNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<UserAchievement> toRet = new ModelCollection<UserAchievement>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;

        } finally {
            pm.close();
        }

    }

    @Override
    public UpdatedPair<UserGameProfile, KeyValuePair>
                addKeyValuePair(UserGameProfile ugp, KeyValuePair kvp)
                        throws ServerException, UserGameProfileNotExists, KeyValuePairNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(ugp, pm);

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());
        KeyValuePairData kvpd = KeyValuePairGetter.getKeyValuePair(pm, kvp.getPrimaryKey());

        // do action
        ugpd.addKeyValue(kvpd);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        kvpd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { kvpd, ugpd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        kvp = ModelWrapper.getClientCloneSafe(kvpd, rv);
        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);

        return new UpdatedPair<UserGameProfile, KeyValuePair>(ugp, kvp);

    }

    @Override
    public UpdatedPair<UserGameProfile, KeyValuePair>
                removeKeyValuePair(UserGameProfile ugp, KeyValuePair kvp)
                        throws ServerException, UserGameProfileNotExists, KeyValuePairNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(ugp, pm);

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());
        KeyValuePairData kvpd = KeyValuePairGetter.getKeyValuePair(pm, kvp.getPrimaryKey());

        // do action
        ugpd.removeKeyValue(kvpd);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        kvpd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { kvpd, ugpd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        kvp = ModelWrapper.getClientCloneSafe(kvpd, rv);
        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);

        return new UpdatedPair<UserGameProfile, KeyValuePair>(ugp, kvp);

    }

    @Override
    public UpdatedPair<UserGameProfile, UserAchievement>
                addUserAchievement(UserGameProfile ugp, UserAchievement userAchievement)
                    throws ServerException, UserGameProfileNotExists, UserAchievementNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(ugp, pm);

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());
        UserAchievementData uad = UserAchievementGetter.getUserAchievement(pm, userAchievement.getPrimaryKey());

        // do action
        ugpd.addUserAchievement(uad);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        uad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, uad };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);
        userAchievement = ModelWrapper.getClientCloneSafe(uad, rv);

        return new UpdatedPair<UserGameProfile, UserAchievement>(ugp, userAchievement);

    }

    @Override
    public UpdatedPair<UserGameProfile, UserAchievement>
                removeUserAchievement(UserGameProfile ugp, UserAchievement userAchievement)
                        throws ServerException, UserGameProfileNotExists, UserAchievementNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();
        RoleAndView rv = checkPermissions(ugp, pm);


        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());
        UserAchievementData uad = UserAchievementGetter.getUserAchievement(pm, userAchievement.getPrimaryKey());

        // do action
        ugpd.removeUserAchievement(uad);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        uad.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, uad };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);
        userAchievement = ModelWrapper.getClientCloneSafe(uad, rv);

        return new UpdatedPair<UserGameProfile, UserAchievement>(ugp, userAchievement);

    }

    @Override
    public UpdatedPair<UserGameProfile, User>
                addToUser(UserGameProfile ugp, User user)
                     throws ServerException, UserGameProfileNotExists, UserNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(ugp, pm);

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());
        UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());

        // do action
        ud.addUserGameProfile(ugpd);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        ud.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, ud };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);
        user = ModelWrapper.getClientCloneSafe(ud, rv);

        return new UpdatedPair<UserGameProfile, User>(ugp, user);

    }

    @Override
    public UpdatedPair<UserGameProfile, User>
                removeFromUser(UserGameProfile ugp, User user)
                    throws ServerException, UserGameProfileNotExists, UserNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(ugp, pm);

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());
        UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());

        // do action
        ud.removeUserGameProfile(ugpd);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        ud.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, ud };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);
        user = ModelWrapper.getClientCloneSafe(ud, rv);

        return new UpdatedPair<UserGameProfile, User>(ugp, user);

    }

    @Override
    public UpdatedPair<UserGameProfile, Game>
                addToGame(UserGameProfile ugp, Game game)
                        throws ServerException, UserGameProfileNotExists, GameNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(ugp, pm);

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());
        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());

        // do action
        gd.addUserGameProfile(ugpd);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);
        game = ModelWrapper.getClientCloneSafe(gd, rv);

        return new UpdatedPair<UserGameProfile, Game>(ugp, game);

    }

    @Override
    public UpdatedPair<UserGameProfile, Game>
                removeFromGame(UserGameProfile ugp, Game game)
                        throws ServerException, UserGameProfileNotExists, GameNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(ugp, pm);

        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());
        GameData gd = GameGetter.getGame(pm, game.getPrimaryKey());

        // do action
        gd.removeUserGameProfile(ugpd);
        Date timestamp = new Date();
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);
        gd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, gd };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);
        game = ModelWrapper.getClientCloneSafe(gd, rv);

        return new UpdatedPair<UserGameProfile, Game>(ugp, game);

    }

    private RoleAndView checkPermissions(UserGameProfile object, PersistenceManager pm)
            throws UserPrivilegeException {
        // get currently logged in user whether they are the user Specified in the UGP
        UserData curUser = UserGetter.getCurrentUser(pm);
        // Only an admin or person in the DevTeam can create the game
        boolean curUserOwned = object.getUserKey().equals(curUser.getPrimaryKey());
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN && !curUserOwned)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified UserGameProfile object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = curUserOwned ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }
}
