/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.model.services;

import java.util.Date;
import java.util.HashSet;
import java.util.List;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import com.google.appengine.api.datastore.Key;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.shared.model.*;
import au.edu.unimelb.csse.mugle.shared.model.services.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.*;

@SuppressWarnings("serial")
public class UserServiceImpl extends RemoteServiceServlet implements UserService {

    @Override
    public User getByKey(Long key) throws UserNotExists, ServerException {

        UserData ud = UserGetter.getUser(key);
        UserData curUser = UserGetter.getCurrentUser();

        ClientView cv = (curUser.getPrimaryKey().equals(ud.getPrimaryKey())) ?
                                    ClientView.PRIVATE : ClientView.PUBLIC;

        return ModelWrapper.getClientCloneSafe(ud, curUser.getRole(), cv);

    }

    @Override
    public User getByName(String urlName) throws UserNotExists, ServerException {

        UserData ud = UserGetter.getUserByName(urlName);
        UserData curUser = UserGetter.getCurrentUser();

        ClientView cv = (curUser.getPrimaryKey().equals(ud.getPrimaryKey())) ?
                                        ClientView.PRIVATE : ClientView.PUBLIC;

        return ModelWrapper.getClientCloneSafe(ud, curUser.getRole(), cv);

    }

    @Override
    public User getLoggedInUser() throws ServerException {
        UserData ud = UserGetter.getCurrentUser();
        return ModelWrapper.getClientCloneSafe(ud, ud.getRole(), ClientView.PRIVATE);
    }

    @Override
    public User create(User object)
                throws ServerException, UserPrivilegeException, UserExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // check for uniqueness
        try {
            if (UserGetter.getUserByName(pm, object.getUrlName()) != null) {
                throw new UserExists(object);
            }
        } catch (UserNotExists e) {
            // this is excepted
        }

        // gets the server-side objects with its properties updated
        // from the client-side object. However, this only happens
        // using permissions.
        UserData data = ModelWrapper.createDataCloneSafe(object, UserData.class, rv);

        // update the database
        try {
            Date timestamp = new Date();
            data.logCreated(rv.getUser().getServerKey(), timestamp);
            data.logUpdated(rv.getUser().getServerKey(), timestamp);
            data = pm.makePersistent(data);
        } finally {
            pm.close();
        }

        // Convert the updated object back to
        return ModelWrapper.getClientCloneSafe(data, rv);

    }

    @Override
    public User write(User object)
                throws ServerException, UserPrivilegeException, UserExists,
                       UserNotExists {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(object, pm);

        // check for uniqueness. Three possibilities:
        // 1. User with same urlname and key. No name change; fine.
        // 2. User with same urlname but different key. Attempt to write over
        //    another user's name: Forbidden.
        // 3. No user with same urlname. Valid name change; fine.
        try {
            UserData userWithName =
                UserGetter.getUserByName(pm, object.getUrlName());
            if (userWithName != null &&
                !userWithName.getPrimaryKey().equals(object.getPrimaryKey()))
            {
                // this is forbidden -- case #2
                throw new UserExists(object);
            }
            // else, fine -- case #1
        } catch (UserNotExists e) {
            // this is fine -- case #3
        }

        // gets the server-side objects with its properties updated
        // from the client-side object, write to the database and
        // get the updated object back. However, this only happens
        // using permissions.
        try {
            return DataServiceHelper.write(object, UserData.class, rv);
        } catch (DataObjectNotExists e) {
            e.printStackTrace();
            throw new UserNotExists(object);
        }

    }

    @Override
    public ModelCollection<DevTeam> getDevTeams(User user) throws UserNotExists, ServerException {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PRIVATE);
        UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());

        try {

            HashSet<DevTeam> list = new HashSet<DevTeam>(ud.getDevTeams().size());
            for (Key k : ud.getDevTeams()) {
                DevTeamData dt;
                try {
                    dt = null;
                    dt = DevTeamGetter.getDevTeam(k.getId());
                    DevTeam wrapped = ModelWrapper.getClientCloneSafe(dt, rv);
                    list.add(wrapped);
                } catch (DevTeamNotExists e) {
                    /*  IF this happens it means the server has screwed
                     *  up entity relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship " +
                            "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<DevTeam> toRet = new ModelCollection<DevTeam>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;

        } finally {
            pm.close();
        }

    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelCollection<UserGameProfile> getUserGameProfiles(User user) throws UserNotExists, ServerException {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PRIVATE);
        UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());

        try {

            Query q = pm.newQuery(UserGameProfile.class, "user == u");
            q.declareParameters("com.google.appengine.api.datastore.Key u");
            List<UserGameProfileData> ugps = (List<UserGameProfileData>)q.execute(ud.getServerKey());
            HashSet<UserGameProfile> list = new HashSet<UserGameProfile>(ugps.size());
            for (UserGameProfileData ugp : ugps) {
                UserGameProfile wrapped = ModelWrapper.getClientCloneSafe(ugp, rv);
                list.add(wrapped);
            }
            ModelCollection<UserGameProfile> toRet = new ModelCollection<UserGameProfile>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;

        } finally {
            pm.close();
        }

    }

    @Override
    public UpdatedPair<User, DevTeam>
                joinDevTeam(User user, DevTeam devTeam)
                    throws ServerException, UserNotExists, DevTeamNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(user, pm);

        UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());
        DevTeamData dtd = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());

        // do action
        dtd.addUser(ud);
        Date timestamp = new Date();
        dtd.logUpdated(rv.getUser().getServerKey(), timestamp);
        ud.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { dtd, ud };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        user = ModelWrapper.getClientCloneSafe(ud, rv);
        devTeam = ModelWrapper.getClientCloneSafe(dtd, rv);

        return new UpdatedPair<User, DevTeam>(user, devTeam);

    }

    @Override
    public UpdatedPair<User, DevTeam>
                leaveDevTeam(User user, DevTeam devTeam)
                    throws ServerException, UserNotExists, DevTeamNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(user, pm);

        UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());
        DevTeamData dtd = DevTeamGetter.getDevTeam(pm, devTeam.getPrimaryKey());

        // do action
        dtd.removeUser(ud);
        Date timestamp = new Date();
        dtd.logUpdated(rv.getUser().getServerKey(), timestamp);
        ud.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { dtd, ud };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        user = ModelWrapper.getClientCloneSafe(ud, rv);
        devTeam = ModelWrapper.getClientCloneSafe(dtd, rv);

        return new UpdatedPair<User, DevTeam>(user, devTeam);

    }

    @Override
    public UpdatedPair<User, UserGameProfile>
                addUserGameProfile(User user, UserGameProfile ugp)
                    throws ServerException, UserNotExists, UserGameProfileNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(user, pm);

        UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        // do action
        ud.addUserGameProfile(ugpd);
        Date timestamp = new Date();
        ud.logUpdated(rv.getUser().getServerKey(), timestamp);
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, ud };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        user = ModelWrapper.getClientCloneSafe(ud, rv);
        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);

        return new UpdatedPair<User, UserGameProfile>(user, ugp);

    }

    @Override
    public UpdatedPair<User, UserGameProfile>
                removeUserGameProfile(User user, UserGameProfile ugp)
                        throws ServerException, UserNotExists, UserGameProfileNotExists, UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();

        RoleAndView rv = checkPermissions(user, pm);

        UserData ud = UserGetter.getUser(pm, user.getPrimaryKey());
        UserGameProfileData ugpd = UserGameProfileGetter.getUserGameProfile(pm, ugp.getPrimaryKey());

        // do action
        ud.removeUserGameProfile(ugpd);
        Date timestamp = new Date();
        ud.logUpdated(rv.getUser().getServerKey(), timestamp);
        ugpd.logUpdated(rv.getUser().getServerKey(), timestamp);

        ModelDataClass<?,?,?> objs[] = new ModelDataClass<?,?,?>[] { ugpd, ud };
        try {
            pm.makePersistentAll(objs);
        } finally {
            pm.close();
        }

        user = ModelWrapper.getClientCloneSafe(ud, rv);
        ugp = ModelWrapper.getClientCloneSafe(ugpd, rv);

        return new UpdatedPair<User, UserGameProfile>(user, ugp);

    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelCollection<User> getUsers(int startIndex, int count)
                throws UserPrivilegeException, ServerException {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        if (curUser.getRole() != Role.ADMIN) {
            throw new UserPrivilegeException("Only admins can retrieve all users!");
        }

        // check the range
        if (startIndex < 0 || count < 1) {
            return new ModelCollection<User>();
        }

        try {

            Query q = pm.newQuery(UserData.class);
            q.setRange(startIndex, startIndex + count);

            List<UserData> results = (List<UserData>)q.execute();

            if (results.isEmpty()) {
                return new ModelCollection<User>();
            }

            int c = results.size();
            if (count < c) { c = count; }

            RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PUBLIC);

            ModelCollection<User> list = new ModelCollection<User>(c);
            list.setStartIndex(startIndex);
            list.setEndIndex(startIndex+c);
            for (int i=0; i<c; i++) {
                User u = ModelWrapper.getClientCloneSafe(results.get(i), rv);
                list.getCollection().add(u);
            }
            return list;

        } finally {
            pm.close();
        }

    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelCollection<User> getUsersByRole(Role role, int startIndex,
            int count) throws UserPrivilegeException, ServerException {

        PersistenceManager pm = PMF.getManager();

        UserData curUser = UserGetter.getCurrentUser(pm);
        if (curUser.getRole() != Role.ADMIN) {
            throw new UserPrivilegeException("Only admins can retrieve all users!");
        }

        // check the range
        if (startIndex < 0 || count < 1) {
            return new ModelCollection<User>();
        }

        try {

            Query q = pm.newQuery(UserData.class, "role == r");
            q.declareParameters("Role r");
            q.setRange(startIndex, startIndex + count);

            List<UserData> results = (List<UserData>)q.execute(role);

            if (results.isEmpty()) {
                return new ModelCollection<User>();
            }

            int c = results.size();
            if (count < c) { c = count; }

            RoleAndView rv = new RoleAndView(curUser.getRole(), ClientView.PUBLIC);

            ModelCollection<User> list = new ModelCollection<User>(c);
            list.setStartIndex(startIndex);
            list.setEndIndex(startIndex+c);
            for (int i=0; i<c; i++) {
                User u = ModelWrapper.getClientCloneSafe(results.get(i), rv);
                list.getCollection().add(u);
            }
            return list;

        } finally {
            pm.close();
        }

    }

    private RoleAndView checkPermissions(User object, PersistenceManager pm)
    throws UserPrivilegeException {
        // get currently logged in user whether they are the user Specified
        UserData curUser = UserGetter.getCurrentUser(pm);
        // Only an admin or person in the DevTeam can create the game
        boolean curUserOwned = curUser.getPrimaryKey().equals(object.getPrimaryKey());
        if (!curUser.isActive() || (curUser.getRole() != Role.ADMIN &&
            !curUserOwned)) {
            String msg = "The current user does not have the the right privilege to"
                                + " modify the specified User object.";
            throw new UserPrivilegeException(msg);
        }
        // figure out the Role and ClientView
        ClientView cv = curUserOwned ? ClientView.PRIVATE : ClientView.PUBLIC;

        return new RoleAndView(curUser.getRole(), cv, curUser);

    }

}
