/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.multiplayer.api;

import java.util.HashSet;
import java.util.Set;

import javax.jdo.PersistenceManager;

import au.edu.unimelb.csse.mugle.client.multiplayer.api.ChannelService;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.UserData;
import au.edu.unimelb.csse.mugle.server.model.UserGetter;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.*;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.InstanceTokenError;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.PlayerNotFoundInGame;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.UserNotExists;

import com.google.appengine.api.channel.ChannelMessage;
import com.google.appengine.api.channel.ChannelServiceFactory;
import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.memcache.MemcacheService;
import com.google.appengine.api.memcache.MemcacheServiceFactory;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

/**
 * The server side implementation of ChannelService
 */
@SuppressWarnings("serial")
public class ChannelServiceImpl extends RemoteServiceServlet implements
        ChannelService {
    public static HashSet<String> channelIds = new HashSet<String>();

    @Override
    public String getChannelToken() {
        String channelId = String.valueOf(UserGetter.getCurrentUser()
                .getPrimaryKey());
        String channelToken = ChannelServiceFactory.getChannelService()
                .createChannel(channelId);
        channelIds.add(channelId);
        return channelToken;
    }

    @Override
    public void sendMessage(String message, String instanceToken)
            throws InstanceTokenError, PlayerNotFoundInGame {
        UserData sender = UserGetter.getCurrentUser();
        GameInstanceData instance = GameInstanceGetter
                .getInstanceByToken(instanceToken);
        GameInstanceGetter.findPlayerInGame(instance, sender);
        sendMessageToInstance(message, sender, instance);
    }

    public static void sendMessage(String message, UserData receiver) {
        ChannelServiceFactory.getChannelService().sendMessage(
                new ChannelMessage(String.valueOf(receiver.getPrimaryKey()),
                        message));
    }

    // Send message to a group within the same instance
    @Override
    public void sendMessage(String message, String instanceToken,
            Set<String> receivers) throws InstanceTokenError,
            PlayerNotFoundInGame, UserNotExists {
        UserData sender = UserGetter.getCurrentUser();
        GameInstanceData instance = GameInstanceGetter
                .getInstanceByToken(instanceToken);
        Set<Key> players = instance.getActivePlayers();
        GameInstanceGetter.findPlayerInGame(instance, sender);
        UserData user;
        for (String name : receivers) {
            user = UserGetter.getUserByName(name);
            if (!players.contains(user.getServerKey())) {
                if (receivers.size() == 1)
                    throw new PlayerNotFoundInGame(user.getUrlName());
                continue;
            }
            ChannelServiceFactory.getChannelService().sendMessage(
                    new ChannelMessage(String.valueOf(user.getPrimaryKey()),
                            message));
        }
    }

    public static void sendMessageToInstance(String message, UserData sender,
            GameInstanceData instance) {
        MemcacheService cache = MemcacheServiceFactory.getMemcacheService();
        Long counter = cache.increment(instance.getPrimaryKey(), 1, (long) 0);
        // when counter is 1, it is a new created cache either when it's a fresh
        // value or the old value is not available in the memache.
        if (counter == 1) {
            // get the counter from datastore.
            counter = (long) PlayerMessageGetter.getMessageCounter(instance);
            if (counter != 1)
                // put the new counter into cache when it is not the first time
                // created.
                cache.put(instance.getPrimaryKey(), counter);
        }
        // add counter to the message
        message = message.substring(0, message.length() - 1) + ",\"counter\":"
                + String.valueOf(counter) + "}";
        putMessageToStore(message, sender, instance);
        Long senderId = sender.getPrimaryKey();
        Set<Key> receivers = instance.getActivePlayers();
        for (Key key : receivers) {
            if (key.getId() != senderId)
                ChannelServiceFactory.getChannelService()
                        .sendMessage(
                                new ChannelMessage(String.valueOf(key.getId()),
                                        message));
        }
    }

    private static void putMessageToStore(String message, UserData sender,
            GameInstanceData instance) {
        PersistenceManager pm = PMF.getManager();
        try {
            PlayerMessageData playerMessage = PlayerMessageGetter
                    .getMessageData(sender, instance);
            playerMessage.addMessage(message);
            pm.makePersistent(playerMessage);
        } finally {
            pm.close();
        }
    }

    public static void sendMessageToPending(String message,
            PendingPlayerData pending) {
        Set<Key> receivers = pending.getPlayers();
        for (Key key : receivers) {
            ChannelServiceFactory.getChannelService().sendMessage(
                    new ChannelMessage(String.valueOf(key.getId()), message));
        }
    }

    @Override
    public String getMessage(String instanceToken, int counter)
            throws InstanceTokenError, PlayerNotFoundInGame {
        PersistenceManager pm = PMF.getManager();
        try {
            UserData player = UserGetter.getCurrentUser(pm);
            GameInstanceData instance = GameInstanceGetter.getInstanceByToken(
                    pm, instanceToken);
            GameInstanceGetter.findPlayerInGame(instance, player);
            return PlayerMessageGetter.getMessage(instance, counter);
        } finally {
            pm.close();
        }
    }

    @Override
    public String[] getMessages(String instanceToken, int[] sequence)
            throws InstanceTokenError, PlayerNotFoundInGame {
        PersistenceManager pm = PMF.getManager();
        try {
            UserData player = UserGetter.getCurrentUser(pm);
            GameInstanceData instance = GameInstanceGetter.getInstanceByToken(
                    pm, instanceToken);
            GameInstanceGetter.findPlayerInGame(instance, player);
            return PlayerMessageGetter.getMessages(instance, sequence);
        } finally {
            pm.close();
        }
    }

    @Override
    public String[] getAllMessages(String instanceToken)
            throws InstanceTokenError, PlayerNotFoundInGame {
        PersistenceManager pm = PMF.getManager();
        try {
            UserData player = UserGetter.getCurrentUser(pm);
            GameInstanceData instance = GameInstanceGetter.getInstanceByToken(
                    pm, instanceToken);
            GameInstanceGetter.findPlayerInGame(instance, player);
            return PlayerMessageGetter.getAllMessages(pm, instance);
        } finally {
            pm.close();
        }

    }

    // TODO sendMessageToAll?
}
