/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.multiplayer.api;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import au.edu.unimelb.csse.mugle.client.multiplayer.api.MultiPlayerGameService;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.*;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.UserNotExists;

import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.KeyFactory;
import com.google.appengine.api.memcache.MemcacheService;
import com.google.appengine.api.memcache.MemcacheServiceFactory;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

/**
 * The server side implementation of MultiPlayerGameService
 */
@SuppressWarnings("serial")
public class MultiPlayerGameServiceImpl extends RemoteServiceServlet implements
        MultiPlayerGameService {
    @Override
    public String createGameInstance(String gameToken, int size,
            boolean isPublic) throws GameTokenError {
        PersistenceManager pm = PMF.getManager();
        String instanceToken;
        try {
            GameInstanceData instance = new GameInstanceData();
            Key gameKey = GameGetter.getCurrentGame(pm, gameToken)
                    .getServerKey();
            instance.setGame(gameKey);
            UserData creator = UserGetter.getCurrentUser();
            instance.setName(creator.getUrlName());
            instance.setCreatedUser(creator.getServerKey());
            instance.setActive(true);
            instance.setJoinable(isPublic);
            instance.setSize(size);
            instance.setCreatedTime(new Date());

            pm.makePersistent(instance);

            instanceToken = instance.getToken();
            if (isPublic) {
                MessageData message = new MessageData();
                message.setServerCommand(ServerCommand.NEW_GAME);
                message.setBody(instanceToken);
                message.setSender(creator.getUrlName());

                HashMap<String, String> attributes = new HashMap<String, String>();
                attributes.put("name", instance.getName());
                attributes.put("playerCount", String.valueOf(1));
                attributes.put("size", String.valueOf(instance.getSize()));
                attributes.put("token", instanceToken);
                message.setAttributeValuePairs(attributes);

                PendingPlayerData pending = PendingPlayerGetter
                        .getPendingByGameKey(pm, gameKey);
                pending.removePlayer(creator.getServerKey());
                ChannelServiceImpl.sendMessageToPending(message.toString(),
                        pending);
            }

        } finally {
            pm.close();
        }
        return instanceToken;
    }

    @Override
    public GameInfo[] getGameList(String gameToken) throws GameTokenError {
        PersistenceManager pm = PMF.getManager();
        try {
            Key gameKey = GameGetter.getCurrentGame(gameToken).getServerKey();
            PendingPlayerData pending = PendingPlayerGetter
                    .getPendingByGameKey(pm, gameKey);
            UserData player = UserGetter.getCurrentUser(pm);
            pending.addPlayer(player.getServerKey());
            pm.makePersistent(pending);

            List<GameInstanceData> instances = GameInstanceGetter
                    .getActiveInstancesByGameKey(pm, gameKey);
            ArrayList<GameInfo> infoList = new ArrayList<GameInfo>();
            boolean canContinue;
            for (GameInstanceData instance : instances) {
                canContinue = instance.containExitPlayer(player.getServerKey());
                if (!instance.joinable() && !canContinue)
                    continue;
                GameInfo info = new GameInfo(instance.getName(),
                        instance.getActivePlayerCount(), instance.getSize(),
                        KeyFactory.keyToString(instance.getServerKey()));
                if (canContinue) {
                    info.setCanContinue();
                }
                infoList.add(info);
            }
            return infoList.toArray(new GameInfo[infoList.size()]);
        } finally {
            pm.close();
        }
    }

    @Override
    public void setGameJoinable(String instanceToken, boolean joinable)
            throws InstanceTokenError, PlayerNotFoundInGame {
        PersistenceManager pm = PMF.getManager();
        try {
            UserData player = UserGetter.getCurrentUser(pm);
            GameInstanceData instance = GameInstanceGetter.getInstanceByToken(
                    pm, instanceToken);
            GameInstanceGetter.findPlayerInGame(instance, player);
            instance.setJoinable(joinable);
        } finally {
            pm.close();
        }
    }

    @Override
    public String[] joinGame(String instanceToken) throws InstanceTokenError,
            GameInstanceNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            GameInstanceData instance = GameInstanceGetter.getInstanceByToken(
                    pm, instanceToken);
            UserData player = UserGetter.getCurrentUser(pm);
            if (!instance.isActive())
                throw new GameInstanceNotExists(instance.getName());
            if (!instance.hasRoom())
                return null; // TODO throws exception
            if (instance.joinable()
                    || instance.containExitPlayer(player.getServerKey())) {
                if (instance.containExitPlayer(player.getServerKey()))
                    instance.removeExitPlayer(player.getServerKey());
                instance.addPlayer(UserGetter.getCurrentUser(pm).getServerKey());
                MessageData message = new MessageData();
                message.setSender(player.getUrlName());
                message.setServerCommand(ServerCommand.PLAYER_JOINED);
                message.setBody(player.getUrlName() + " Joined");
                message.setCreatedTime(new Date());
                message.setInstanceToken(instance.getToken());
                ChannelServiceImpl.sendMessageToInstance(message.toString(),
                        player, instance);
                PendingPlayerData pending = PendingPlayerGetter
                        .getPendingByGameKey(pm, instance.getGame());
                pending.removePlayer(player.getServerKey());
                updateGameState(instance, pending, player);

                String[] players = new String[instance.getActivePlayerCount()];
                int i = 0;
                for (Key key : instance.getActivePlayers())
                    players[i++] = UserGetter.getUser(key).getUrlName();
                return players;
            }

        } catch (UserNotExists e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } finally {
            pm.close();
        }
        return null;
    }

    @Override
    public void removePlayer(String instanceToken, String name)
            throws InstanceTokenError, PlayerNotFoundInGame {
        PersistenceManager pm = PMF.getManager();
        try {
            UserData player = UserGetter.getCurrentUser(pm);
            GameInstanceData instance = GameInstanceGetter.getInstanceByToken(
                    pm, instanceToken);
            if (!instance.containPlayer(player.getServerKey()))
                throw new PlayerNotFoundInGame(player.getUrlName());
            PendingPlayerData pending = PendingPlayerGetter
                    .getPendingByGameKey(pm, instance.getGame());
            removePlayer(pm, pending, instance,
                    UserGetter.getUserByName(pm, name),
                    ServerCommand.PLAYER_REMOVED);
        } catch (UserNotExists e) {
            throw new PlayerNotFoundInGame(name);
        } finally {
            pm.close();
        }
    }

    @Override
    public void quitGame(String instanceToken) throws InstanceTokenError,
            PlayerNotFoundInGame {
        PersistenceManager pm = PMF.getManager();
        try {
            GameInstanceData instance = GameInstanceGetter.getInstanceByToken(
                    pm, instanceToken);
            PendingPlayerData pending = PendingPlayerGetter
                    .getPendingByGameKey(pm, instance.getGame());
            UserData player = UserGetter.getCurrentUser(pm);
            instance.moveToExit(player.getServerKey());
            removePlayer(pm, pending, instance, player,
                    ServerCommand.PLAYER_QUIT);
        } finally {
            pm.close();
        }

    }

    public void disconnectGame(PersistenceManager pm,
            PendingPlayerData pending, GameInstanceData instance,
            UserData player) throws PlayerNotFoundInGame {
        instance.moveToExit(player.getServerKey());
        removePlayer(pm, pending, instance, player,
                ServerCommand.PLAYER_DISCONNECTED);
    }

    // TODO needs transaction
    private void removePlayer(PersistenceManager pm, PendingPlayerData pending,
            GameInstanceData instance, UserData player, ServerCommand command)
            throws PlayerNotFoundInGame {
        String playerName = player.getUrlName();
        if (!instance.getActivePlayers().contains(player.getServerKey()))
            throw new PlayerNotFoundInGame(playerName);
        instance.removePlayer(player.getServerKey());
        MessageData message = new MessageData();
        message.setSender(playerName);
        message.setCreatedTime(new Date());
        message.setInstanceToken(instance.getToken());

        switch (command) {
        case PLAYER_QUIT:
            message.setServerCommand(ServerCommand.PLAYER_QUIT);
            message.setBody(playerName + " Quit");
            pending.addPlayer(player.getServerKey());
            break;
        case PLAYER_REMOVED:
            message.setServerCommand(ServerCommand.PLAYER_REMOVED);
            message.setBody(playerName + " Removed");
            ChannelServiceImpl.sendMessage(message.toString(), player);
            pending.addPlayer(player.getServerKey());
            break;
        case PLAYER_DISCONNECTED:
            message.setServerCommand(ServerCommand.PLAYER_DISCONNECTED);
            message.setBody(playerName + " Dsconnected");
            break;
        default:
            break;
        }
        ChannelServiceImpl.sendMessageToInstance(message.toString(), player,
                instance);

        // Remove the game when there are no players
        if (instance.getActivePlayerCount() == 0) {
            message.setServerCommand(ServerCommand.GAME_REMOVED);
            message.setSender(instance.getName());
            instance.setActive(false);
            instance.setEndedTime(new Date());

            ChannelServiceImpl
                    .sendMessageToPending(message.toString(), pending);

            removeUnLogMessages(instance);
            removeKeyValuePairs(instance);
            MemcacheService cache = MemcacheServiceFactory.getMemcacheService();
            cache.delete(instance.getPrimaryKey());
        } else
            updateGameState(instance, pending, player);

    }

    private void removeUnLogMessages(GameInstanceData instance) {
        PersistenceManager pm = PMF.getManager();
        try {
            Query q = pm.newQuery(PlayerMessageData.class,
                    "gameInstance == i && hasLogMessages == false");
            q.declareParameters(Key.class.getName() + " i");
            q.deletePersistentAll(instance.getServerKey());

        } finally {
            pm.close();
        }
    }

    private void removeKeyValuePairs(GameInstanceData instance) {
        PersistenceManager pm = PMF.getManager();
        try {
            Query q = pm.newQuery(GameKeyValuePairData.class,
                    "gameInstance == i");
            q.declareParameters(Key.class.getName() + " i");
            q.deletePersistentAll(instance.getServerKey());

        } finally {
            pm.close();
        }
    }

    private void updateGameState(GameInstanceData instance,
            PendingPlayerData pending, UserData sender) {
        if (!instance.joinable())
            return;
        MessageData message = new MessageData();
        message.setServerCommand(ServerCommand.GAME_UPDATED);
        message.setSender(instance.getName());

        HashMap<String, String> paras = new HashMap<String, String>();
        paras.put("name", instance.getName());
        paras.put("playerCount",
                String.valueOf(instance.getActivePlayerCount()));
        paras.put("size", String.valueOf(instance.getSize()));
        paras.put("token", KeyFactory.keyToString(instance.getServerKey()));
        message.setAttributeValuePairs(paras);
        ChannelServiceImpl.sendMessageToPending(message.toString(), pending);
    }

}
