/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package au.edu.unimelb.csse.mugle.server.multiplayer.api;

import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.PersistenceManager;

import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.KeyFactory;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

import au.edu.unimelb.csse.mugle.client.multiplayer.api.TransactionService;
import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.*;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.*;
import au.edu.unimelb.csse.mugle.shared.api.GameTokenError;
import au.edu.unimelb.csse.mugle.shared.api.KeyError;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.*;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.UserNotExists;

@SuppressWarnings("serial")
public class TransactionServiceImpl extends RemoteServiceServlet implements
        TransactionService {

    @Override
    public String createTransaction(String instanceToken,
            Transaction transaction) throws InstanceTokenError,
            GameInstanceNotExists, UserNotExists {
        GameInstanceData gameInstance = GameInstanceGetter
                .getInstanceByToken(instanceToken);
        String keyString = null;
        if (gameInstance.isActive()) {
            PersistenceManager pm = PMF.getManager();
            try {
                TransactionData td = new TransactionData();
                pm.makePersistent(td);
                td.setTransaction(transaction);
                keyString = KeyFactory.keyToString(td.getServerKey());
                UserData u = UserGetter.getCurrentUser();
                MessageData message = new MessageData();
                message.setServerCommand(ServerCommand.TRANSACTION_CREATED);
                message.setBody(transaction.toString());
                message.addAttribute("ID", keyString);
                message.setSender(u.getUrlName());
                UserData ud;

                for (String user : transaction.getUsers()) {
                    if (user != null && !user.equals(u.getUrlName())) {
                        ud = UserGetter.getUserByName(user);
                        td.addUser(ud.getServerKey());
                        ChannelServiceImpl.sendMessage(message.toString(), ud);
                    }
                }
            } finally {
                pm.close();
            }
        }
        return keyString;
    }

    @Override
    public void confirmTransaction(String instanceToken, String transactionId)
            throws InstanceTokenError, GameInstanceNotExists,
            TransactionNotExists, UserNotExists, KeyError {
        UserData u = UserGetter.getCurrentUser();
        Key key = KeyFactory.stringToKey(transactionId);
        PersistenceManager pm = PMF.getManager();
        try {
            pm.currentTransaction().begin();
            TransactionData td = pm.getObjectById(TransactionData.class, key);
            td.addConfirmedUsers(u.getServerKey());
            if (td.isAllConfirmed()) {
                Transaction transaction = td.getTransaction();
                runTransaction(pm, transaction);
            }
            pm.currentTransaction().commit();
        } catch (JDOObjectNotFoundException e) {
            throw new TransactionNotExists(transactionId);
        } finally {
            if (pm.currentTransaction().isActive()) {
                System.err.println("Transaction with ID: " + transactionId
                        + " failed");
                pm.currentTransaction().rollback();
            }
        }
    }

    private void runTransaction(PersistenceManager pm, Transaction transaction)
            throws UserNotExists, KeyError, InstanceTokenError {

        for (Action action : transaction.getActions())
            execute(pm, transaction, action);

    }

    private void execute(PersistenceManager pm, Transaction transaction,
            Action action) throws UserNotExists, KeyError, InstanceTokenError {
        if (action instanceof Transfer)
            execute(pm, transaction, (Transfer) action);
        else if (action instanceof Bidding)
            execute(pm, transaction, (Bidding) action);

    }

    private void execute(PersistenceManager pm, Transaction transaction,
            Transfer transfer) throws UserNotExists, KeyError,
            InstanceTokenError {
        KeyValuePairData transferFrom = null;
        KeyValuePairData transferTo = null;

        if (!transaction.isInstanceGameProfile()) {

            try {
                transferFrom = KeyValuePairGetter.getUserKeyValuePair(pm,
                        transfer.getTransferFrom(), transfer.getKey(),
                        transaction.getGameToken(), false);
                transferTo = KeyValuePairGetter.getUserKeyValuePair(pm,
                        transfer.getTransferTo(), transfer.getKey(),
                        transaction.getGameToken(), true);
            } catch (GameTokenError e) {
                // should never happen
                e.printStackTrace();
            }
        } else {
            transferFrom = KeyValuePairGetter.getUserInstanceKeyValuePair(pm,
                    transfer.getTransferFrom(), transfer.getKey(),
                    transaction.getInstanceToken(), false);
            transferTo = KeyValuePairGetter.getUserInstanceKeyValuePair(pm,
                    transfer.getTransferTo(), transfer.getKey(),
                    transaction.getInstanceToken(), true);
        }
        transferTo.setValue(transferFrom.getValue());
        pm.deletePersistent(transferFrom);
    }

    private void execute(PersistenceManager pm, Transaction transaction,
            Bidding bidding) {
        // TODO
        System.out.println("Executing bidding");
    }

}
