/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.multiplayer.model;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.jdo.annotations.IdGeneratorStrategy;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;
import javax.jdo.annotations.PrimaryKey;

import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.KeyFactory;

import au.edu.unimelb.csse.mugle.server.model.ModelDataClass;
import au.edu.unimelb.csse.mugle.server.model.annotations.GetterUserLevel;
import au.edu.unimelb.csse.mugle.server.model.annotations.MugleData;
import au.edu.unimelb.csse.mugle.shared.model.Role;
import au.edu.unimelb.csse.mugle.shared.multiplayer.model.GameInstance;

@MugleData
@PersistenceCapable
public class GameInstanceData extends
        ModelDataClass<Long, GameInstance, GameInstanceData> {
    @PrimaryKey
    @Persistent(valueStrategy = IdGeneratorStrategy.IDENTITY)
    private Key id;

    @Persistent
    private Key game;

    @Persistent
    private Set<Key> playerMessages;

    @Persistent
    private String name; // the same as the creator's url name

    @Persistent
    private Set<Key> activePlayers;

    @Persistent
    private Set<Key> exitPlayers;

    @Persistent
    private Set<Key> gameLogs;

    @Persistent
    private boolean active;

    @Persistent
    private boolean joinable;

    @Persistent
    private int size;

    @Persistent
    private Key createdUser;

    @Persistent
    private Date createdTime;

    @Persistent
    private Date endedTime;

    public GameInstanceData() {
        this.activePlayers = new HashSet<Key>();
        this.playerMessages = new HashSet<Key>();
        this.exitPlayers = new HashSet<Key>();
        this.gameLogs = new HashSet<Key>();
        this.setJoinable(true);
    }

    public String getToken() {
        return KeyFactory.keyToString(this.getServerKey());
    }

    public void addPlayer(Key player) {
        this.activePlayers.add(player);
    }

    public void removePlayer(Key player) {
        this.activePlayers.remove(player);
    }

    public boolean containPlayer(Key player) {
        return this.activePlayers.contains(player);
    }

    public void moveToExit(Key player) {
        if (this.activePlayers.contains(player))
            this.exitPlayers.add(player);
    }

    public Set<Key> getExitPlayers() {
        return this.exitPlayers;
    }

    public void removeExitPlayer(Key player) {
        this.exitPlayers.remove(player);
    }

    public boolean containExitPlayer(Key player) {
        return this.exitPlayers.contains(player);
    }

    public Set<Key> getGameLogs() {
        return gameLogs;
    }

    public void addGameLog(Key gameLog) {
        this.gameLogs.add(gameLog);
    }

    @GetterUserLevel(privateView = Role.GUEST, publicView = Role.GUEST, mappedTo = "setHasLog")
    public boolean hasLog() {
        return this.gameLogs.size() != 0;
    }

    @GetterUserLevel(privateView = Role.GUEST, publicView = Role.GUEST, mappedTo = "setActive")
    public boolean isActive() {
        return active;
    }

    public void setActive(boolean active) {
        this.active = active;
    }

    public int getActivePlayerCount() {
        return activePlayers.size();
    }

    @GetterUserLevel(privateView = Role.GUEST, publicView = Role.GUEST, mappedTo = "setPlayerCount")
    public int getAllPlayerCount() {
        return getActivePlayerCount() + getExitPlayers().size();
    }

    public boolean hasRoom() {
        return activePlayers.size() != this.size;
    }

    public void setGame(Key game) {
        this.game = game;
    }

    public Key getGame() {
        return game;
    }

    public Set<Key> getMessages() {
        return playerMessages;
    }

    public void addPlayerMessage(Key playerMessage) {
        this.activePlayers.add(playerMessage);
    }

    @GetterUserLevel(privateView = Role.GUEST, publicView = Role.GUEST, mappedTo = "setName")
    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public Set<Key> getActivePlayers() {
        return activePlayers;
    }

    public boolean joinable() {
        return joinable && getActivePlayerCount() <= size;
    }

    public void setJoinable(boolean joinable) {
        this.joinable = joinable;
    }

    public int getSize() {
        return size;
    }

    public void setSize(int size) {
        this.size = size;
    }

    @Override
    @GetterUserLevel(privateView = Role.GUEST, publicView = Role.GUEST, mappedTo = "setToken", overrideBy = "keyToToken")
    public Key getServerKey() {
        return this.id;
    }

    @Override
    public void setServerKey(Key key) {
        this.id = key;
    }

    @Override
    @GetterUserLevel(privateView = Role.GUEST, publicView = Role.GUEST, mappedTo = "setPrimaryKey")
    public Long getPrimaryKey() {
        if (this.id == null) {
            return null;
        }
        return this.id.getId();
    }

    @Override
    public Class<GameInstance> getClientClassType() {
        return GameInstance.class;
    }

    @Override
    protected void ClearFields() {
        throw new UnsupportedOperationException();
    }

    @Override
    public Key getCreatedUser() {
        return this.createdUser;
    }

    @Override
    @GetterUserLevel(privateView = Role.GUEST, publicView = Role.GUEST, mappedTo = "setCreatedTime")
    public Date getCreatedTime() {
        return this.createdTime;
    }

    @Override
    public Key getUpdatedUser() {
        throw new UnsupportedOperationException();
    }

    @Override
    public Date getUpdatedTime() {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setCreatedUser(Key user) {
        this.createdUser = user;
        addPlayer(user);
    }

    @Override
    public void setCreatedTime(Date time) {
        this.createdTime = time;
    }

    @Override
    public void setUpdatedUser(Key user) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void setUpdatedTime(Date time) {
        throw new UnsupportedOperationException();
    }

    @GetterUserLevel(privateView = Role.GUEST, publicView = Role.GUEST, mappedTo = "setEndedTime")
    public Date getEndedTime() {
        return endedTime;
    }

    public void setEndedTime(Date endedTime) {
        this.endedTime = endedTime;
    }

    public String keyToToken() {
        return KeyFactory.keyToString(this.id);
    }

}
