/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.multiplayer.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.UserData;

import com.google.appengine.api.datastore.Key;

public class PlayerMessageGetter {
    public static PlayerMessageData getMessageData(UserData sender,
            GameInstanceData instance) {
        PersistenceManager pm = PMF.getManager();
        try {
            return getMessageData(pm, sender, instance);
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    public static PlayerMessageData getMessageData(PersistenceManager pm,
            UserData sender, GameInstanceData instance) {
        Query q = pm.newQuery(PlayerMessageData.class,
                "gameInstance == i && sender == s");
        q.declareParameters(Key.class.getName() + " i, " + Key.class.getName()
                + " s");
        List<PlayerMessageData> results = (List<PlayerMessageData>) q.execute(
                instance.getServerKey(), sender.getServerKey());
        if (!results.isEmpty())
            return results.get(0);
        else {
            PlayerMessageData playerMessage = new PlayerMessageData();
            playerMessage.setGameInstance(instance.getServerKey());
            playerMessage.setSender(sender.getServerKey());
            pm.makePersistent(playerMessage);
            return playerMessage;
        }
    }

    public static int getMessageCounter(GameInstanceData instance) {
        PersistenceManager pm = PMF.getManager();
        try {
            return getMessageCounter(pm, instance);
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    public static int getMessageCounter(PersistenceManager pm,
            GameInstanceData instance) {
        pm.currentTransaction().begin();
        try {
            Query q = pm.newQuery(PlayerMessageData.class, "gameInstance == i");
            int counter = 1;
            q.declareParameters(Key.class.getName() + " i");
            List<PlayerMessageData> results = (List<PlayerMessageData>) q
                    .execute(instance.getServerKey());
            if (!results.isEmpty()) {
                for (PlayerMessageData messageData : results)
                    counter += messageData.getCounter();
                pm.currentTransaction().commit();
            }
            return counter;
        } finally {
            if (pm.currentTransaction().isActive())
                pm.currentTransaction().rollback();
        }
    }

    public static String getMessage(GameInstanceData instance, int counter) {
        PersistenceManager pm = PMF.getManager();
        try {
            return getMessage(pm, instance, counter);
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    public static String getMessage(PersistenceManager pm,
            GameInstanceData instance, int counter) {
        Query q = pm.newQuery(PlayerMessageData.class,
                "gameInstance == i && sequence == s");
        q.declareParameters(Key.class.getName() + " i, "
                + Integer.class.getName() + " s");
        List<PlayerMessageData> results = (List<PlayerMessageData>) q.execute(
                instance.getServerKey(), counter);
        if (!results.isEmpty()) {
            PlayerMessageData pmd = results.get(0);
            for (String message : pmd.getMessages()) {
                if (MessageData.getCounter(message) == counter)
                    return message;
            }
        }
        return null;
    }

    public static String[] getMessages(GameInstanceData instance, int[] sequence) {
        PersistenceManager pm = PMF.getManager();
        try {
            return getMessages(pm, instance, sequence);
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    public static String[] getMessages(PersistenceManager pm,
            GameInstanceData instance, int[] sequence) {
        Query q = pm.newQuery(PlayerMessageData.class,
                "gameInstance == i && sequence == s");
        q.declareParameters(Key.class.getName() + " i, "
                + Integer.class.getName() + " s");
        List<PlayerMessageData> results;
        ArrayList<String> messages = new ArrayList<String>();

        for (int i : sequence) {
            results = (List<PlayerMessageData>) q.execute(
                    instance.getServerKey(), i);
            if (!results.isEmpty()) {
                PlayerMessageData pmd = results.get(0);
                for (String message : pmd.getMessages()) {
                    if (MessageData.getCounter(message) == i)
                        messages.add(message);
                }
            }
        }
        if (messages.size() != 0)
            return messages.toArray(new String[messages.size()]);
        else
            return null;
    }

    @SuppressWarnings("unchecked")
    public static List<PlayerMessageData> getAllMessageData(
            PersistenceManager pm, GameInstanceData instance) {
        Query q = pm.newQuery(PlayerMessageData.class, "gameInstance == i");
        q.declareParameters(Key.class.getName() + " i");
        List<PlayerMessageData> results = (List<PlayerMessageData>) q
                .execute(instance.getServerKey());
        return results;
    }

    public static String[] getAllMessages(GameInstanceData instance) {
        PersistenceManager pm = PMF.getManager();
        try {
            return getAllMessages(pm, instance);
        } finally {
            pm.close();
        }
    }

    public static String[] getAllMessages(PersistenceManager pm,
            GameInstanceData instance) {
        List<PlayerMessageData> results = getAllMessageData(pm, instance);
        CounterComparator compare = new CounterComparator();
        if (!results.isEmpty()) {
            ArrayList<String> messages = new ArrayList<String>();
            for (PlayerMessageData messageData : results)
                messages.addAll(messageData.getMessages());
            Collections.sort(messages, compare);
            return messages.toArray(new String[messages.size()]);
        } else
            return null;
    }

    private static class CounterComparator implements Comparator<String> {
        @Override
        public int compare(String m1, String m2) {
            if (getCounter(m1) > getCounter(m2))
                return 1;
            return -1;
        }

        private int getCounter(String message) {
            return MessageData.getCounter(message);
        }
    }
}
