/*  Melbourne University Game-based Learning Environment
 *  Copyright (C) 2011 The University of Melbourne
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published
 *  by the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package au.edu.unimelb.csse.mugle.server.multiplayer.model.services;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;

import javax.jdo.PersistenceManager;
import javax.jdo.Query;

import au.edu.unimelb.csse.mugle.server.PMF;
import au.edu.unimelb.csse.mugle.server.model.GameGetter;
import au.edu.unimelb.csse.mugle.server.model.ModelWrapper;
import au.edu.unimelb.csse.mugle.server.model.RoleAndView;
import au.edu.unimelb.csse.mugle.server.model.UserData;
import au.edu.unimelb.csse.mugle.server.model.UserGetter;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.GameInstanceData;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.GameInstanceGetter;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.GameLogData;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.LogMessageData;
import au.edu.unimelb.csse.mugle.server.multiplayer.model.MessageData;
import au.edu.unimelb.csse.mugle.shared.model.ClientView;
import au.edu.unimelb.csse.mugle.shared.model.Game;
import au.edu.unimelb.csse.mugle.shared.model.ModelCollection;
import au.edu.unimelb.csse.mugle.shared.model.ModelList;
import au.edu.unimelb.csse.mugle.shared.model.Role;
import au.edu.unimelb.csse.mugle.shared.model.User;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.GameInstanceNotExists;
import au.edu.unimelb.csse.mugle.shared.multiplayer.api.InstanceTokenError;
import au.edu.unimelb.csse.mugle.shared.multiplayer.model.GameInstance;
import au.edu.unimelb.csse.mugle.shared.multiplayer.model.GameLog;
import au.edu.unimelb.csse.mugle.shared.multiplayer.model.LogMessage;
import au.edu.unimelb.csse.mugle.shared.multiplayer.model.services.GameMonitorService;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.GameNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.ServerException;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.UserNotExists;
import au.edu.unimelb.csse.mugle.shared.platform.exceptions.UserPrivilegeException;

import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.KeyFactory;
import com.google.gwt.user.server.rpc.RemoteServiceServlet;

@SuppressWarnings("serial")
public class GameMonitorServiceImpl extends RemoteServiceServlet implements
        GameMonitorService {

    @Override
    public ModelList<GameInstance> getGameInstances(Game game, int startIndex,
            int count) throws ServerException, UserPrivilegeException,
            GameNotExists {
        return getSortedGameInstances(game, startIndex, count, "name");
    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelList<GameInstance> getSortedGameInstances(Game game,
            int startIndex, int count, String sortBy) throws GameNotExists,
            ServerException, UserPrivilegeException {
        PersistenceManager pm = PMF.getManager();
        try {
            UserData curUser = checkPermission(pm);

            if (startIndex < 0 || count < 1)
                return new ModelList<GameInstance>();

            Query q = pm.newQuery(GameInstanceData.class, "game == g");
            q.declareParameters(Key.class.getName() + " g");
            q.setOrdering(sortBy + " asc");
            q.setRange(startIndex, startIndex + count);
            List<GameInstanceData> results = (List<GameInstanceData>) q
                    .execute(GameGetter.getGame(game.getPrimaryKey())
                            .getServerKey());

            if (results.isEmpty())
                return new ModelList<GameInstance>();

            int c = results.size();
            if (count < c)
                c = count;

            RoleAndView rv = new RoleAndView(curUser.getRole(),
                    ClientView.PUBLIC);

            ModelList<GameInstance> list = new ModelList<GameInstance>(c);
            list.setStartIndex(startIndex);
            list.setEndIndex(startIndex + c);
            list.setCapacity(results.size());
            for (int i = 0; i < c; i++) {
                GameInstance gi = ModelWrapper.getClientCloneSafe(
                        results.get(i), rv);
                list.getList().add(gi);
            }
            return list;

        } finally {
            pm.close();
        }
    }

    @Override
    public ModelCollection<User> getPlayers(GameInstance instance)
            throws ServerException, UserPrivilegeException,
            GameInstanceNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            UserData curUser = checkPermission(pm);

            RoleAndView rv = new RoleAndView(curUser.getRole(),
                    ClientView.PUBLIC);

            GameInstanceData id = GameInstanceGetter.getInstance(instance
                    .getPrimaryKey());
            HashSet<User> list = new HashSet<User>(id.getActivePlayers().size());
            HashSet<Key> players = new HashSet<Key>();
            players.addAll(id.getActivePlayers());
            players.addAll(id.getExitPlayers());

            for (Key k : players) {
                UserData ud;
                try {
                    ud = null;
                    ud = UserGetter.getUser(k.getId());
                    User wrapped = ModelWrapper.getClientCloneSafe(ud, rv);
                    list.add(wrapped);
                } catch (UserNotExists e) {
                    /*
                     * IF this happens it means the server has screwed up entity
                     * relations somewhere
                     */
                    throw new ServerException("Server Entity Relationship "
                            + "Management Error: " + e.getMessage());
                }
            }
            ModelCollection<User> toRet = new ModelCollection<User>();
            toRet.setCapacity(list.size());
            toRet.setEndIndex(list.size());
            toRet.setStartIndex(0);
            toRet.setCollection(list);
            return toRet;
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    @Override
    public int getInstanceCount(Game game) throws ServerException,
            UserPrivilegeException, GameNotExists {
        PersistenceManager pm = PMF.getManager();
        try {
            Query q = pm.newQuery(GameInstanceData.class, "game == g");
            q.declareParameters(Key.class.getName() + " g");
            List<GameInstanceData> results = (List<GameInstanceData>) q
                    .execute(GameGetter.getGame(game.getPrimaryKey())
                            .getServerKey());
            return results.size();
        } finally {
            pm.close();
        }
    }

    @Override
    public ModelList<GameLog> getGameLogs(String instanceToken)
            throws InstanceTokenError, ServerException, UserPrivilegeException {
        PersistenceManager pm = PMF.getManager();
        try {
            UserData curUser = checkPermission(pm);

            GameInstanceData gi = GameInstanceGetter.getInstanceByToken(pm,
                    instanceToken);
            ArrayList<GameLogData> results = new ArrayList<GameLogData>();

            for (Key k : gi.getGameLogs()) {
                GameLogData gl = pm.getObjectById(GameLogData.class, k);
                results.add(gl);
            }

            if (results.isEmpty())
                return new ModelList<GameLog>();

            int c = results.size();

            RoleAndView rv = new RoleAndView(curUser.getRole(),
                    ClientView.PUBLIC);

            ModelList<GameLog> list = new ModelList<GameLog>(c);
            list.setCapacity(results.size());
            for (int i = 0; i < c; i++) {
                GameLog gl = ModelWrapper
                        .getClientCloneSafe(results.get(i), rv);
                list.getList().add(gl);
            }
            return list;
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    public int getLogMessageCount(String logToken) throws ServerException {
        PersistenceManager pm = PMF.getManager();
        try {
            Query q = pm.newQuery(LogMessageData.class, "gameLog == gl");
            q.declareParameters(Key.class.getName() + " gl");
            q.setOrdering("id asc");
            Key k = KeyFactory.stringToKey(logToken);
            List<LogMessageData> results = (List<LogMessageData>) q.execute(k);
            int count = 0;
            for (LogMessageData lm : results) {
                count += lm.getMessages().size();
            }

            return count;
        } finally {
            pm.close();
        }
    }

    @SuppressWarnings("unchecked")
    @Override
    public ModelList<LogMessage> getLogMessages(String logToken,
            int startIndex, int count) throws ServerException,
            UserPrivilegeException {

        PersistenceManager pm = PMF.getManager();
        try {
            UserData curUser = checkPermission(pm);
            Query q = pm.newQuery(LogMessageData.class, "gameLog == gl");
            q.declareParameters(Key.class.getName() + " gl");
            q.setOrdering("id asc");
            Key k = KeyFactory.stringToKey(logToken);
            List<LogMessageData> results = (List<LogMessageData>) q.execute(k);

            RoleAndView rv = new RoleAndView(curUser.getRole(),
                    ClientView.PUBLIC);

            ModelList<LogMessage> list = new ModelList<LogMessage>();
            MessageData md = new MessageData();
            int index = 0;
            for (LogMessageData lmd : results) {
                for (String m : lmd.getMessages()) {
                    if (index++ < startIndex)
                        continue;
                    md.setMessageBean(MessageData.deserialiseFromJson(m));
                    LogMessage lm = ModelWrapper.getClientCloneSafe(md, rv);
                    list.getList().add(lm);
                    if (index - startIndex == count)
                        break;
                }
                if (index - startIndex == count)
                    break;
            }

            return list;
        } finally {
            pm.close();
        }
    }

    private UserData checkPermission(PersistenceManager pm)
            throws UserPrivilegeException {
        UserData curUser = UserGetter.getCurrentUser(pm);
        if (curUser.isActive() && curUser.getRole() != Role.ADMIN)
            throw new UserPrivilegeException("Only admins can monitor game!");
        return curUser;
    }

}
